%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%   Peter Hawkins  <hawkinsp@cs.stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Useful abstractions and short-cuts for CIL constructs

using translatecil.
using strops.
using intops.

% PREDICATES

% the type of machine we are to simulate. this affects the size and
% signedness of the various integer and floating point types. default is i386.
% if you want a different machine kind, add the declaration:
% +machine(x86_64).
type machine_kind ::=
    i386
  | x86_64.
predicate machine(machine_kind).

predicate ikind_bytes(ikind,bool,int).  % sign and byte size for an ikind
predicate ikind_bits(ikind,bool,int).   % sign and number of bits for an ikind
predicate fkind_bytes(fkind,int).       % byte size for an fkind
predicate fkind_bits(fkind,int).        % number of bits for an fkind
predicate pointer_bytes(int).           % byte size for a pointer 
predicate pointer_bits(int).            % number of bits for a pointer 

% lightweight type representation
type t_type ::=
    y_void
  | y_int{ikind}
  | y_flt{fkind}
  | y_ptr{t_type}
  | y_array{t_type,LEN:int}
  | y_func{RTYP:t_type,ATYPS:list[t_type],VARARG:bool}
  | y_comp{string}
  | y_enum{string}
  | y_valist.

% convert regular CIL types to the t_type representation.
% FN/C/G indicates which session this TYP is stored in
predicate convert_func_type(in FN:string,in TYP:c_type,Y:t_type)
    succeeds [once].
predicate convert_comp_type(in C:string,in TYP:c_type,Y:t_type)
    succeeds [once].
predicate convert_glob_type(in G:string,in TYP:c_type,Y:t_type)
    succeeds [once].
predicate convert_init_type(in G:string,in TYP:c_type,Y:t_type)
    succeeds [once].

% convert CIL types to t_type according to whichever func/glob/comp
% we are currently analyzing
predicate convert_type(in TYP:c_type,Y:t_type).

% get the width of a type, in bytes
predicate type_width(in Y:t_type,BS:int) succeeds [once].

% sign/length for an int/flt/ptr type. pointers are unsigned, floats are signed
predicate type_bits(in Y:t_type,SIGN:bool,BITS:int).

predicate lval_type(c_lval,t_type).        % lval -> type
predicate exp_type(c_exp,t_type).          % exp -> type
predicate off_type(c_offset,t_type).       % off -> result type
predicate base_off_type(c_offset,t_type).  % off -> base type

predicate exp_tptr(c_exp).        % pointer exps
predicate exp_tint(c_exp,ikind).  % integer exps
predicate exp_tflt(c_exp,fkind).  % floating point exps

% field F is the Nth field of type C, at byte offset BS and with type Y
predicate comp_field(in C:string,N:int,BS:int,F:string,Y:t_type).

% classify variables beyond the existing cil_var_global
% These predicates are declared but not inferred in the translatecil package.
% predicate cil_var_formal(X:c_var).
% predicate cil_var_local(X:c_var).
% predicate cil_var_temporary(X:c_var).
% predicate cil_var_return(X:c_var).

% get the full set of statements and instructions in the function
predicate cil_stmt(S:c_stmt).
predicate cil_instr(I:c_instr).

% I is a direct call to FN
predicate dircall(I:c_instr,FN:string).

% I is either a direct call to FN or an indirect call which *could* target FN
type calltype ::= direct | indirect.
predicate anycall(I:c_instr,FN:string,T:calltype).

% I is a call assigning to temporary X
predicate callretvar(I:c_instr,X:c_var).

% classify each operation
predicate logic_op(binop).        % boolean args, boolean result
predicate compare_op(binop).      % integer args, boolean result
predicate arith_op(binop).        % integer args, integer result
predicate pointer_op(binop).      % pointer LHS, integer RHS, pointer result
predicate pointer_int_op(binop).  % pointer args, integer result

% T0 and T1 are equivalent types
predicate type_alias(in T0:c_type,in T1:c_type).

% call instruction I implicitly casts its result to a value of type T
predicate call_cast(I:c_instr,T:c_type).

% identify loops in a function. we just identify these based on the
% corresponding CIL statement, either a native loop or the target of a goto
type loop = c_stmt.

% locations we commonly attach summary information to
type sum ::=
    s_func           % the current function
  | s_loop{loop}     % a particular loop in the current function
  | s_call{c_instr}  % a call instruction
  | s_asm{c_instr}   % an assembly instruction
  .

% loops in this function body
predicate loophead(loop).

% stringifying and output

predicate lval_string(c_lval,string).
predicate off_string(c_offset,string).
predicate exp_string(c_exp,string).
predicate type_string(c_type,string).
predicate location_string(in FILE:string,in LINE:int,string).

% clean up a global function or variable name, removing any 'static' variable
% uniquification /path/to/file.c:
predicate cleanup_glob_string(in G:string,GSTR:string).

% SUM is at FILE/LINE
predicate sum_location(in SUM:sum,FILE:string,LINE:int) succeeds [once].

% UI location information for types, globals, and the current function.
% get an appropriate MINL and MAXL for use in generating a display,
% and for the current function get the cleaned up function name FS
predicate comp_location(in C:string,FILE:string,MINL:int,MAXL:int) succeeds [once].
predicate glob_location(in G:string,FILE:string,MINL:int,MAXL:int) succeeds [once].
predicate func_location(in FN:string,FILE:string,MINL:int,MAXL:int) succeeds [once].

predicate curfn_location(FS:string,FILE:string,MINL:int,MAXL:int).
cil_curfn(FN), cleanup_glob_string(FN,FS), func_location(FN,FILE,MINL,MAXL),
    +curfn_location(FS,FILE,MINL,MAXL).

% SUMBODY SESSIONS

% session containing only the CFG for a particular loop or outer function body,
% as well as all the regular predicates in cil_body. SUM is s_func or s_loop.
% constructed by sumbody.clp
session cil_sum_body(FN:string,SUM:sum).

% the sum we are currently analyzing. this predicate is ONLY defined
% if iterating over the cil_sum_body session.
predicate cil_cursum(sum).

% shorthand for 'cil_curfn(FN), cil_cursum(SUM)', fetches both for cil_sum_body
predicate cursum(FN:string,SUM:sum).
cil_curfn(FN), cil_cursum(SUM), +cursum(FN,SUM).

% INDIRECT CALL SESSIONS

% indirect call target session
session sum_funptr(FN:string) containing [sindirect].
predicate sindirect(I:c_instr,CFN:string).  % call I to possible callee CFN

type t_call_context ::=
    cxt_any
  | cxt_call{PFN:string,PI:c_instr,PCXT:t_call_context}.

% result of refinement analysis which has narrowed down the possible targets,
% including calling-context sensitive information.
session sum_funptr_refine(FN:string) containing [sindirect_context].
predicate sindirect_context(I:c_instr,CXT:t_call_context,CFN:string).

% further refinement which has ruled out particular call sites as not reachble
% under certain calling contexts
session sum_funptr_omit(FN:string) containing [sindirect_omit].
predicate sindirect_omit(I:c_instr,CXT:t_call_context).

% client-extensible overriding of call site edges
predicate omit_call_edge(in FN:string,in I:c_instr,in CFN:string).

cil_curfn(CURFN), dircall(I,F), ~omit_call_edge(CURFN,I,F),
    +anycall(I,F,direct).
cil_curfn(CURFN), sum_funptr(CURFN)->sindirect(I,F), 
    ~sum_funptr_refine(CURFN)->sindirect_context(I,_,_),
    ~omit_call_edge(CURFN,I,F), +anycall(I,F,indirect).
cil_curfn(CURFN), sum_funptr_refine(CURFN)->sindirect_context(I,_,F),
    ~omit_call_edge(CURFN,I,F), +anycall(I,F,indirect).

% CALLGRAPH SESSIONS

% forward definition of program points, filled in by cilcfg.clp
type pp.

% cil_sum_caller.db and cil_sum_callee.db are filled in by sumcallers_infer.clp

% FN/SUM may be called through isum PI/PP within caller PFN/PSUM
session cil_sum_caller(FN:string,SUM:sum) containing [caller].
predicate caller(PFN:string,PSUM:sum,PP:pp,PI:sum).

% FN/SUM may invoke a callee CFN/CSUM through isum I/P
session cil_sum_callee(FN:string,SUM:sum) containing [callee].
predicate callee(P:pp,I:sum,CFN:string,CSUM:sum).

% FN is a root of the callgraph (i.e. has no callers)
predicate callgraph_root(in FN:string).
callgraph_root(FN) :- ~cil_sum_caller(FN,s_func)->caller(_,_,_,_).

% for baked-in behaviors ignoring particular call sites
predicate ignore_callgraph_site(in PFN:string,in PI:c_instr,in FN:string).

predicate function_caller(in FN:string,PFN:string,PI:c_instr).
function_caller(FN,PFN,PI) :-
    cil_sum_caller(FN,s_func)->caller(PFN,_,_,s_call{PI}),
    ~ignore_callgraph_site(PFN,PI,FN).

predicate function_callee(in FN:string,in SUM:sum,I:c_instr,CFN:string).
function_callee(FN,SUM,I,CFN) :-
    cil_sum_callee(FN,SUM)->callee(_,s_call{I},CFN,s_func).
function_callee(FN,SUM,I,CFN) :-
    cil_sum_callee(FN,SUM)->callee(_,s_loop{PH},FN,CSUM), CSUM\=s_loop{PH},
    function_callee(FN,CSUM,I,CFN).

predicate sum_caller(in FN:string,in SUM:sum,PFN:string,PSUM:sum,PP:pp,PI:sum).
sum_caller(FN,SUM,PFN,PSUM,PP,PI) :-
    cil_sum_caller(FN,SUM)->caller(PFN,PSUM,PP,PI),
    ~(PI=s_call{XPI}, ignore_callgraph_site(PFN,XPI,FN)).

% RULES

% cil modifications

predicate no_convert_start_addr().

% each cil_exp_start(E,LV) is equivalent to cil_exp_addr(E,LV[0]),
% make the conversion painful though it is
predicate want_convert_start_addr(OFF:c_offset).
predicate convert_start_addr(OFF:c_offset,NOFF:c_offset).
want_convert_start_addr(OFF),
    cil_off_none(OFF),
    cil_make_const(OFF,"cvtcst",C), +cil_const_int(C,iint,0),
    cil_make_exp(OFF,"cvtexp",E), +cil_exp_const(E,C), +cil_exp_x_intval(E,0),
    cil_make_offset(OFF,"empoff",EOFF), +cil_off_none(EOFF),
    cil_make_offset(OFF,"cvtoff",NOFF), +cil_off_index(NOFF,E,EOFF),
    +convert_start_addr(OFF,NOFF).
want_convert_start_addr(OFF),
    cil_off_field(OFF,F,FOFF),
    +want_convert_start_addr(FOFF),
    convert_start_addr(FOFF,NFOFF),
    cil_make_offset(OFF,"cvtoff",NOFF), +cil_off_field(NOFF,F,NFOFF),
    +convert_start_addr(OFF,NOFF).
want_convert_start_addr(OFF),
    cil_off_index(OFF,E,IOFF),
    +want_convert_start_addr(IOFF), convert_start_addr(IOFF,NIOFF),
    cil_make_offset(OFF,"cvtoff",NOFF), +cil_off_index(NOFF,E,NIOFF),
    +convert_start_addr(OFF,NOFF).
~no_convert_start_addr(), cil_exp_start(E,LV), cil_lval_var(LV,X,OFF),
    +want_convert_start_addr(OFF),
    convert_start_addr(OFF,NOFF), cil_make_lval(LV,"cvtlv",NLV),
    +cil_lval_var(NLV,X,NOFF), +cil_exp_addr(E,NLV).
~no_convert_start_addr(), cil_exp_start(E,LV), cil_lval_mem(LV,ME,OFF),
    +want_convert_start_addr(OFF),
    convert_start_addr(OFF,NOFF), cil_make_lval(LV,"cvtlv",NLV),
    +cil_lval_mem(NLV,ME,NOFF), +cil_exp_addr(E,NLV).

% convert all 'index' pointer arithmetic operations into 'plus' operations,
% which are equivalent
cil_exp_binop(E,b_indexpi,L,R,T), +cil_exp_binop(E,b_pluspi,L,R,T).

% we don't replace sizeof/alignof's with const expressions. these should be
% manually handled by client analyses. check for such non-const exps though
predicate cil_exp_x_sizeof_alignof(c_exp).
cil_exp_sizeof(E,_), +cil_exp_x_sizeof_alignof(E).
cil_exp_sizeofe(E,_), +cil_exp_x_sizeof_alignof(E).
cil_exp_sizeofstr(E,_), +cil_exp_x_sizeof_alignof(E).
cil_exp_alignof(E,_), +cil_exp_x_sizeof_alignof(E).
cil_exp_alignofe(E,_), +cil_exp_x_sizeof_alignof(E).

cil_exp_x_sizeof_alignof(E), ~cil_exp_x_intval(E,_),
    +warning("non-const sizeof/alignof exp",E).

% add temporary/local opflags to non-argument non-static locals
cil_fundec_formal(_,_,X), +cil_var_formal(X).
cil_fundec_local(_,X), cil_var_name(X,NAME,_), str_cat("ciltmp",_,NAME),
    +cil_var_temporary(X).
cil_fundec_local(_,X), ~cil_var_temporary(X), ~cil_var_static(X),
    +cil_var_local(X).

% get all statements in the function
cil_stmt_instrs(S,_), +cil_stmt(S).
cil_stmt_return(S), +cil_stmt(S).
cil_stmt_goto(S,_), +cil_stmt(S).
cil_stmt_break(S), +cil_stmt(S).
cil_stmt_continue(S), +cil_stmt(S).
cil_stmt_if(S,_,_,_), +cil_stmt(S).
cil_stmt_switch(S,_,_,_), +cil_stmt(S).
cil_stmt_loop(S,_), +cil_stmt(S).
cil_stmt_block(S,_), +cil_stmt(S).
cil_stmt_tryfinally(S), +cil_stmt(S).
cil_stmt_tryexcept(S), +cil_stmt(S).

% get all instructions in the function
cil_instr_set(I,_,_), +cil_instr(I).
cil_instr_call(I,_), +cil_instr(I).
cil_instr_asm(I,_), +cil_instr(I).

% classify ikind/fkind signs and sizes
predicate machine_int(machine_kind).
machine(K), +machine_int(K).
~machine(_), +machine_int(i386).

machine_int(i386), +ikind_bytes(ischar,true,1),
    +ikind_bytes(ichar,    true,1), +ikind_bytes(iuchar,    false,1),
    +ikind_bytes(ishort,   true,2), +ikind_bytes(iushort,   false,2),
    +ikind_bytes(iint,     true,4), +ikind_bytes(iuint,     false,4),
    +ikind_bytes(ilong,    true,4), +ikind_bytes(iulong,    false,4),
    +ikind_bytes(ilonglong,true,8), +ikind_bytes(iulonglong,false,8).
machine_int(i386), +fkind_bytes(ffloat,4), +fkind_bytes(fdouble,8), 
    +fkind_bytes(flongdouble,12).
machine_int(i386), +pointer_bytes(4).

machine_int(x86_64), +ikind_bytes(ischar,true,1),
    +ikind_bytes(ichar,    true,1), +ikind_bytes(iuchar,    false,1),
    +ikind_bytes(ishort,   true,2), +ikind_bytes(iushort,   false,2),
    +ikind_bytes(iint,     true,4), +ikind_bytes(iuint,     false,4),
    +ikind_bytes(ilong,    true,8), +ikind_bytes(iulong,    false,8),
    +ikind_bytes(ilonglong,true,8), +ikind_bytes(iulonglong,false,8).
machine_int(x86_64), +fkind_bytes(ffloat,4), +fkind_bytes(fdouble,8), 
    +fkind_bytes(flongdouble,16).
machine_int(x86_64), +pointer_bytes(8).

ikind_bits(Kind, Signed, Bytes * 8) :- ikind_bytes(Kind, Signed, Bytes).
fkind_bits(Kind, Bytes * 8) :- fkind_bytes(Kind, Bytes).
pointer_bits(Bytes * 8) :- pointer_bytes(Bytes).

% type conversion

predicate convert_func_type_args(in FN:string,in TYP:c_type,in SA:int,
                                 YLIST:list[t_type]).
convert_func_type_args(FN,TYP,SA,[]) :-
    ~cil_body(FN)->cil_type_func_arg(TYP,SA,_,_).
convert_func_type_args(FN,TYP,SA,[Y|TAIL]) :-
    cil_body(FN)->cil_type_func_arg(TYP,SA,_,ATYP),
    convert_func_type(FN,ATYP,Y),
    int_add(SA,1,NSA), convert_func_type_args(FN,TYP,NSA,TAIL).

convert_func_type(FN,TYP,y_void) :- cil_body(FN)->cil_type_void(TYP).
convert_func_type(FN,TYP,y_int{K}) :- cil_body(FN)->cil_type_int(TYP,K).
convert_func_type(FN,TYP,y_flt{K}) :- cil_body(FN)->cil_type_float(TYP,K).
convert_func_type(FN,TYP,y_ptr{DY}) :- 
    cil_body(FN)->cil_type_ptr(TYP,DTYP), convert_func_type(FN,DTYP,DY).
convert_func_type(FN,TYP,y_array{EY,N}) :- 
    cil_body(FN)->cil_type_array(TYP,ETYP,LEN),
    cil_body(FN)->cil_exp_x_intval(LEN,N),
    convert_func_type(FN,ETYP,EY).
convert_func_type(FN,TYP,y_func{RY,AYLIST,VA}) :-
    cil_body(FN)->cil_type_func(TYP,RTYP,VA),
    convert_func_type(FN,RTYP,RY),
    convert_func_type_args(FN,TYP,0,AYLIST).
convert_func_type(FN,TYP,y_comp{DC}) :- cil_body(FN)->cil_type_comp(TYP,DC).
convert_func_type(FN,TYP,y_enum{E}) :- cil_body(FN)->cil_type_enum(TYP,E).
convert_func_type(FN,TYP,y_valist) :- cil_body(FN)->cil_type_valist(TYP).
convert_func_type(FN,TYP,Y) :- 
    cil_body(FN)->cil_type_named(TYP,_,ATYP),
    convert_func_type(FN,ATYP,Y).

predicate convert_comp_type_args(in C:string,in TYP:c_type,in SA:int,
                                 YLIST:list[t_type]).
convert_comp_type_args(C,TYP,SA,[]) :-
    ~cil_comp(C)->cil_type_func_arg(TYP,SA,_,_).
convert_comp_type_args(C,TYP,SA,[Y|TAIL]) :-
    cil_comp(C)->cil_type_func_arg(TYP,SA,_,ATYP),
    convert_comp_type(C,ATYP,Y),
    int_add(SA,1,NSA), convert_comp_type_args(C,TYP,NSA,TAIL).

convert_comp_type(C,TYP,y_void) :- cil_comp(C)->cil_type_void(TYP).
convert_comp_type(C,TYP,y_int{K}) :- cil_comp(C)->cil_type_int(TYP,K).
convert_comp_type(C,TYP,y_flt{K}) :- cil_comp(C)->cil_type_float(TYP,K).
convert_comp_type(C,TYP,y_ptr{DY}) :- 
    cil_comp(C)->cil_type_ptr(TYP,DTYP), convert_comp_type(C,DTYP,DY).
convert_comp_type(C,TYP,y_array{EY,N}) :- 
    cil_comp(C)->cil_type_array(TYP,ETYP,LEN),
    cil_comp(C)->cil_exp_x_intval(LEN,N),
    convert_comp_type(C,ETYP,EY).
convert_comp_type(C,TYP,y_func{RY,AYLIST,VA}) :-
    cil_comp(C)->cil_type_func(TYP,RTYP,VA),
    convert_comp_type(C,RTYP,RY),
    convert_comp_type_args(C,TYP,0,AYLIST).
convert_comp_type(C,TYP,y_comp{DC}) :- cil_comp(C)->cil_type_comp(TYP,DC).
convert_comp_type(C,TYP,y_enum{E}) :- cil_comp(C)->cil_type_enum(TYP,E).
convert_comp_type(C,TYP,y_valist) :- cil_comp(C)->cil_type_valist(TYP).
convert_comp_type(C,TYP,Y) :- 
    cil_comp(C)->cil_type_named(TYP,_,ATYP),
    convert_comp_type(C,ATYP,Y).

predicate convert_glob_type_args(in G:string,in TYP:c_type,in SA:int,
                                 YLIST:list[t_type]).
convert_glob_type_args(G,TYP,SA,[]) :-
    ~cil_glob(G)->cil_type_func_arg(TYP,SA,_,_).
convert_glob_type_args(G,TYP,SA,[Y|TAIL]) :-
    cil_glob(G)->cil_type_func_arg(TYP,SA,_,ATYP),
    convert_glob_type(G,ATYP,Y),
    int_add(SA,1,NSA), convert_glob_type_args(G,TYP,NSA,TAIL).

convert_glob_type(G,TYP,y_void) :- cil_glob(G)->cil_type_void(TYP).
convert_glob_type(G,TYP,y_int{K}) :- cil_glob(G)->cil_type_int(TYP,K).
convert_glob_type(G,TYP,y_flt{K}) :- cil_glob(G)->cil_type_float(TYP,K).
convert_glob_type(G,TYP,y_ptr{DY}) :- 
    cil_glob(G)->cil_type_ptr(TYP,DTYP), convert_glob_type(G,DTYP,DY).
convert_glob_type(G,TYP,y_array{EY,N}) :- 
    cil_glob(G)->cil_type_array(TYP,ETYP,LEN),
    cil_glob(G)->cil_exp_x_intval(LEN,N),
    convert_glob_type(G,ETYP,EY).
convert_glob_type(G,TYP,y_func{RY,AYLIST,VA}) :-
    cil_glob(G)->cil_type_func(TYP,RTYP,VA),
    convert_glob_type(G,RTYP,RY),
    convert_glob_type_args(G,TYP,0,AYLIST).
convert_glob_type(G,TYP,y_comp{C}) :- cil_glob(G)->cil_type_comp(TYP,C).
convert_glob_type(G,TYP,y_enum{E}) :- cil_glob(G)->cil_type_enum(TYP,E).
convert_glob_type(G,TYP,y_valist) :- cil_glob(G)->cil_type_valist(TYP).
convert_glob_type(G,TYP,Y) :-
    cil_glob(G)->cil_type_named(TYP,_,ATYP),
    convert_glob_type(G,ATYP,Y).

predicate convert_init_type_args(in G:string,in TYP:c_type,in SA:int,
                                 YLIST:list[t_type]).
convert_init_type_args(G,TYP,SA,[]) :-
    ~cil_init(G)->cil_type_func_arg(TYP,SA,_,_).
convert_init_type_args(G,TYP,SA,[Y|TAIL]) :-
    cil_init(G)->cil_type_func_arg(TYP,SA,_,ATYP),
    convert_init_type(G,ATYP,Y),
    int_add(SA,1,NSA), convert_init_type_args(G,TYP,NSA,TAIL).

convert_init_type(G,TYP,y_void) :- cil_init(G)->cil_type_void(TYP).
convert_init_type(G,TYP,y_int{K}) :- cil_init(G)->cil_type_int(TYP,K).
convert_init_type(G,TYP,y_flt{K}) :- cil_init(G)->cil_type_float(TYP,K).
convert_init_type(G,TYP,y_ptr{DY}) :- 
    cil_init(G)->cil_type_ptr(TYP,DTYP), convert_init_type(G,DTYP,DY).
convert_init_type(G,TYP,y_array{EY,N}) :- 
    cil_init(G)->cil_type_array(TYP,ETYP,LEN),
    cil_init(G)->cil_exp_x_intval(LEN,N),
    convert_init_type(G,ETYP,EY).
convert_init_type(G,TYP,y_func{RY,AYLIST,VA}) :-
    cil_init(G)->cil_type_func(TYP,RTYP,VA),
    convert_init_type(G,RTYP,RY),
    convert_init_type_args(G,TYP,0,AYLIST).
convert_init_type(G,TYP,y_comp{C}) :- cil_init(G)->cil_type_comp(TYP,C).
convert_init_type(G,TYP,y_enum{E}) :- cil_init(G)->cil_type_enum(TYP,E).
convert_init_type(G,TYP,y_valist) :- cil_init(G)->cil_type_valist(TYP).
convert_init_type(G,TYP,Y) :-
    cil_init(G)->cil_type_named(TYP,_,ATYP),
    convert_init_type(G,ATYP,Y).

convert_type(TYP,Y) :- cil_curfn(FN), convert_func_type(FN,TYP,Y).
convert_type(TYP,Y) :- cil_comp(C), convert_comp_type(C,TYP,Y).
convert_type(TYP,Y) :- cil_glob(G), convert_glob_type(G,TYP,Y).
convert_type(TYP,Y) :- cil_init(G), convert_init_type(G,TYP,Y).

% type reverse conversion

% sometimes we need to convert t_type values back into c_type values.
% this is tricky because the new types that we fabricate are not stored
% in any databases, so convert_type will not work on them unless special cased

% the c_type for Y should be computed.
% must be added for all Y where reconvert_type is used
predicate do_reconvert_type(Y:t_type).

% TYP is the c_type for Y. this should always succeed except for y_func,
% where a warning will be generated. TODO: if this becomes a problem y_func
% needs to be expanded to include return and argument types
predicate reconvert_type(Y:t_type,TYP:c_type).

do_reconvert_type(Y), cil_make_type(Y,"type",TYP), +reconvert_type(Y,TYP),
    +convert_func_type(_,TYP,Y),
    +convert_comp_type(_,TYP,Y),
    +convert_glob_type(_,TYP,Y),
    +convert_init_type(_,TYP,Y).

reconvert_type(y_void,TYP), +cil_type_void(TYP).
reconvert_type(y_int{K},TYP), +cil_type_int(TYP,K).
reconvert_type(y_flt{K},TYP), +cil_type_float(TYP,K).
reconvert_type(y_ptr{DY},TYP), +do_reconvert_type(DY), reconvert_type(DY,DTYP),
    +cil_type_ptr(TYP,DTYP).
reconvert_type(y_array{EY,N},TYP), +do_reconvert_type(EY), reconvert_type(EY,ETYP),
    cil_make_const(N,"const",NC), cil_make_exp(N,"exp",NE),
    +cil_const_int(NC,iint,N), +cil_exp_const(NE,NC), +cil_exp_x_intval(NE,N),
    +cil_type_array(TYP,ETYP,NE).
reconvert_type(y_comp{C},TYP), +cil_type_comp(TYP,C).
reconvert_type(y_enum{E},TYP), +cil_type_enum(TYP,E).
reconvert_type(y_valist,TYP), +cil_type_valist(TYP).

reconvert_type(Y,TYP), type_width(Y,BS), +cil_type_x_bytes(TYP,BS).

% type widths

type_width(y_int{K},BS) :- ikind_bytes(K,_,BS).
type_width(y_enum{_},BS) :- ikind_bytes(iint,_,BS).
type_width(y_flt{K},BS) :- fkind_bytes(K,BS).
type_width(y_ptr{_},BS) :- pointer_bytes(BS).
type_width(y_array{Y,N},BS) :- type_width(Y,YBS), int_mul(YBS,N,BS).
type_width(y_comp{C},BS) :- cil_comp(C)->cil_comp_bytes(_,BS).

% void, function, and undefined composite types don't really have widths
+type_width(y_void,1), +type_width(y_func{_,_,_},1).
type_width(y_comp{C},1) :- ~cil_comp(C)->cil_comp_name(_,_,_).

% hack for databases that don't have cil_comp_bytes. TODO: remove me
type_width(y_comp{C},BS) :- cil_comp(C)->cil_comp_name(_,_,_),
    ~cil_comp(C)->cil_comp_bytes(_,_),
    cil_type_comp(TYP,C), cil_type_x_bytes(TYP,BS).

type_bits(y_int{K},S,LEN) :- ikind_bits(K,S,LEN).
type_bits(y_enum{_},S,LEN) :- ikind_bits(iint,S,LEN).
type_bits(y_flt{K},true,LEN) :- fkind_bits(K,LEN).
type_bits(y_ptr{_},false,LEN) :- pointer_bits(LEN).

% type fields

% get all the fields of a struct/union. we need to special case the
% __missing_field_name field, which is shared by all fields without names
% (i.e. padding fields which cannot be directly accessed)
comp_field(C,N,BS,F,Y) :-
    cil_comp(C)->cil_comp_name(COMP,C,_),
    cil_comp(C)->cil_comp_field(COMP,N,FLD),
    cil_comp(C)->cil_field_name(FLD,F,TYP,BS,_),
    (F="___missing_field_name", Y=y_void;
     F\="___missing_field_name", convert_comp_type(C,TYP,Y)).

% C has a value of type Y at offset zero, with FLIST field chain to reach Y.
% split cases based on whether C is a union type, where every field is at
% offset zero
predicate first_field(in C:string,Y:t_type,
                      STRUCT:bool,FLIST:list[t_pair[string,string]]).

first_field(C,Y,STRUCT,[pair{F,C}]) :-
    cil_comp(C)->cil_comp_name(_,_,STRUCT),
    (STRUCT=true, comp_field(C,0,_,F,Y);
     STRUCT=false, comp_field(C,_,_,F,Y)).

first_field(C,Y,XSTRUCT,[pair{F,C}|FLIST]) :-
    cil_comp(C)->cil_comp_name(_,_,STRUCT),
    (STRUCT=true, comp_field(C,0,_,F,y_comp{NC});
     STRUCT=false, comp_field(C,_,_,F,y_comp{NC})),
    first_field(NC,Y,TSTRUCT,FLIST),
    (STRUCT=true, TSTRUCT=true, XSTRUCT=true;
     STRUCT=false, XSTRUCT = false;
     TSTRUCT=false, XSTRUCT = false).

% lval/offset types

cil_lval_var(LV,X,OFF), cil_var_name(X,_,XT), convert_type(XT,XY),
    +base_off_type(OFF,XY), off_type(OFF,Y), +lval_type(LV,Y).
cil_lval_mem(LV,ME,OFF), exp_type(ME,y_ptr{DY}),
    +base_off_type(OFF,DY), off_type(OFF,Y), +lval_type(LV,Y).

cil_off_none(OFF), base_off_type(OFF,Y), +off_type(OFF,Y).
cil_off_field(OFF,F,NOFF), cil_field_name(F,_,FT,_,_), convert_type(FT,FY),
    +base_off_type(NOFF,FY), off_type(NOFF,Y), +off_type(OFF,Y).
cil_off_index(OFF,_,NOFF), base_off_type(OFF,y_array{EY,_}),
    +base_off_type(NOFF,EY), off_type(NOFF,Y), +off_type(OFF,Y).

% exp types

% get the types of constant expressions
cil_const_int(C,K,_), cil_exp_const(E,C), +exp_type(E,y_int{K}).
cil_const_char(C,_), cil_exp_const(E,C), +exp_type(E,y_int{iint}).
cil_const_str(C,_), cil_exp_const(E,C), +exp_type(E,y_ptr{y_int{ichar}}).
cil_const_wstr(C,_), cil_exp_const(E,C), +exp_type(E,y_ptr{y_int{iushort}}).
cil_const_real(C,K,_), cil_exp_const(E,C), +exp_type(E,y_flt{K}).
cil_exp_x_sizeof_alignof(E), +exp_type(E,y_int{iuint}).

% lval exps have same type as the lval
cil_exp_lval(E,LV), lval_type(LV,Y), +exp_type(E,Y).

% address exps have type of pointer to the lval type
cil_exp_addr(E,LV), lval_type(LV,DY), +exp_type(E,y_ptr{DY}).

% types for cast/unop/binop come straight off the exp
cil_exp_cast(E,_,T), convert_type(T,Y), +exp_type(E,Y).
cil_exp_unop(E,_,_,T), convert_type(T,Y), +exp_type(E,Y).
cil_exp_binop(E,_,_,_,T), convert_type(T,Y), +exp_type(E,Y).

% get pointer/integer/float exps
exp_type(E,y_ptr{_}), +exp_tptr(E).
exp_type(E,y_int{K}), +exp_tint(E,K).
exp_type(E,y_enum{_}), +exp_tint(E,iint).
exp_type(E,y_flt{K}), +exp_tflt(E,K).

% classify ops

+logic_op(b_land), +logic_op(b_lor).
+compare_op(b_eq), +compare_op(b_ne), +compare_op(b_lt), +compare_op(b_gt),
    +compare_op(b_le), +compare_op(b_ge).
+arith_op(b_plusa), +arith_op(b_minusa),
    +arith_op(b_band), +arith_op(b_bxor), +arith_op(b_bor),
    +arith_op(b_mult), +arith_op(b_div), +arith_op(b_mod),
    +arith_op(b_shiftlt), +arith_op(b_shiftrt).
+pointer_op(b_pluspi), +pointer_op(b_minuspi).
+pointer_int_op(b_minuspp).

% function call predicates

% identify direct calls
cil_instr_call(I,FEXP), cil_exp_lval(FEXP,LV),
    cil_lval_var(LV,X,OFF), cil_off_none(OFF), cil_var_name(X,FN,_),
    +dircall(I,FN).

% identify call return values
cil_instr_call_ret(I,RLV), cil_lval_var(RLV,X,ROFF), cil_off_none(ROFF),
    cil_var_temporary(X), +callretvar(I,X).

% stringifying

% convert an exp to a string using exp_string, wrapping binops with '()'
predicate b_exp_string(in c_exp,string).
b_exp_string(E,S) :- exp_string(E,S), ~cil_exp_binop(E,_,_,_,_).
b_exp_string(E,S) :- exp_string(E,SS), cil_exp_binop(E,_,_,_,_),
    S = "(" ^ SS ^ ")".

% stringify lvalues
cil_lval_var(LV,X,OFF), cil_var_name(X,XS,_), cleanup_glob_string(XS,NXS),
    off_string(OFF,OS), +lval_string(LV,NXS ^ OS).
cil_lval_mem(LV,ME,OFF), cil_off_none(OFF), b_exp_string(ME,MS),
    +lval_string(LV, "*" ^ MS).
cil_lval_mem(LV,ME,OFF), cil_off_field(OFF,F,NOFF), b_exp_string(ME,MS),
    cil_field_name(F,FS,_,_,_), off_string(NOFF,OS),
    S = MS ^ "->" ^ FS ^ OS, +lval_string(LV,S).
cil_lval_mem(LV,ME,OFF), cil_off_index(OFF,_,_), b_exp_string(ME,MS),
    off_string(OFF,OS), +lval_string(LV, MS ^ OS).

% stringify offsets
cil_off_none(OFF), +off_string(OFF,"").
cil_off_field(OFF,F,NOFF), cil_field_name(F,FS,_,_,_), off_string(NOFF,OS),
    S = "." ^ FS ^ OS, +off_string(OFF,S).
cil_off_index(OFF,E,NOFF), exp_string(E,IS), off_string(NOFF,OS),
    S = "[" ^ IS ^ "]" ^ OS, +off_string(OFF,S).

% stringify constants
predicate const_string(c_const,string).
cil_const_int(C,_,N), tostring(N,S), +const_string(C,S).
cil_const_char(C,N), tostring(N,S), +const_string(C,S).
cil_const_str(C,S), R = "\"" ^ S ^ "\"", +const_string(C,R).
cil_const_wstr(C,_), +const_string(C,"wchar_t[]").
cil_const_real(C,_,N), tostring(N,S), +const_string(C,S).

% stringify unops
predicate unop_str(unop,string).
+unop_str(u_neg,"-"), +unop_str(u_bnot,"~"), +unop_str(u_lnot,"!").

% stringify binops
predicate binop_str(binop,string).
+binop_str(b_plusa,"+"), +binop_str(b_minusa,"-"),
    +binop_str(b_pluspi,"+"),
    +binop_str(b_minuspi,"-"), +binop_str(b_minuspp,"-"),
    +binop_str(b_mult,"*"), +binop_str(b_div,"/"), +binop_str(b_mod,"%"),
    +binop_str(b_shiftlt,"<<"), +binop_str(b_shiftrt,">>"),
    +binop_str(b_lt,"<"), +binop_str(b_gt,">"),
    +binop_str(b_le,"<="), +binop_str(b_ge,">="),
    +binop_str(b_eq,"=="), +binop_str(b_ne,"!="),
    +binop_str(b_band,"&"), +binop_str(b_bxor,"^"), +binop_str(b_bor,"|"),
    +binop_str(b_land,"&&"), +binop_str(b_lor,"||").

% stringify expressions
cil_exp_const(E,C), const_string(C,S), +exp_string(E,S).
cil_exp_lval(E,LV), lval_string(LV,S), +exp_string(E,S).
cil_exp_sizeof(E,_), cil_exp_x_intval(E,N), tostring(N,S), +exp_string(E,S).
cil_exp_sizeofe(E,_), cil_exp_x_intval(E,N), tostring(N,S), +exp_string(E,S).
cil_exp_sizeofstr(E,_), cil_exp_x_intval(E,N), tostring(N,S), +exp_string(E,S).
cil_exp_alignof(E,_), cil_exp_x_intval(E,N), tostring(N,S), +exp_string(E,S).
cil_exp_alignofe(E,_), cil_exp_x_intval(E,N), tostring(N,S), +exp_string(E,S).
cil_exp_addr(E,LV), lval_string(LV,LS), +exp_string(E,"&" ^ LS).
cil_exp_cast(E,CE,T), type_string(T,TS), b_exp_string(CE,CS),
    +exp_string(E, "(" ^ TS ^ ")" ^ CS).
cil_exp_unop(E,OP,RE,_), unop_str(OP,US), b_exp_string(RE,RS),
    +exp_string(E,US ^ RS).
cil_exp_binop(E,OP,LE,RE,_), binop_str(OP,US),
    b_exp_string(LE,LS), b_exp_string(RE,RS), +exp_string(E,LS ^ US ^ RS).

% stringify integer kinds
predicate ikind_str(ikind,string).
+ikind_str(ichar,"char"),
    +ikind_str(ischar,"char"), +ikind_str(iuchar,"u_char"),
    +ikind_str(iint,"int"), +ikind_str(iuint,"u_int"),
    +ikind_str(ishort,"short"), +ikind_str(iushort,"u_short"),
    +ikind_str(ilong,"long"), +ikind_str(iulong,"u_long"),
    +ikind_str(ilonglong,"longlong"),
    +ikind_str(iulonglong,"u_longlong").

% stringify float kinds
predicate fkind_str(fkind,string).
+fkind_str(ffloat,"float"), +fkind_str(fdouble,"double"),
    +fkind_str(flongdouble,"longdouble").

% stringify types

cil_type_void(T), +type_string(T,"void").
cil_type_int(T,K), ikind_str(K,S), +type_string(T,S).
cil_type_float(T,K), fkind_str(K,S), +type_string(T,S).
cil_type_ptr(T,DT), type_string(DT,DS), +type_string(T,DS ^ "*").
cil_type_array(T,ET,_), type_string(ET,ES), +type_string(T,ES ^ "[]").
cil_type_func(T,_,_), +type_string(T,"function").
cil_type_comp(T,NAME), +type_string(T,"struct " ^ NAME).
cil_type_enum(T,NAME), +type_string(T,"enum " ^ NAME).
cil_type_named(T,_,NT), type_string(NT,S), +type_string(T,S).
cil_type_valist(T), +type_string(T,"builtin_va_list").

% other stringification and output

location_string(FILE,LINE, FILE ^ ":" ^ tostring(LINE)) :- .

% clean functions by stripping everything before the ':' if present.
% ':' can't appear in a regular function name
cleanup_glob_string(G,GSTR) :-
    str_sub(G,POS,1,":"), str_sub(G,0,POS + 1,STATIC),
    str_cat(STATIC,GSTR,G).
cleanup_glob_string(G,G) :- ~str_sub(G,_,1,":").

sum_location(s_func,FILE,LINE) :- cil_fundec_location(_,FILE,LINE).
sum_location(s_loop{PH},FILE,LINE) :- cil_stmt_x_location(PH,FILE,LINE).
sum_location(s_call{I},FILE,LINE) :- cil_instr_x_location(I,FILE,LINE).
sum_location(s_asm{I},FILE,LINE) :- cil_instr_x_location(I,FILE,LINE).

% get composite type locations
comp_location(C,FILE,MINLINE - 5,MAXLINE + 5) :-
    cil_comp(C)->cil_comp_location(_,FILE,MINLINE),
    cil_comp(C)->cil_comp_end_location(_,_,MAXLINE).

% get global variable locations. for variables with static initializers look at the
% initializer expressions to find a suitable end location
glob_location(G,FILE,MINLINE - 5,MAXLINE + 5) :-
    cil_glob(G)->cil_var_name(_,_,_),
    cil_glob(G)->cil_var_location(_,FILE,MINLINE),
    (~cil_init(G)->cil_exp_x_location(_,_,_,_,_,_), MAXLINE=MINLINE;
     \/cil_init(G)->cil_exp_x_location(_,_,L,_,_,_):int_max_all(L,MAXLINE)).

% get function locations
func_location(FN,FILE,MINLINE - 5,MAXLINE + 5) :-
    cil_body(FN)->cil_fundec_location(_,FILE,MINLINE),
    cil_body(FN)->cil_fundec_end_location(_,_,MAXLINE).
