import "../smemory/traces.clp".
import "../smemory/memory.clp".

% SESSIONS

type t_trace_loc ::=
    tl_func{FN:string,T:t_trace,CXT:t_call_context}
  | tl_glob{T:t_trace}
  | tl_comp{C:string,RT:t_trace}.

type t_escape_where ::=
    te_func{FN:string,SUM:sum,P:pp}
  | te_glob{G:string}.

% forward/backward propagation edges for L. only includes cxt_any for tl_func.
% SL/TL are always tl_func{...} for assignments within functions,
% or tl_glob{...} for assignments from static initializers
session salias_escape(L:t_trace_loc,FWD:bool) containing [sescape].
predicate sescape(W:t_escape_where,SL:t_trace_loc,TL:t_trace_loc).

% all assignments performed by the static initializer for IG
session salias_init(IG:string) containing [initial_value].
predicate initial_value(T:t_trace,V:scalar).

% direct uses of locations in a read, write, index, rfld or fld operation
session salias_used(L:t_trace_loc)
    containing [sread,swrite,sdrf,sindex,srfld,sfld].
predicate sread(W:t_escape_where,L:t_trace_loc).
predicate swrite(W:t_escape_where,L:t_trace_loc).
predicate sdrf(W:t_escape_where,L:t_trace_loc).
predicate sindex(W:t_escape_where,L:t_trace_loc).
predicate srfld(W:t_escape_where,L:t_trace_loc,C:string,F:string).
predicate sfld(W:t_escape_where,L:t_trace_loc,C:string,F:string).

% rfld/fld information indexed by the field being rfld'ed/fld'ed
session salias_rfld(C:string,F:string) containing [srfld].
session salias_fld(C:string,F:string) containing [sfld].

% side effect information for each FN/SUM
session salias_write_func(FN:string,SUM:sum)
    containing [wfunc_impure,wfunc_write,
                wfunc_write_inline,wfunc_allocate,wfunc_indirect].
predicate wfunc_impure().
predicate wfunc_write(T:t_trace).
predicate wfunc_write_inline(T:t_trace,V:scalar).
predicate wfunc_allocate(T:t_trace,ESCAPE:bool).
predicate wfunc_indirect().  % this function might make indirect calls

% initialization information for each C/F
session salias_write_comp(C:string,F:string)
    containing [wcomp_impure,wcomp_init].
predicate wcomp_impure(FN:string).
predicate wcomp_init(FN:string,T:t_trace).

% per-location relative write information for each C/F.
% when the structure is embedded at L, writes to it may occur through WL
session salias_comp_relative(C:string,F:string)
    containing [wcomp_nonrelative,wcomp_relative].
predicate wcomp_nonrelative(WL:t_trace_loc).
predicate wcomp_relative(L:t_trace_loc,WL:t_trace_loc).

% argument trace T passed to FN will not be used for a write, either within FN
% or in any of its callers through locations where T escapes to
session salias_func_readonly(FN:string)
    containing [wfunc_readonly,wfunc_localwrite,wfunc_nonescape].
predicate wfunc_readonly(T:t_trace).
predicate wfunc_localwrite(T:t_trace).
predicate wfunc_nonescape(T:t_trace).

% POLYDATA ANALYSIS SESSIONS

session spoly_comp(C:string,F:string) containing [wpoly_search_comp].
session spoly_func(FN:string) containing [wpoly_search_func,wpoly_write_func].

% we are looking for correlations between FT and T relative to XC.
% writes to fld{STR,F,C} may affect this relationship, where STR is at
% offset RT from the value of type XC
predicate wpoly_search_comp(XC:string,RT:t_trace,
                            FT:t_trace,T:t_trace,AT:t_trace).

% we are looking for correlations between FT and T as passed into FN
predicate wpoly_search_func(FT:t_trace,T:t_trace,AT:t_trace).

% there is a write of V to T that affects some wpoly_search_comp,
% and which should be inlined in all callers. this is a may-update
predicate wpoly_write_func(T:t_trace,V:t_trace).

session spoly_receive_comp(C:string,FT:t_trace,T:t_trace,AT:t_trace)
    containing [wpoly_receive,wpoly_fail].
session spoly_receive_func(FN:string,FT:t_trace,T:t_trace,AT:t_trace)
    containing [wpoly_receive,wpoly_fail].
predicate wpoly_receive(FNPTR:string,FN:string,SUM:sum,P:pp,DT:t_trace).
predicate wpoly_fail().

session spoly_receive_attribute(C:string,F:string,AT:t_trace)
    containing [wpoly_receive].
predicate receive_attribute(in FN:string,in I:c_instr,
                            in PFT:t_trace,in AT:t_trace,
                            C:string,F:string).

% the indirect call at FN/I is done polymorphically, such that argument
% trace AT corresponds to trace T and the function pointer used corresponds
% to trace FT
session spoly_call(FN:string,I:c_instr)
    containing [wpoly_call_target,wpoly_call_comp,wpoly_call_func].
predicate wpoly_call_target(FT:t_trace).
predicate wpoly_call_comp(FT:t_trace,C:string,FT:t_trace,T:t_trace,AT:t_trace).
predicate wpoly_call_func(FT:t_trace,T:t_trace,AT:t_trace).

% AT is an interesting trace which should be used to fill in spoly_call.
% poly_data.clp will use a normal set of rules to populate the argument
% traces AT to consider; this session can be used to add more candidates
% before polydata is run. Y indicates the expected type of AT.
% any relationships which violate this typing (tested with check_cast)
% will lead to propagation failure
session spoly_call_try(FN:string) containing [wpoly_call_try].
predicate wpoly_call_try(AT:t_trace,TYP:t_type).

% STRINGIFY PREDICATES

predicate trace_loc_string(in L:t_trace_loc,LS:string) succeeds [once].

trace_loc_string(tl_func{FN,T,CXT},STR) :-
    tostring(FN,FNS), trace_string(T,TS), tostring(CXT,CXTS),
    STR = "tl_func{"^FNS^","^TS^","^CXTS^"}".
trace_loc_string(tl_glob{T},STR) :-
    trace_string(T,TS),
    STR = "tl_glob{"^TS^"}".
trace_loc_string(tl_comp{C,RT},STR) :-
    tostring(C,CS), trace_string(RT,RTS),
    STR = "tl_comp{"^CS^","^RTS^"}".

predicate trace_loc_sub(in L:t_trace_loc,NL:t_trace_loc,RT:t_trace).
trace_loc_sub(tl_func{FN,T,CXT},tl_func{FN,ST,CXT},RT) :- trace_sub(T,ST,RT).
trace_loc_sub(tl_glob{T},tl_glob{ST},RT) :- trace_sub(T,ST,RT).
trace_loc_sub(tl_comp{C,T},tl_comp{C,ST},RT) :- trace_sub(T,ST,RT).

predicate trace_loc_compose(in L:t_trace_loc,in RT:t_trace,NL:t_trace_loc).
trace_loc_compose(tl_func{FN,T,CXT},RT,tl_func{FN,NT,CXT}) :-
    trace_compose(T,RT,NT).
trace_loc_compose(tl_glob{T},RT,tl_glob{NT}) :- trace_compose(T,RT,NT).
trace_loc_compose(tl_comp{C,T},RT,tl_comp{C,NT}) :- trace_compose(T,RT,NT).

% PREDICATES

predicate trace_loc_trace(in L:t_trace_loc,T:t_trace).
+trace_loc_trace(tl_comp{_,T},T).
+trace_loc_trace(tl_glob{T},T).
+trace_loc_trace(tl_func{_,T,_},T).

% I is an allocation function FN returning fresh data. WRAP indicates this
% is a wrapper for some other primitive allocator
predicate allocate_call(in I:c_instr,in P:pp,
                        FN:string,CT:t_trace,WRAP:bool,ESCAPE:bool).
allocate_call(I,P,FN,root{return},false,false) :-
    isum_target_p(P,s_call{I},FN,s_func,direct),
    (FN="malloc";
     FN="mm/slab.c:__cache_alloc";
     FN="include/linux/mm.h:page_zone";
     FN="__alloc_pages";
     FN="vmap";
     FN="__vmalloc_area";
     FN="mempool_alloc";
     FN="__builtin_alloca").  % CIL inserted function call
allocate_call(I,P,FN,CT,true,ESCAPE) :-
    isum_target_p(P,s_call{I},FN,s_func,direct),
    salias_write_func(FN,s_func)->wfunc_allocate(CT,ESCAPE).

% T is the root of a location freshly allocated within I through CT.
% CT is the location in the callee where the pointer was stored;
% T is the target of the pointer after the call.
% ESCAPE indicates that T might have otherwise escaped within the allocation
predicate allocate_root(T:t_trace,I:c_instr,CT:t_trace,ESCAPE:bool).
icall(P,_,I), allocate_call(I,P,_,CT,_,ESCAPE),
    inst_trace(s_call{I},P,CT,s_trace{T},G),
    guard(P,PG), guard_implies(PG,G),
    +allocate_root(uc_sum{s_call{I},P,T},I,CT,ESCAPE).

% T is a location allocated within this function which might have escaped
% at or before point P, except for possibly being directly assigned to XT
% (XT may be empty). if allocate_escape does not hold then the fields of T
% are effectively local variables at point P
predicate allocate_escape(in T:t_trace,in P:pp,in XT:t_trace).
allocate_escape(T,_,_) :-
    allocate_root(T,_,_,true).
allocate_escape(T,P,_) :-
    allocate_root(T,I,_,_), icall(BP,_,I), BP\=P,
    icall(XP,_,XI), pp_reach(BP,XP), pp_xreach(XP,P),
    argument(XP,XI,AT,s_trace{VT},_), trace_fld_sub(VT,T,_), AT=root{arg{_}},
    anycall(XI,CFN,_), ~salias_func_readonly(CFN)->wfunc_nonescape(AT),
    ~memcpy_func(CFN,_,_,_), ~memset_func(CFN,_,_,_).
allocate_escape(T,P,XT) :-
    allocate_root(T,I,_,_), icall(BP,_,I), BP\=P,
    iset(XP,_,_), pp_reach(BP,XP), pp_xreach(XP,P),
    assign(XP,WXT,s_trace{VT},_), trace_fld_sub(VT,T,_), WXT\=XT,
    ~trace_alloc(s_func,WXT,ak_stack_local).

% locations cast under conditions where they indicate an error code,
% not an actual value (as indicated by a call to ERR_PTR or a call to IS_ERR
% which returns true). these pointers can't be dereferenced
predicate error_trace(T:t_trace,G:g_guard).
dircall(I,"include/linux/err.h:ERR_PTR"), icall(P,_,I),
    inst_trace(s_call{I},P,drf{root{arg{0}}},s_trace{T},_),
    guard(P,G), +error_trace(T,G).
dircall(I,"include/linux/err.h:ERR_PTR"), icall(P,_,I),
    callret(P,I,RT,_), T=uc_sum{s_call{I},P,RT},
    guard(P,G), +error_trace(T,G).
dircall(I,"include/linux/err.h:IS_ERR"), icall(P,_,I),
    inst_trace(s_call{I},P,drf{root{arg{0}}},s_trace{T},_),
    callret(P,I,RT,_), cfg_assign_write(P,RT,PV,_),
    #id_g(sc_eqz{PV,true,32},NG1), #not(NG1,G0),
    guard(P,G1), #and(G0,G1,G), +error_trace(T,G).

% T within FN/SUM was passed out through a call I and callee trace CT
predicate match_call_exit(in FN:string,in SUM:sum,in T:t_trace,
                          I:sum,P:pp,CFN:string,CSUM:sum,CT:t_trace).
match_call_exit(FN,SUM,T,s_call{I},P,CFN,s_func,drf{root{return}}) :-
    T=uc_sum{s_call{I},P,BT}, smemory_remote(FN,SUM)->callret(P,I,BT,_),
    isum_target_p(p_remote{FN,SUM,P},s_call{I},CFN,s_func,direct).
match_call_exit(FN,SUM,uc_sum{I,P,T},I,P,CFN,CSUM,drf{CT}) :-
    isum_target_p(p_remote{FN,SUM,P},I,CFN,CSUM,direct),
    (salias_write_func(CFN,CSUM)->wfunc_write(CT);
     salias_write_func(CFN,CSUM)->wfunc_allocate(CT,_)),
    trace_root(CT,arg{_}), inst_trace(I,p_remote{FN,SUM,P},CT,s_trace{T},G),
    guard(p_remote{FN,SUM,P},PG), guard_implies(PG,G).
match_call_exit(FN,SUM,drf{T},I,P,CFN,CSUM,drf{CT}) :-
    match_call_exit(FN,SUM,T,I,P,CFN,CSUM,CT).
match_call_exit(FN,SUM,uc_sum{I,P,T},I,P,CFN,CSUM,drf{CT}) :-
    match_call_exit(FN,SUM,T,I,P,CFN,CSUM,CT).
match_call_exit(FN,SUM,fld{T,F,C},I,P,CFN,CSUM,FCT) :-
    match_call_exit(FN,SUM,T,I,P,CFN,CSUM,CT),
    trace_simplify(fld{CT,F,C},FCT).
match_call_exit(FN,SUM,rfld{T,F,C},I,P,CFN,CSUM,FCT) :-
    match_call_exit(FN,SUM,T,I,P,CFN,CSUM,CT),
    trace_simplify(rfld{CT,F,C},FCT).
match_call_exit(FN,SUM,index{T,Y,N},I,P,CFN,CSUM,ICT) :- N=s_const{_},
    match_call_exit(FN,SUM,T,I,P,CFN,CSUM,CT),
    trace_simplify(index{CT,Y,N},ICT).

% T is part of a just-allocated location
predicate allocate_filter(in FN:string,in SUM:sum,in T:t_trace).
allocate_filter(FN,SUM,T) :- match_call_exit(FN,SUM,T,_,_,CFN,CSUM,CT),
    trace_sub(CT,fld{drf{PCT},_,_},_),
    salias_write_func(CFN,CSUM)->wfunc_allocate(PCT,_),
    ~trace_type(CFN,PCT,y_ptr{y_comp{_}}).
allocate_filter(_,s_func,T) :- trace_root(T,local{_}).

% location of the very important but hardware alignment dependent
% netdevice private field
predicate netdev_priv_field(PRIV:string).
+netdev_priv_field("$net_device$704").
%netdev_priv_field(PRIV) :-
%    FN="include/linux/netdevice.h:netdev_priv",
%    salias_write_func(FN,s_func)->wfunc_write_inline(root{return},s_trace{T}),
%    T=fld{drf{root{arg{0}}},PRIV,"net_device"}.

% PRIVT is a private data trace for net_device PT
predicate netdev_priv(in PT:t_trace,PRIVT:t_trace).
netdev_priv(PT,PRIVT) :-
    netdev_priv_field(PRIV),
    PRIVT=fld{PT,PRIV,"net_device"}.
netdev_priv(PT,PRIVT) :-
    PRIVT=drf{fld{PT,"priv","net_device"}}.

% internal to polydata analysis, ignore type polymorphism with relative trace T
predicate poly_candidate_ignore(in T:t_trace).
poly_candidate_ignore(T) :- trace_sub(T,index{_,_,I},_), I\=s_const{0}.
poly_candidate_ignore(T) :- trace_sub(T,rfld{_,_,C},_),
    % exception for usb drivers
    C\="usb_driver".
poly_candidate_ignore(T) :- trace_derefs(T,N), int_gt(N,3).
poly_candidate_ignore(T) :-
    trace_sub(T,fld{STR,_,C},_), STR\=empty,
    cil_curfn(FN), ~trace_type(FN,STR,y_comp{C}),
    % exception for hdlc net devices
    C\="hdlcdrv_state".

predicate poly_caller_fail(in FN:string,in T:t_trace).
poly_caller_fail(FN,_) :- function_caller(FN,PFN,PI),
    ~spoly_call(PFN,PI)->wpoly_call_target(_).
poly_caller_fail(FN,T) :- function_caller(FN,PFN,PI),
    spoly_call(PFN,PI)->wpoly_call_target(PFT),
    ~spoly_call(PFN,PI)->wpoly_call_comp(PFT,_,_,_,T),
    ~spoly_call(PFN,PI)->wpoly_call_func(PFT,_,T).
poly_caller_fail(FN,T) :- function_caller(FN,PFN,PI),
    spoly_call(PFN,PI)->wpoly_call_target(PFT),
    spoly_call(PFN,PI)->wpoly_call_comp(PFT,C,FT,DT,T),
    spoly_receive_comp(C,FT,DT,T)->wpoly_fail().
poly_caller_fail(FN,T) :- function_caller(FN,PFN,PI),
    spoly_call(PFN,PI)->wpoly_call_target(PFT),
    spoly_call(PFN,PI)->wpoly_call_func(PFT,DT,T),
    spoly_receive_func(PFN,PFT,DT,T)->wpoly_fail().

% attribute functions
+receive_attribute("drivers/base/class.c:class_device_attr_show",_,_,drf{root{arg{0}}},"class_device_attribute","show").
+receive_attribute("drivers/base/class.c:class_device_attr_store",_,_,drf{root{arg{0}}},"class_device_attribute","store").
+receive_attribute("drivers/base/core.c:dev_attr_show",_,_,drf{root{arg{0}}},"device_attribute","show").
+receive_attribute("drivers/base/core.c:dev_attr_store",_,_,drf{root{arg{0}}},"device_attribute","store").
+receive_attribute("block/genhd.c:disk_attr_show",_,_,drf{root{arg{0}}},"disk_attribute","show").
+receive_attribute("block/genhd.c:disk_attr_store",_,_,drf{root{arg{0}}},"disk_attribute","store").
+receive_attribute("block/elevator.c:elv_attr_show",_,_,drf{root{arg{0}}},"elv_fs_entry","show").
+receive_attribute("block/elevator.c:elv_attr_store",_,_,drf{root{arg{0}}},"elv_fs_entry","store").
