% check per-function safety

import "typecheck.clp".
import "../daliasing/zerofield.clp".
import "../base/sumcallers.clp".

analyze session_name("cil_sum_body").
analyze analysis_name("casting_func").

analyze priority(1).

% PREDICATES

type h_check ::=
     % check always holds
    hc_true{P:maybe[pp],TEXT:string}
     % check never holds
  | hc_false{P:maybe[pp],TEXT:string}
     % G is unsatisfiable at point P
  | hc_unsat{P:pp,G:g_guard}
     % T has no explicit type or an explicit type compatible with V
  | hc_explicit_type{T:t_trace,V:t_typed_val}
     % uses of T with type V are compatible with the value FV fixed for T
  | hc_explicit_fixed{T:t_trace,V:t_typed_val}
     % there is an explicit use of EXP with value compatible with V at P
  | hc_explicit_use{P:pp,T:t_trace,V:t_typed_val}
     % sanitized forms for propagating TP to callers or glob/type
  | hc_quantify_sanitize{ST:t_trace,QTP:t_property}
  | hc_inst_sanitize{TP:t_property}
  | hc_glob_sanitize{TP:t_property}
  | hc_comp_sanitize{TP:t_property}
     % taking address NT of internal trace T cannot lead to future writes
     % affecting TAIL
  | hc_impact{UK:t_update_where,L:t_trace_loc,NL:t_trace_loc,TAIL:t_trace}
     % writing XV to XT at P nulls out the pointer XT
  | hc_null_assign{P:pp,TP:t_property,XT:t_trace,XV:scalar}
     % there is no direct conflicting use of T with XV/XG
  | hc_direct_use{P:pp,T:t_trace,V:t_typed_val,G:g_guard,
                  XV:t_typed_val,XG:g_guard}
     % all correlations between FT/T have been identified by polydata
  | hc_full_poly_comp_receive{C:string,FT:t_trace,T:t_trace,AT:t_trace}
  | hc_full_poly_func_receive{FN:string,FT:t_trace,T:t_trace,AT:t_trace}
     % TP holds at call site within caller PFN/PSUM
  | hc_caller{TP:t_property,PFN:string,PSUM:sum,PI:sum,PP:pp}
     % CTP holds at exit from callee CFN/CSUM
  | hc_callee{P:pp,CFN:string,CSUM:sum,CTP:t_property}
     % TP holds for some other function FN/SUM at P.
     % TP is relative to FN/SUM entry
  | hc_function{FN:string,SUM:sum,TP:t_property,P:pp}
     % ditto for hk_future properties
  | hc_function_future{FN:string,SUM:sum,PT:t_trace,V:t_typed_val,P:pp}
     % TP is a global invariant over GLOB
  | hc_glob_invariant{P:maybe[pp],GLOB:string,TP:t_property}
     % TP is a type invariant over COMP
  | hc_comp_invariant{P:maybe[pp],COMP:string,TP:t_property}
     % HK holds within this same summary
  | hc_sub_kind{HK:h_kind}
  .

% RULES

% fill in current location info for typecheck.clp

predicate explicit_kind(HK:h_kind,P:pp,T:t_trace,V:t_typed_val).
cursum(FN,SUM), scasting_explicit(FN,SUM)->sexplicit(P,T,V,G,_),
    +explicit_kind(hk_prior{l_value{T,V,G},P,true},P,T,V).

cursum(FN,SUM), +curloc(hl_func{FN,SUM}),
    sum_location(SUM,_,LINE), +curloc_line(LINE),
    curfn_location(SNAME,FILE,MINLINE,MAXLINE), dstring_sum(SUM,STITLE),
    +scasting_post(hl_func{FN,SUM})->
        sloc_info(SNAME,STITLE,FILE,MINLINE,MAXLINE),
    casting_display_name(hl_func{FN,SUM},td_casting_overall,DNAME),
    +search_add(SNAME,DNAME),
    explicit_kind(HK,_,_,_), hash_kind(HK,HKH),
    casting_display_name(hl_func{FN,SUM},td_casting_kind{HKH},NDNAME),
    +search_add(SNAME,NDNAME).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING FAILURES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hc_true

predicate maybe_point_line(in maybe[pp],out maybe[int]).
maybe_point_line(no,no) :- .
maybe_point_line(yes{P},yes{LINE}) :- point_location(P,_,LINE).

check_info(hc_true{MP,TEXT},MLINE,TEXT) :- maybe_point_line(MP,MLINE).

% ------------------------------------ hc_false

+check_fail(hc_false{_,_}).

check_info(hc_false{MP,TEXT},MLINE,TEXT) :- maybe_point_line(MP,MLINE).

% ------------------------------------ hc_unsat

predicate unsat_point_guard(in P:pp,in G:g_guard).
unsat_point_guard(P,G) :- guard(P,PG), #and(PG,G,MG), ~guard_sat(MG).

check_fail(hc_unsat{P,G}) :- ~unsat_point_guard(P,G).

check_info(hc_unsat{P,G},yes{LINE},TEXT) :-
    point_location(P,_,LINE), guard_string(G,GS),
    TEXT = GS^" is unsatisfiable".

% ------------------------------------ hc_explicit_type

check_fail(hc_explicit_type{T,C}) :-
    explicit_type(T,XC), ~known_compatible_types(C,XC).

check_info(hc_explicit_type{T,V},no,TEXT) :-
    trace_string(T,TS), typed_val_dstring(V,VS),
    TEXT = TS^" has explicit type compatible with "^VS.

% ------------------------------------ hc_explicit_fixed

check_fail(hc_explicit_fixed{T,C}) :-
    cursum(FN,SUM), scasting_explicit(FN,SUM)->sexplicit_fixed(T,XC),
    ~known_compatible_types(C,XC).

check_info(hc_explicit_fixed{T,V},no,TEXT) :-
    trace_string(T,TS), typed_val_dstring(V,VS),
    TEXT = TS^" has fixed type compatible with "^VS.

% ------------------------------------ hc_explicit_use

check_fail(hc_explicit_use{P,T,C}) :-
    explicit_kind(_,P,T,XC), ~known_compatible_types(C,XC).

check_info(hc_explicit_use{P,T,V},yes{LINE},TEXT) :-
    point_location(P,_,LINE), trace_string(T,TS), typed_val_dstring(V,VS),
    TEXT = "Earlier access to "^TS^" is an explicit use of "^VS.

% ------------------------------------ hc_inst_sanitize

predicate unsanitary_inst(in TP:t_property).
unsanitary_inst(TP) :-
    property_trace_used(TP,T0,_), property_trace_used(TP,T1,_), T0\=T1,
    trace_sub(T0,fld{STR,_,_},_), trace_sub(T1,STR,_).
unsanitary_inst(TP) :-
    ~(property_trace_used(TP,T,_), ~trace_root(T,glob{_})).
unsanitary_inst(TP) :-
    property_trace_used(TP,T,_), cursum(_,SUM),
    ~trace_alloc(SUM,T,ak_external).
unsanitary_inst(l_value{T,_,_}) :- trace_root(T,glob{_}); trace_root(T,cstr{_}).
unsanitary_inst(l_value{T,_,G}) :- ~#bool_g(true,G), T\=drf{root{_}}.

check_fail(hc_inst_sanitize{TP}) :- unsanitary_inst(TP).
check_info(hc_inst_sanitize{TP},no,TEXT) :-
    property_dstring(TP,TPS),
    TEXT = "Interesting bottom-up property: "^TPS.

% ------------------------------------ hc_glob_sanitize

predicate unsanitary_glob(in TP:t_property).
unsanitary_glob(TP) :-
    property_trace_used(TP,T,_),
    ~trace_root(T,glob{_}).
unsanitary_glob(TP) :-  % needed for soundness
    ~(property_trace_used(TP,T,_), ~trace_derefs(T,0)).

check_fail(hc_glob_sanitize{TP}) :- unsanitary_glob(TP).
check_info(hc_glob_sanitize{TP},no,TEXT) :-
    property_dstring(TP,TPS),
    TEXT = "Interesting global property: "^TPS.

% ------------------------------------ hc_comp_sanitize

predicate unsanitary_comp(in TP:t_property).
unsanitary_comp(TP) :-
    property_trace_used(TP,T,_),
    ~trace_relative(T), ~trace_root(T,glob{_}).
unsanitary_comp(TP) :-
    property_trace_used(TP,T,_), trace_sub(T,rfld{empty,_,_},_).
unsanitary_comp(TP) :-
    property_trace_used(TP,T,_), trace_sub(T,fld{XT,_,_},_),
    XT\=empty, XT\=index{empty,_,_},
    cil_curfn(FN), (~trace_type(FN,XT,_); trace_type(FN,XT,y_void)).
unsanitary_comp(TP) :-
    property_trace_used(TP,T0,_), property_trace_used(TP,T1,_),
    trace_sub(T0,fld{empty,_,C0},_), trace_sub(T1,fld{empty,_,C1},_), C0\=C1.
unsanitary_comp(TP) :-  % needed for soundness
    ~(property_trace_used(TP,T,_), trace_sub(T,fld{empty,_,_},RT),
      ~trace_derefs(RT,0)).
unsanitary_comp(l_value{T,_,_}) :- ~trace_relative(T).
unsanitary_comp(l_value{T,_,G}) :- ~#bool_g(true,G),
    trace_derefs(T,N), int_gt(N,1).

% don't want type invariants involving recursive structures
unsanitary_comp(TP) :-
    property_trace_used(TP,T,_), trace_sub(T,fld{_,_,"list_entry"},_).

check_fail(hc_comp_sanitize{TP}) :- unsanitary_comp(TP).
check_info(hc_comp_sanitize{TP},no,TEXT) :-
    property_dstring(TP,TPS),
    TEXT = "Interesting type property: "^TPS.

% ------------------------------------ hc_impact

check_fail(hc_impact{_,L,NL,TAIL}) :- ~trace_loc_any_write(L,NL,TAIL,_).
check_fail(hc_impact{_,L,NL,TAIL}) :- trace_loc_any_write(L,NL,TAIL,NLS), ~set_empty(NLS).
check_info(hc_impact{UK,_,_,_},no,TEXT) :-
    update_where_dstring(UK,UKS),
    TEXT = UKS^" does not impact whether property holds".

% ------------------------------------ hc_null_assign

predicate null_assign(in TP:t_property,in XT:t_trace,in XV:scalar).
null_assign(TP,XT,s_const{_}) :- property_trace_used(TP,drf{XT},true).
null_assign(TP,XT,s_const{_}) :- property_trace_used(TP,SXT,_),
    trace_sub(SXT,drf{XT},RT), RT\=empty.

check_fail(hc_null_assign{_,TP,XT,XV}) :- ~null_assign(TP,XT,XV).

check_info(hc_null_assign{P,TP,XT,XV},yes{LINE},TEXT) :-
    point_location(P,_,LINE), trace_string(XT,XTS),
    TEXT = "Write nulls out pointer "^XTS.

% ------------------------------------ hc_direct_use

check_fail(hc_direct_use{P,T,V,G,XV,XG}) :-
    ~known_compatible_types(V,XV), #and(G,XG,MG), guard_sat(MG).

check_info(hc_direct_use{P,T,V,G,XV,XG},yes{LINE},TEXT) :-
    point_location(P,_,LINE),
    trace_string(T,TS), typed_when_dstring(V,G,VS),
    TEXT = TS^" is directly used with type compatible with "^VS.

% ------------------------------------ hc_full_poly_comp_receive

check_fail(hc_full_poly_comp_receive{C,FT,DT,AT}) :-
    spoly_receive_comp(C,FT,DT,AT)->wpoly_fail().

check_info(hc_full_poly_comp_receive{C,FT,DT,_},no,TEXT) :-
    trace_string(FT,FTS), trace_string(DT,DTS),
    TEXT = "Full polymorphic type propagation for "^C
         ^ ": funptr "^FTS^", data "^DTS.

% ------------------------------------ hc_full_poly_func_receive

check_fail(hc_full_poly_func_receive{FN,FT,DT,AT}) :-
    spoly_receive_func(FN,FT,DT,AT)->wpoly_fail().

check_info(hc_full_poly_func_receive{FN,FT,DT,_},no,TEXT) :-
    trace_string(FT,FTS), trace_string(DT,DTS),
    TEXT = "Full polymorphic function propagation for "^FN
         ^ ": funptr "^FTS^", data "^DTS.

% ------------------------------------ hc_caller

check_require(hc_caller{TP,PFN,PSUM,PI,PP},XHKL,XHK) :-
    XHKL = hl_func{PFN,PSUM},
    (cursum(PFN,PSUM), RECURSE=true;
     ~cursum(PFN,PSUM), RECURSE=false),
    XHK = hk_prior_inst{PI,PP,TP,RECURSE}.

check_info(hc_caller{TP,PFN,PSUM,PI,PP},no,TEXT) :-
    property_dstring(TP,TPS),
    (cursum(PFN,PSUM), TEXT = "Holds for recursive caller that "^TPS;
     ~cursum(PFN,PSUM), TEXT = "Holds for caller that "^TPS).

% ------------------------------------ hc_callee

check_require(hc_callee{P,CFN,CSUM,CTP},XHKL,XHK) :-
    XHKL = hl_func{CFN,CSUM}, XHK = hk_prior_exit{CTP}.

check_info(hc_callee{P,CFN,CSUM,CTP},yes{LINE},TEXT) :-
    point_location(P,_,LINE), property_dstring(CTP,CTPS),
    (cursum(CFN,_), dstring_sum(CSUM,IS); ~cursum(CFN,_), IS=CFN),
    TEXT = "Holds for callee "^IS^" that "^CTPS.

% ------------------------------------ hc_function

check_require(hc_function{FN,SUM,TP,P},XHKL,XHK) :-
    XHKL = hl_func{FN,SUM}, XHK = hk_prior{TP,P,true}.

check_info(hc_function{FN,SUM,TP,_},no,TEXT) :-
    property_dstring(TP,TPS),
    TEXT = "Holds for function "^FN^" that "^TPS.

% ------------------------------------ hc_function_future

check_require(hc_function_future{FN,SUM,PT,V,P},XHKL,XHK) :-
    XHKL = hl_func{FN,SUM}, XHK = hk_future{PT,V,P,false}.

check_info(hc_function_future{FN,SUM,PT,V,P},no,TEXT) :-
    trace_string(PT,PTS), tostring(V,VS),
    TEXT = "Holds for function "^FN^" that future writes to "^PTS
         ^ " preserve type of "^VS.

% ------------------------------------ hc_glob_invariant

check_require(hc_glob_invariant{_,GLOB,TP},XHKL,XHK) :-
    XHKL = hl_glob{GLOB}, XHK = hk_invariant{TP}.

check_info(hc_glob_invariant{MP,GLOB,TP},MLINE,TEXT) :-
    maybe_point_line(MP,MLINE), property_dstring(TP,TPS),
    TEXT = "Global invariant that "^TPS.

% ------------------------------------ hc_comp_invariant

check_require(hc_comp_invariant{_,COMP,TP},XHKL,XHK) :-
    XHKL = hl_comp{COMP}, XHK = hk_invariant{TP}.

check_info(hc_comp_invariant{MP,COMP,TP},MLINE,TEXT) :-
    maybe_point_line(MP,MLINE), property_dstring(TP,TPS),
    TEXT = "Type invariant over "^COMP^" that "^TPS.

% ------------------------------------ hc_sub_kind

check_require(hc_sub_kind{HK},HKL,HK) :- curloc(HKL).

check_info(hc_sub_kind{HK},yes{LINE},TEXT) :-
    kind_line(HK,LINE), kind_dstring(HK,HKS), TEXT = "Prove "^HKS.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING METHODS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hk_prior

% traces used under an unsatisfiable guard are safe
trivial_property(l_value{T,V,G},P,HC,no) :-
    unsat_point_guard(P,G),
    HC = hc_unsat{P,G}.

% traces with a matching or conflicting explicit type don't need further work
trivial_property(l_value{T,V,G},P,HC,yes{hx_explicit_type}) :-
    ~unsat_point_guard(P,G), known_explicit_super_type(T,V),
    HC = hc_explicit_type{T,V}.

% similarly, handle an explicit type fixed at allocation for the trace
trivial_property(l_value{T,V,G},P,HC,yes{HX}) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    cursum(FN,SUM), scasting_explicit(FN,SUM)->sexplicit_fixed(T,_),
    HC = hc_explicit_fixed{T,V}, HX = hx_explicit_fixed.
trivial_property(l_value{XT,V,G},P,HC,yes{HX}) :-
    XT=index{T,_,s_const{0}},
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    cursum(FN,SUM), scasting_explicit(FN,SUM)->sexplicit_fixed(T,_),
    HC = hc_explicit_fixed{T,V}, HX = hx_explicit_fixed.
trivial_property(l_value{XT,V,G},P,HC,yes{HX}) :-
    XT=uc_sum{_,_,PT}, PT=fld{_,_,_},
    ~unsat_point_guard(P,G), ~known_explicit_super_type(drf{PT},V),
    cursum(FN,SUM), scasting_explicit(FN,SUM)->sexplicit_fixed(drf{PT},_),
    (HC = hc_explicit_fixed{drf{PT},V};
     HC = hc_sub_kind{hk_future{PT,V,P,true}}),
    HX = hx_explicit_fixed.

% accesses we've previously proved safe are safe
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    cursum(FN,SUM), scasting_explicit_safe(FN,SUM)->sexplicit(P,T,V,G,_),
    HC = hc_true{yes{P},"Access proved safe in a previous run"}.

% handle accesses to untyped fields (int[] or hidden) which are only ever
% accessed as this type, rendering them defacto instances of that type
trivial_property(l_value{T,V,G},P,HC,yes{HX}) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    T=fld{_,F,C},
    scasting_explicit_field(C,F)->sexplicit_field(V,_,_),
    ~(scasting_explicit_field(C,F)->sexplicit_field(XV,_,_), V\=XV),
    HC = hc_true{no,"Untyped field is only ever used with this type"},
    HX = hx_explicit_fixed.

% dominated by an earlier explicit use of the trace
trivial_property(l_value{T,V,G},P,HC,yes{HX}) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    explicit_kind(_,XP,T,_), P\=XP, pp_dominate(XP,P),
    HC = hc_explicit_use{XP,T,V}, HX = hx_explicit_use.

% contradicted by an earlier assertion on the trace
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    cursum(FN,SUM),
    scasting_restrict_assign(FN,SUM)->wrestrict_assign(XP,PT,XV,_,_),
    val(P,PT,s_trace{T},XG), guard_sat(XG),
    ~check_subtype(y_comp{V},y_comp{XV}),
    HC = hc_false{yes{XP},"Trace asserted as a different type"}.

% locations which are known to be NULL. these have no type
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    trace_sub(T,ST,RT), ~trace_derefs(RT,0), allocate_root(ST,I,CT,_),
    icall(XP,_,I), allocate_call(I,XP,FN,CT,WRAP,_),
    (WRAP=false;
     WRAP=true, scasting_allocate_untyped(FN)->wfunc_allocate(CT,_)),
    HC = hc_true{yes{XP},"Uninitialized data in allocation by "^FN}.
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    trace_sub(T,uc_sum{s_call{I},_,XPT},_), trace_sub(XPT,ST,_),
    allocate_root(ST,I,CT,_),
    icall(XP,_,I), allocate_call(I,XP,FN,CT,_,_),
    HC = hc_true{yes{XP},"Uninitialized data in allocation by "^FN}.
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    trace_sub(T,ST,_), zero_initialized(P,ST), trace_string(ST,STS),
    HC = hc_true{no,"Known zero data: "^STS}.

% locations produced by the ERR_PTR linux macro.
trivial_property(l_value{T,V,G},P,HC,no) :-
    ~unsat_point_guard(P,G), ~known_explicit_super_type(T,V),
    trace_sub(T,ST,_), error_trace(ST,_),
    \/error_trace(ST,EG):#or_all(EG,MEG),
    guard(P,PG), #and(PG,G,NG), guard_implies(NG,MEG),
    HC = hc_true{no,"Location indicates an error code"}.

predicate interesting_guard_trace(in T:t_trace,in XT:t_trace).
interesting_guard_trace(drf{root{R}},drf{root{XR}}) :-
    R\=glob{_}, XR\=glob{_}, R\=temp{_,_,_}, XR\=temp{_,_,_}.
interesting_guard_trace(T,drf{fld{STR,_,_}}) :- trace_fld_sub(T,STR,_).
interesting_guard_trace(drf{T},drf{fld{STR,_,_}}) :- trace_fld_sub(T,STR,_).

predicate interesting_guard(in T:t_trace,in P:pp,G:g_guard).
interesting_guard(T,P,G) :-
    guard(P,PG), #g_id(PG,ID), #id_g(ID,XG), (XG=G; #not(XG,G)),
    (ID=sc_eqz{_,_,_}; ID=sc_cmp{b_eq,_,_,_,_}; ID=sc_cmp{b_ne,_,_,_,_}),
    \/guard_used(G,su_scalar{s_trace{MT}}):list_all(MT,[XT]),
    trace_remove_uc_sum(T,NT), trace_remove_uc_sum(XT,NXT),
    interesting_guard_trace(NT,NXT), guard_implies(PG,G).

% point guard includes tests which we are interested in.
% split out any candidate component guard
method_check(hk_prior{TP,P,true},hm_guard{GS},HC) :-
    ~trivial_property(TP,P,_,_), TP=l_value{T,V,_},
    interesting_guard(T,P,G),
    ~(interesting_guard(T,P,XG), XG\=G, guard_implies(XG,G)),
    NTP=l_value{T,V,G}, guard_string(G,GS),
    HC = hc_sub_kind{hk_prior{NTP,P,false}}.

% traces passed back out through a callee. check within the callee that the
% trace has the corresponding type
method_check(hk_prior{TP,P,_},hm_callee{CFN},HC) :-
    ~trivial_property(TP,P,_,_), TP=l_value{T,V,_},
    cursum(FN,SUM), match_call_exit(FN,SUM,T,_,CP,CFN,CSUM,CT),
    #bool_g(true,TG), CTP=l_value{CT,V,TG},
    HC = hc_callee{CP,CFN,CSUM,CTP}.

% traces which might (or might not) have been modified by a callee.
% assume here that any assignments performed within the callee preserve the
% type of the location
method_check(hk_prior{TP,P,_},hm_preserve,HC) :-
    ~trivial_property(TP,P,_,_), TP=l_value{T,V,G}, impure_value(T,_,_),
    (follow_impure_values(P,s_trace{T},NP,s_trace{NT}),
         HC = hc_sub_kind{hk_prior{l_value{NT,V,G},NP,false}};
     ~follow_impure_values(P,s_trace{T},_,s_trace{_}),
         HC = hc_true{no,"No trace value prior to overwrite by call"}).
method_check(hk_prior{TP,P,_},hm_preserve,HC) :-
    ~trivial_property(TP,P,_,_), TP=l_value{T,_,_}, impure_value(T,_,PT),
    TP=l_value{_,V,_}, cursum(_,SUM), sum_bound(SUM,XP,_),
    HC = hc_sub_kind{hk_future{PT,V,XP,true}}.

% generate a check that the type we are tracking for T is compatible
% with its explicitly declared type.
predicate explicit_check(in TP:t_property,HC:h_check).
explicit_check(l_value{T,V,_},HC) :-
    explicit_type(T,_), HC = hc_explicit_type{T,V}.

predicate any_poly_caller(in PFN:string,in PI:c_instr,in PFT:t_trace,
                          in T:t_trace,ST:t_trace,RT:t_trace).
any_poly_caller(PFN,PI,PFT,T,ST,RT) :-
    trace_sub(T,ST,RT), spoly_call(PFN,PI)->wpoly_call_comp(PFT,_,_,_,ST).
any_poly_caller(PFN,PI,PFT,T,ST,RT) :-
    trace_sub(T,ST,RT), spoly_call(PFN,PI)->wpoly_call_func(PFT,_,ST).
any_poly_caller(PFN,PI,PFT,T,ST,RT) :-
    trace_sub(T,ST,RT), receive_attribute(PFN,PI,PFT,ST,_,_).

predicate min_poly_caller(in PFN:string,in PI:c_instr,in PFT:t_trace,
                          in T:t_trace,ST:t_trace,RT:t_trace).
min_poly_caller(PFN,PI,PFT,T,ST,RT) :-
    any_poly_caller(PFN,PI,PFT,T,ST,RT),
    ~(any_poly_caller(PFN,PI,PFT,T,XST,_), XST\=ST, trace_sub(XST,ST,_)).

predicate poly_check_single(in FN:string,in SUM:sum,in P:pp,in T:t_trace,
                            in RT:t_trace,in V:t_typed_val,HC:h_check).
poly_check_single(FN,SUM,P,T,RT,V,HC) :-
    trace_compose_point(p_remote{FN,SUM,P},T,RT,s_trace{NT},_),
    #bool_g(true,G), HC = hc_function{FN,SUM,l_value{NT,V,G},P}.
poly_check_single(FN,SUM,P,T,RT,V,HC) :-
    (~trace_derefs(RT,0), RT=drf{fld{empty,XF,XC}},
       HC = hc_function_future{FN,SUM,fld{T,XF,XC},V,P};
     ~trace_derefs(RT,0), RT\=drf{fld{empty,_,_}},
       HC = hc_false{no,"extra dereferences for polymorphic propagation"}).

% caller PFN/PI is a polymorphic call, where the poly_data.clp analysis has
% identified a few sources where this data could have originated from.
predicate poly_check_caller(in PFN:string,in PSUM:sum,
                            in PI:c_instr,in PP:pp,in PFT:t_trace,
                            in TP:t_property,HC:h_check,HX:h_method_extra).
poly_check_caller(PFN,PSUM,PI,PP,PFT,l_value{T,V,_},HC,HX) :-
    cil_curfn(FN), min_poly_caller(PFN,PI,PFT,T,ST,RT),
    spoly_call(PFN,PI)->wpoly_call_comp(PFT,C,FT,DT,ST),
    HX = hx_poly_comp{C,FT,DT},
    (HC = hc_full_poly_comp_receive{C,FT,DT,ST};
     spoly_receive_comp(C,FT,DT,ST)->wpoly_receive(FN,XFN,XSUM,XP,XT),
         poly_check_single(XFN,XSUM,XP,XT,RT,V,HC)).
poly_check_caller(PFN,PSUM,PI,PP,PFT,l_value{T,V,_},HC,HX) :-
    cil_curfn(FN), min_poly_caller(PFN,PI,PFT,T,ST,RT),
    spoly_call(PFN,PI)->wpoly_call_func(PFT,DT,ST),
    HX = hx_poly_func{PFN,PFT,DT},
    (HC = hc_full_poly_func_receive{PFN,PFT,DT,ST};
     spoly_receive_func(PFN,PFT,DT,ST)->wpoly_receive(FN,XFN,XSUM,XP,XT),
         poly_check_single(XFN,XSUM,XP,XT,RT,V,HC)).
poly_check_caller(PFN,PSUM,PI,PP,PFT,l_value{T,V,_},HC,HX) :-
    cil_curfn(FN), min_poly_caller(PFN,PI,PFT,T,ST,RT),
    receive_attribute(PFN,PI,PFT,ST,C,F),
    HX = hx_poly_attr{C,F,ST},
    spoly_receive_attribute(C,F,ST)->wpoly_receive(FN,XFN,XSUM,XP,XT),
        poly_check_single(XFN,XSUM,XP,XT,RT,V,HC).

% checks for all callers of this summary
predicate any_check_callers(in TP:t_property,HC:h_check,
                            MHX:maybe[h_method_extra]).
any_check_callers(TP,HC,no) :-
    cursum(FN,SUM), sum_caller(FN,SUM,PFN,PSUM,PP,PI),
    ~(PI=s_call{XI}, sum_funptr(PFN)->sindirect(XI,FN)),
    HC = hc_caller{TP,PFN,PSUM,PI,PP}.
any_check_callers(TP,HC,MHX) :-
    cursum(FN,SUM), sum_caller(FN,SUM,PFN,PSUM,PP,s_call{PI}),
    sum_funptr(PFN)->sindirect(PI,FN),
    feasible_caller(PFN,PI,FN,PFT),
    (poly_check_caller(PFN,PSUM,PI,PP,PFT,TP,HC,HX), MHX=yes{HX};
     ~poly_check_caller(PFN,PSUM,PI,PP,PFT,TP,_,_),
         HC=hc_caller{TP,PFN,PSUM,s_call{PI},PP}, MHX=no).

% trace was passed into this sum. check prior accesses to T within this sum,
% as well as uses within callers
method_check(hk_prior{TP,P,_},hm_external,HC) :-
    ~trivial_property(TP,P,_,_),
    (unsanitary_inst(TP), HC = hc_inst_sanitize{TP};
     explicit_check(TP,HC);
     any_check_callers(TP,HC,_);
     ~any_check_callers(TP,_,_), HC = hc_true{no,"No known callers"}).
method_extra(hk_prior{TP,P,_},hm_external,HX) :-
    ~trivial_property(TP,P,_,_),
    any_check_callers(TP,_,yes{HX}).

predicate global_property(in TP:t_property,in P:pp,
                          GLOB:string,NTP:t_property).
global_property(TP,P,GLOB,NTP) :-
    ~trivial_property(TP,P,_,_), property_remove_uc_sum(TP,NTP),
    property_trace_used(NTP,XT,true), trace_root(XT,glob{GLOB}).

% property is a global variable accessed by this function. check the usage is
% consistent with all other accesses to the global
method_check(hk_prior{TP,P,_},hm_glob,HC) :-
    global_property(TP,P,GLOB,NTP),
    (unsanitary_glob(NTP), HC = hc_glob_sanitize{TP};
     explicit_check(TP,HC);
     HC = hc_glob_invariant{no,GLOB,NTP}).
method_extra(hk_prior{TP,P,_},hm_glob,HX) :-
    global_property(TP,P,GLOB,_),
    HX = hx_invariant_glob{GLOB}.

predicate type_property(in TP:t_property,in P:pp,
                        COMP:string,RTP:t_property,STRS:string).
type_property(TP,P,COMP,RTP,STRS) :-
    ~trivial_property(TP,P,_,_), property_remove_uc_sum(TP,NTP),
    property_trace_used(NTP,XT,true), trace_sub(XT,STR,_),
    property_trace_used(NTP,WXT,_), trace_sub(WXT,fld{STR,_,COMP},_),
    trace_string(STR,STRS), reptrace_property(STR,empty,NTP,RTP).

% property is data reachable from a structure. check the usage is consistent
% with all other accesses to the structure's type
method_check(hk_prior{TP,P,_},hm_comp{STRS},HC) :-
    type_property(TP,P,COMP,RTP,STRS),
    (unsanitary_comp(RTP), HC = hc_comp_sanitize{RTP};
     explicit_check(TP,HC);
     HC = hc_comp_invariant{no,COMP,RTP}).
method_extra(hk_prior{TP,P,_},hm_comp{STRS},HX) :-
    type_property(TP,P,COMP,_,STRS),
    HX = hx_invariant_type{COMP}.

% ------------------------------------ hk_prior_inst

% convert a callee property to a caller property.
% we ignore the guard under which the callee->caller map holds
predicate inst_property(in I:sum,in P:pp,in CTP:t_property,TP:t_property).
inst_property(I,P,l_value{CT,V,CG},l_value{T,V,G}) :-
    inst_trace(I,P,CT,s_trace{T},TG), guard_sat(TG), inst_guard(I,P,CG,G).

% get a list of all possible caller properties
predicate inst_property_list(in I:sum,in P:pp,in CTP:t_property,
                             TPLIST:list[t_property]).
inst_property_list(I,P,CTP,TPLIST) :-
    \/inst_property(I,P,CTP,TP):list_all(TP,TPLIST).

% handle case without direct recursion: just check that all the
% possible corresponding caller properties hold
method_check(hk_prior_inst{I,P,CTP,false},hm_trivial,HC) :-
    inst_property_list(I,P,CTP,[]),
    HC = hc_true{yes{P},"No caller property"}.
method_check(hk_prior_inst{I,P,CTP,false},HM,HC) :-
    inst_property_list(I,P,CTP,[TP]),
    method_check(hk_prior{TP,P,true},HM,HC).
method_check(hk_prior_inst{I,P,CTP,false},hm_xmethod{"prior_inst"},HC) :-
    inst_property_list(I,P,CTP,TPLIST), TPLIST\=[], TPLIST\=[_],
    list_mem(TPLIST,TP), HC = hc_sub_kind{hk_prior{TP,P,true}}.

predicate inst_property_not_preserved(in I:sum,in P:pp,in CTP:t_property).
inst_property_not_preserved(I,P,CTP) :-
    inst_property(I,P,CTP,TP), CTP\=TP,
    ~(TP=l_value{T,_,_}, impure_value(T,_,_)).
inst_property_not_preserved(I,P,CTP) :-
    inst_property(I,P,CTP,TP), CTP\=TP,
    TP=l_value{T,V,G}, impure_value(T,_,_),
    follow_impure_values(P,s_trace{T},_,s_trace{NT}), CTP\=l_value{NT,V,G}.

predicate inst_property_check_preserved(in I:sum,in P:pp,in CTP:t_property,
                                        PT:t_trace).
inst_property_check_preserved(I,P,CTP,PT) :-
    inst_property(I,P,CTP,TP), CTP\=TP,
    TP=l_value{T,_,_}, impure_value(T,_,PT).

% direct recursion. the caller property must imply the callee property holds
method_check(hk_prior_inst{I,P,CTP,true},hm_local,HC) :-
    (inst_property_not_preserved(I,P,CTP),
         HC = hc_false{yes{P},"Changes over recursive call"};
     ~inst_property_not_preserved(I,P,CTP),
         HC = hc_true{yes{P},"Preserved over recursive call"}).
method_check(hk_prior_inst{I,P,CTP,true},hm_local,HC) :-
    inst_property_check_preserved(I,P,CTP,PT),
    CTP=l_value{_,V,_}, HC = hc_sub_kind{hk_future{PT,V,P,true}}.

% ------------------------------------ hk_prior_exit

% get a list of all possible exit properties
predicate exit_property_list(in TP:t_property,
                             TPLIST:list[t_pair[pp,t_property]]).
exit_property_list(TP,TPLIST) :- cursum(_,SUM),
    \/(sum_bound(SUM,_,XP), convert_point_property(XP,TP,XTP,G),
       ~unsat_point_guard(XP,G)):list_all(pair{XP,XTP},TPLIST).

method_check(hk_prior_exit{TP},hm_trivial,HC) :-
    exit_property_list(TP,[]),
    HC = hc_true{no,"No exit property"}.
method_check(hk_prior_exit{TP},HM,HC) :-
    exit_property_list(TP,[pair{XP,XTP}]),
    method_check(hk_prior{XTP,XP,true},HM,HC).
method_check(hk_prior_exit{TP},hm_xmethod{"prior_exit"},HC) :-
    exit_property_list(TP,TPLIST), TPLIST\=[], TPLIST\=[_],
    list_mem(TPLIST,pair{XP,XTP}), HC = hc_sub_kind{hk_prior{XTP,XP,true}}.

% ------------------------------------ hk_preserve

predicate convert_point_property_container(in P:pp,in TP:t_property,NTP:t_property,G:g_guard).
convert_point_property_container(P,l_value{drf{PT},V,G},l_value{drf{NPT},V,NG},WG) :-
    convert_point_trace(P,PT,s_trace{NPT},WG0),
    gmap_guard(ts_convert,P,G,NG,WG1), #and(WG0,WG1,WG).
convert_point_property_container(P,l_value{T,V,G},l_value{NT,V,NG},WG) :- T\=drf{_},
    convert_point_trace(P,T,s_trace{NT},WG0),
    gmap_guard(ts_convert,P,G,NG,WG1), #and(WG0,WG1,WG).

predicate convert_update_property(in P:pp,
                                  in TP:t_property,in TY:t_property_where,
                                  NTP:t_property,NTY:t_property_where).
convert_update_property(P,TP,TY,NTP,NTY) :-
    convert_point_property_container(P,TP,NTP,G0),
    convert_point_property_where(P,TY,NTY,G1),
    #and(G0,G1,G), ~unsat_point_guard(P,G).

method_check(hk_preserve{P,TP,TY},hm_xmethod{"any_write"},HC) :-
    convert_update_property(P,TP,TY,NTP,NTY),
    update_property(NTP,NTY,P,UK),
    HC = hc_sub_kind{hk_update{P,UK,NTP,NTY}}.
method_check(hk_preserve{P,TP,TY},hm_xmethod{"any_write"},HC) :-
    ~(convert_update_property(P,TP,TY,NTP,NTY), update_property(NTP,NTY,_,_)),
    HC = hc_true{no,"No potential updates"}.

% ------------------------------------ hk_update

% taking address of a subtrace which cannot lead to future writes
method_check(hk_update{_,UK,TP,TY},hm_write_noimpact,HC) :-
    (UK=lu_assign{XT,s_trace{DXT}}; UK=lu_argument{_,XT,s_trace{DXT}}),
    property_trace_used_where(TP,TY,DXT,TAIL), TAIL\=empty,
    reduce_trace_skolem(XT,NXT), reduce_trace_skolem(DXT,NDXT),
    cil_curfn(FN), (UK=lu_assign{_,_}, NFN=FN; UK=lu_argument{s_call{I},_,_}, anycall(I,NFN,_)),
    HC = hc_impact{UK,tl_func{FN,NDXT,cxt_any},tl_func{NFN,drf{NXT},cxt_any},TAIL}.

% property holds of the new value for XT
method_check(hk_update{P,lu_assign{XT,s_trace{DXT}},TP,_},HM,HC) :-
    property_trace_used(TP,WT,_), trace_sub(WT,drf{XT},_),
    reptrace_property(drf{XT},DXT,TP,NTP),
    method_check(hk_prior{NTP,P,true},HM,HC).

% property NULL's out a trace that is required to be non-NULL by the property
method_check(hk_update{P,lu_assign{XT,XV},TP,_},hm_write_null,HC) :-
    null_assign(TP,XT,XV),
    HC = hc_null_assign{P,TP,XT,XV}.

% property copies data from another instance of the structure
method_check(hk_update{P,lu_assign{XT,XV},TP,lw_comp{STR}},hm_write_copy,HC) :-
    XT=fld{STR,F,C}, XV=s_trace{drf{fld{XSTR,F,C}}},
    assign(P,fld{STR,F,C},_,G0), guard(P,PG), #and(PG,G0,G),
    ~(property_trace_used(TP,drf{fld{STR,XF,C}},_),
      ~(iset(XP,_,_), assign(XP,fld{STR,XF,C},s_trace{drf{fld{XSTR,XF,C}}},XG0), guard(XP,XPG),
        #and(XPG,XG0,XG), guard_equivalent(G,XG))),
    HC=hc_true{no,"Copied from other instance of the structure"}.

% there is at least one path from P to summary exit which
% does not go through a point in SET
predicate pp_exit_non_restrict(in P:pp,in SET:set[pp]).
pp_exit_non_restrict(P,SET) :-
    cursum(_,SUM), sum_bound(SUM,_,EXIT), pp_path_edge(P,EXIT,XP0,XP1),
    ~(set_member(SET,FRP), pp_xreach(FRP,XP0)),
    ~(set_member(SET,BRP), pp_xreach(XP1,BRP)).

% get a set of points XP where T is directly written, such that all points
% from P to summary exit pass through at least one point in the set
predicate dominating_set(in P:pp,in T:t_trace,SET:set[pp]).
dominating_set(P,T,SET) :-
    \/(iset(XP,_,_), assign(XP,T,_,G), #bool_g(true,G), pp_reach(P,XP)):set_all(XP,SET),
    ~set_empty(SET), ~pp_exit_non_restrict(P,SET).

%{
% handle synchronized writes of two fields by reducing to the last field write.
% TODO: handle temporarily violated invariants between the two writes
method_check(hk_update{P,lu_assign{XT0,XV0},TP,TY},hm_write_sync{TS},HC) :-
    ~null_assign(TP,XT0,XV0),
    property_trace_used_where(TP,TY,drf{XT1},_), trace_string(XT1,TS),
    dominating_set(P,XT1,SET), set_member(SET,XP), assign(XP,XT1,XV1,_),
    HC = hc_sub_kind{hk_update{XP,lu_assign{XT1,XV1},TP,TY}}.
%}

% ------------------------------------ hk_future

method_check(hk_future{PT,V,P,true},hm_local,HC) :-
    HC = hc_true{yes{P},"future updates constrained to current frame"}.

method_check(hk_future{PT,V,P,false},hm_local,HC) :-
    cursum(FN,SUM),
    (allocate_filter(FN,SUM,PT),
        HC = hc_true{yes{P},"updated location created in this frame"};
     ~allocate_filter(FN,SUM,PT),
        HC = hc_false{yes{P},"updated location not created in this frame"}).

% V must have been used previously in the function (could consider
% callers as well), indicating it is non-NULL, or explicitly set with
% a typed value.
predicate previous_value_used(in P:pp,in V:scalar).
previous_value_used(P,s_trace{T}) :-
    explicit_kind(_,XP,T,_), pp_dominate(XP,P).
previous_value_used(P,s_trace{T}) :-
    follow_impure_values(P,s_trace{T},_,s_trace{XT}),
    (explicit_kind(_,XP,XT,_), pp_dominate(XP,P);
     explicit_type(XT,_)).

% all possible values of the trace must already be initialized.
method_check(hk_future{PT,V,P,false},hm_trivial,HC) :-
    ~(val(P,PT,DV,G), guard_sat(G), ~previous_value_used(P,DV)),
    HC = hc_true{yes{P},"updated location values are already in use"}.

% all other requirements for hk_future covered by the extra information
% we attach to these proofs.
method_extra(hk_future{PT,V,P,FRAME},HM,HX) :-
    method_check(hk_future{PT,V,P,false},HM,_),
    (HX = hx_future{FRAME};
     PT=fld{_,F,C}, HX = hx_future_preserves{C,F};
     cursum(FN,SUM), HX = hx_future_assign{FN,SUM,P,PT,V}).

% baked in behavior

import "typebaked_check.clp".

% seq_operations

% for seq_operations the private data is not a heap pointer, but the return
% values of the start and next operations in the seq_operations.
% the polydata analysis does not have the expressiveness to handle this so
% we do it by hand here
poly_check_caller(_,_,_,_,drf{fld{_,FLD,"seq_operations"}},l_value{T,V,_},HC,HX) :-
    FLD\="start", T=drf{root{arg{1}}}, cil_curfn(FN),
    escape_edge(true,tl_glob{root{glob{FN}}},
                     tl_glob{drf{fld{GT,FLD,"seq_operations"}}}),
    (XFLD="start"; XFLD="next"),
    escape_edge(false,tl_glob{drf{fld{GT,XFLD,"seq_operations"}}},
                      tl_glob{root{glob{XFN}}}),
    cil_sum_body(XFN,s_func)->sum_bound(_,_,XP),
    HX = hx_poly_comp{"seq_operations",empty,empty},
    (HC = hc_true{no,XFN^" is possible start/next function for seq_operations"};
     val(p_remote{XFN,s_func,XP},root{return},s_trace{XT},_),
         #bool_g(true,G), HC = hc_function{XFN,s_func,l_value{XT,V,G},XP}).

% elevator

% for elevator requests, the private data is stored in a data structure
% separate from the elevator itself. the ->elevator_set_req_fn method is called
% on the elevator ops for each attached request, which fills in elevator_private
poly_check_caller(_,_,_,_,drf{fld{_,FLD,"elevator_ops"}},l_value{T,V,_},HC,HX) :-
    FLD\="elevator_set_req_fn", T=drf{fld{_,"elevator_private","request"}}, cil_curfn(FN),
    escape_edge(true,tl_glob{root{glob{FN}}},
                     tl_glob{drf{fld{GT,FLD,"elevator_ops"}}}),
    escape_edge(false,tl_glob{drf{fld{GT,"elevator_set_req_fn","elevator_ops"}}},
                      tl_glob{root{glob{XFN}}}),
    cil_sum_body(XFN,s_func)->sum_bound(_,_,XP),
    HX = hx_poly_comp{"request",empty,empty},
    (HC = hc_true{no,XFN^" is set_request function for elevator_ops"};
     val(p_remote{XFN,s_func,XP},fld{drf{root{arg{1}}},"elevator_private","request"},s_trace{XT},_),
         #bool_g(true,G), HC = hc_function{XFN,s_func,l_value{XT,V,G},XP}).

% fb_info

% there is a definite use of this value in the immediate parent of this loop.
cil_curfn("fb_init_class_device"),
    +previous_value_used(_,s_trace{drf{fld{_,"class_data","class_device"}}}).
