% check per-global safety

import "typecheck.clp".
import "../smemory/memoryinit.clp".

analyze session_name("cil_glob").
analyze analysis_name("casting_glob").

analyze priority(1).

% PREDICATES

type h_check ::=
     % check always holds
    hc_true{maybe[pp],TEXT:string}
     % check never holds
  | hc_false{maybe[pp],TEXT:string}
     % TP is preserved within XHKL
  | hc_preserve{TP:t_property,XHKL:h_kind_loc,P:pp}
     % writing to XT does not impact whether TP holds
  | hc_impact{UK:t_update_where,TP:t_property}
     % writing XV to XT at P nulls out the pointer XT
  | hc_null_assign{P:pp,TP:t_property,XT:t_trace,XV:scalar}
     % check that HK also holds within this same global
  | hc_sub_kind{HK:h_kind}
  .

% RULES

% fill in current location info for typecheck.clp

cil_glob(GLOB), +curloc(hl_glob{GLOB}),
    cil_var_location(_,_,LINE), +curloc_line(LINE),
    glob_location(GLOB,FILE,MINLINE,MAXLINE),
    cleanup_glob_string(GLOB,SNAME), STITLE="global "^SNAME,
    +scasting_post(hl_glob{GLOB})->
        sloc_info(SNAME,STITLE,FILE,MINLINE,MAXLINE),
    casting_display_name(hl_glob{GLOB},td_casting_overall,DNAME),
    +search_add(SNAME,DNAME).

cil_var_location(_,FILE,LINE), +point_location(p_abort,FILE,LINE).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING FAILURES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hc_true

+check_info(hc_true{_,TEXT},no,TEXT).

% ------------------------------------ hc_false

+check_fail(hc_false{_,_}).

+check_info(hc_false{_,TEXT},no,TEXT).

% ------------------------------------ hc_preserve

check_require(hc_preserve{TP,XHKL,P},XHKL,XHK) :-
    XHK = hk_preserve{P,TP,lw_glob}.

check_info(hc_preserve{TP,_,_},no,TEXT) :- property_dstring(TP,TEXT).

% ------------------------------------ hc_impact

check_fail(hc_impact{lu_assign{XT,_},l_value{T,_,G}}) :-
    trace_sub(T,XT,_),
    ~(guard_used(G,su_scalar{s_trace{GT}}), trace_sub(GT,XT,_)),
    ~(trace_used(T,su_scalar{s_trace{IT}}), trace_sub(IT,XT,_)).
check_fail(hc_impact{lu_assign{XT,_},l_value{T,_,G}}) :-
    guard_used(G,su_scalar{s_trace{GT}}), trace_sub(GT,XT,_), ~trace_sub(T,XT,_).
check_fail(hc_impact{lu_assign{XT,_},TP}) :- TP\=l_value{_,_,_}.

check_info(hc_impact{UK,TP},no,TEXT) :-
    update_where_dstring(UK,UKS), property_dstring(TP,TPS),
    TEXT = UKS^" does not impact property that "^TPS.

% ------------------------------------ hc_null_assign

predicate null_relevant(in TP:t_property,XT:t_trace).
null_relevant(TP,XT) :- property_trace_used(TP,drf{XT},true).
null_relevant(TP,XT) :- property_trace_used(TP,SXT,_),
    trace_sub(SXT,drf{XT},RT), RT\=empty.

check_fail(hc_null_assign{_,TP,XT,XV}) :-
    (XV\=s_const{_}; ~null_relevant(TP,XT)).

check_info(hc_null_assign{P,TP,XT,XV},yes{LINE},TEXT) :-
    point_location(P,_,LINE), trace_string(XT,XTS),
    TEXT = "Write nulls out pointer "^XTS.

% ------------------------------------ hc_sub_kind

check_require(hc_sub_kind{HK},HKL,HK) :- curloc(HKL).

check_info(hc_sub_kind{HK},yes{LINE},TEXT) :-
    kind_line(HK,LINE), kind_dstring(HK,HKS), TEXT = "Prove "^HKS.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CHECKING METHODS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ------------------------------------ hk_invariant

predicate relevant_property(in TP:t_property,XHKL:h_kind_loc,P:pp).
relevant_property(TP,XHKL,P) :-
    property_trace_used(TP,T,_), ~trace_derefs(T,0), trace_sub(T,XRT,WRT),
    (salias_escape(tl_glob{XRT},false)->sescape(W,_,_);
     ~trace_derefs(WRT,0), salias_escape(tl_glob{XRT},true)->sescape(W,_,_)),
    (W=te_func{FN,SUM,P}, XHKL=hl_func{FN,SUM};
     W=te_glob{G}, P=p_abort, XHKL=hl_glob{G}).

predicate relevant_update(in TP:t_property,NK:h_kind).
relevant_update(TP,NK) :-
    update_property(TP,lw_glob,p_abort,UK),
    NK = hk_update{p_abort,UK,TP,lw_glob}.

method_check(hk_invariant{TP},hm_local,HC) :-
    ~trivial_property(TP,p_abort,_,_),
    (relevant_property(TP,XHKL,P), HC = hc_preserve{TP,XHKL,P};
     relevant_update(TP,XHK), HC = hc_sub_kind{XHK};
     ~relevant_property(TP,_,_), ~relevant_update(TP,_),
         HC = hc_true{no,"Globals are never written"}).

% ------------------------------------ hk_preserve

method_check(hk_preserve{_,TP,_},hm_local,HC) :-
    (relevant_update(TP,XHK), HC = hc_sub_kind{XHK};
     ~relevant_update(TP,_), HC = hc_true{no,"Globals have no intializers"}).

% ------------------------------------ hk_update

% writing to XT does not impact whether TP holds or not
method_check(hk_update{_,UK,TP,_},hm_write_noimpact,HC) :-
    HC = hc_impact{UK,TP}.

method_check(hk_update{P,_,TP,_},hm_write_null,HC) :-
    cil_glob(GLOB), null_relevant(TP,XT), trace_root(XT,glob{GLOB}),
    salias_init(GLOB)->initial_value(XT,DXT), DXT=s_const{_},
    HC = hc_null_assign{P,TP,XT,DXT}.

% baked in behavior

import "typebaked_check.clp".
