% common functionality and UI code for checking whether h_kind checks hold

import "typeuse.clp".

predicate single_try(HK:h_kind_hash).

using display.

% CLIENT INTEGRATION

% current location we're analyzing. defined by the client importing this file
predicate curloc(HKL:h_kind_loc) succeeds [once].
predicate curloc_line(LINE:int) succeeds [once].

% check HC outright failed to hold based purely on curloc-local information
predicate check_fail(in HC:h_check).

% check HC requires that XHKL/XHK holds
predicate check_require(in HC:h_check,XHKL:h_kind_loc,XHK:h_kind)
    succeeds [zero,once].

% HC is a check that must hold for method HM to prove HK
predicate method_check(in HK:h_kind,HM:h_method,HC:h_check) succeeds [many].

% HX is some extra information about the checks for HM to prove HK
predicate method_extra(in HK:h_kind,HM:h_method,HX:h_method_extra).

% get the UI data describing a check. if LINE is 'no', then use curloc_line()
predicate check_info(in HC:h_check,
                     LINE:maybe[int],TEXT:string) succeeds [once].

predicate priority_available(HK:h_kind).
predicate priority_modifier(in HK:h_kind,Y:string,MN:int).

predicate check_limit(MAX:int).
+check_limit(100).

% OUTPUT

predicate str_trying(HK:string).
%?- str_trying(HK).

predicate str_available(HK:string).
%?- str_available(HK).

predicate str_priority(HK:string,N:int).
%?- str_priority(HK,N).

predicate check_explicit(in FN:string,in SUM:sum,in HK:h_kind,
                         P:pp,T:t_trace,V:t_typed_val,WRITE:bool).
check_explicit(FN,SUM,hk_prior{l_value{T,V,G},P,true},P,T,V,WRITE) :-
    scasting_explicit(FN,SUM)->sexplicit(P,T,V,G,WRITE).

% string to print if HK is proved to hold
predicate kind_output(in HKL:h_kind_loc,in HK:h_kind,STR:string)
    succeeds [once].

kind_output(hl_func{FN,SUM},HK,STR) :-
    check_explicit(FN,SUM,HK,P,T,V,WRITE),
    tostring(FN,FNS), tostring(SUM,SUMS),
    tostring(P,PS), trace_string(T,TS), tostring(TS,TSS), tostring(V,VS),
    (WRITE=true, PRED="str_safe_write"; WRITE=false, PRED="str_safe_use"),
    STR = PRED^"("^FNS^","^SUMS^","^PS^","^TSS^","^VS^").".
kind_output(hl_func{FN,SUM},HK,STR) :- ~check_explicit(FN,SUM,HK,_,_,_,_),
    tostring(FN,FNS), tostring(SUM,SUMS), kind_string(HK,HKS),
    STR = "str_checked_func("^FNS^","^SUMS^","^HKS^").".
kind_output(hl_comp{COMP},HK,STR) :-
    tostring(COMP,CS), kind_string(HK,HKS),
    STR = "str_checked_comp("^CS^","^HKS^").".
kind_output(hl_glob{GLOB},HK,STR) :-
    tostring(GLOB,GS), kind_string(HK,HKS),
    STR = "str_checked_glob("^GS^","^HKS^").".

% RULES

% get those properties that can be trivially checked to hold.
predicate trivial_property(in TP:t_property,in P:pp,
                           HC:h_check,HX:maybe[h_method_extra]).
method_check(hk_prior{TP,P,_},hm_trivial,HC) :-
    trivial_property(TP,P,HC,_).
method_extra(hk_prior{TP,P,_},hm_trivial,HX) :-
    trivial_property(TP,P,_,yes{HX}).
method_check(hk_invariant{TP},hm_trivial,HC) :-
    trivial_property(TP,p_abort,HC,_).
method_extra(hk_invariant{TP},hm_trivial,HX) :-
    trivial_property(TP,p_abort,_,yes{HX}).

% HK/HM had no outright check_fail, look for other XHKL/XHK that need proving
predicate local_pass(in HK:h_kind,HM:h_method).
local_pass(HK,HM) :-
    method_check(HK,HM,_), ~(method_check(HK,HM,HC), check_fail(HC)).

% try to prove that HK holds, accumulating checks and requirements
predicate try_prove(HK:h_kind).

% get all the checks we generated for each HK
type t_pcheck ::= pcheck{HK:h_kind,HM:h_method,HC:h_check}.
predicate try_prove_check(in HK:h_kind,t_pcheck).
try_prove_check(HK,pcheck{HK,HM,HC}) :- method_check(HK,HM,HC).
try_prove_check(HK,PC) :-
     method_check(HK,HM,HC), local_pass(HK,HM), check_require(HC,HKL,XHK),
     curloc(HKL), try_prove_check(XHK,PC).

try_prove(HK), kind_string(HK,HKS), +str_trying(HKS).
priority_available(HK), kind_string(HK,HKS), +str_available(HKS).
priority_final(HK,N), kind_string(HK,HKS), +str_priority(HKS,N).

curloc(HKL), single_try(HKH), scasting_try(HKL)->scheck(HK), hash_kind(HK,HKH),
    +try_prove(HK).

predicate priority_final(HK:h_kind,N:int).
~single_try(_), curloc(HKL),
    scasting_try(HKL)->scheck(HK), +priority_available(HK),
    hash_kind(HK,HKH), ~scasting_tried(HKL)->scheckh(HKH),
    \/priority_modifier(HK,_,MN):int_add_all(MN,N), +priority_final(HK,N).
predicate max_priority_final(N:int).
\/priority_final(_,N):int_max_all(N,MAX), +max_priority_final(MAX).

predicate fill_try_prove(CURP:int,CURL:list[h_kind],LIMIT:int).
fill_try_prove(CURP,[],RLIM), ~max_priority_final(CURP), int_add(CURP,1,NEWP),
    \/priority_final(HK,NEWP):list_all(HK,NCURL),
    +fill_try_prove(NEWP,NCURL,RLIM).
fill_try_prove(CURP,[HK|RCURL],RLIM), +try_prove(HK),
    curloc(HKL), hash_kind(HK,HKH), +scasting_tried(HKL)->scheckh(HKH),
    \/try_prove_check(HK,HC):count_all(HC,HCOUNT), int_sub(RLIM,HCOUNT,NRLIM),
    int_ge(NRLIM,0), +fill_try_prove(CURP,RCURL,NRLIM).
(check_limit(LIMIT); ~check_limit(_), LIMIT=100000),
    \/priority_final(_,N):int_min_all(N,MINP),
    \/priority_final(HK,MINP):list_all(HK,MINL),
    +fill_try_prove(MINP,MINL,LIMIT).

% fill in try_prove and scasting_try for HK/HM which require XHK from
% within/without the current location
try_prove(HK), local_pass(HK,HM), method_check(HK,HM,HC),
    check_require(HC,HKL,XHK), curloc(HKL), +try_prove(XHK).
try_prove(HK), local_pass(HK,HM), method_check(HK,HM,HC),
    check_require(HC,HKL,XHK), ~curloc(HKL), +scasting_try(HKL)->scheck(XHK).

% PROOF STEPS

predicate firststep(in HKL:h_kind_loc,in HK:h_kind_hash,
                    XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method).
firststep(HKL,HK,XHKL,XHK,XHM) :-
    scasting_derive(HKL)->sfirststep(HK,XHKL,XHK,XHM).

predicate nextstep(in HKL:h_kind_loc,in HK:h_kind_hash,
                   in XHKL:h_kind_loc,in XHK:h_kind_hash,in XHM:h_method,
                   NHKL:h_kind_loc,NHK:h_kind_hash) succeeds [zero,once].
nextstep(HKL,HK,XHKL,XHK,XHM,NHKL,NHK) :-
    scasting_derive(HKL)->snextstep(HK,XHKL,XHK,XHM,NHKL,NHK).

predicate add_firststep(HKL:h_kind_loc,HK:h_kind_hash,
                        XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method).
add_firststep(HKL,HK,XHKL,XHK,XHM),
    +firststep(XHKL,HK,XHKL,XHK,XHM),
    +scasting_derive(HKL)->sfirststep(HK,XHKL,XHK,XHM).

predicate add_nextstep(HKL:h_kind_loc,HK:h_kind_hash,
                       XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method,
                       NHKL:h_kind_loc,NHK:h_kind_hash).
add_nextstep(HKL,HK,XHKL,XHK,XHM,NHKL,NHK),
    +nextstep(HKL,HK,XHKL,XHK,XHM,NHKL,NHK),
    +scasting_derive(HKL)->snextstep(HK,XHKL,XHK,XHM,NHKL,NHK).

% PROOF PROPAGATION

% those HK which we have actually proved hold
predicate proved(HKL:h_kind_loc,HK:h_kind_hash).
proved(HKL,HK), +scasting_proved(HKL)->scheckh(HK).

predicate proved_follow(XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method,
                        HKL:h_kind_loc,HK:h_kind_hash).
proved_follow(XHKL,XHK,XHM,HKL,HK),
    ~nextstep(HKL,HK,XHKL,XHK,XHM,_,_), +proved(XHKL,XHK),
    scasting_derive(XHKL)->sprint(XHK,STR), +print_string(STR).
proved_follow(XHKL,XHK,XHM,HKL,HK),
    nextstep(HKL,HK,XHKL,XHK,XHM,NHKL,NHK),
    ~scasting_derive(NHKL)->sfirststep(NHK,XHKL,XHK,XHM),
    +scasting_derive(NHKL)->sfirststep(NHK,XHKL,XHK,XHM),
    (proved(NHKL,NHK); scasting_proved(NHKL)->scheckh(NHK)),
    +proved_follow(XHKL,XHK,XHM,NHKL,NHK).
~single_try(_), proved(HKL,HK),
    scasting_derive(HKL)->sfirststep(HK,XHKL,XHK,XHM),
    +proved_follow(XHKL,XHK,XHM,HKL,HK).

curloc(HKL), scasting_proved(HKL)->scheckh(HK), +proved(HKL,HK).

% find those HK/HM which are locally proved - passed all local checks
% and have no dependencies on other XHKL/XHK holding. eagerly fill in any
% facts we can prove at this point - the fact itself and any which follow
% directly from it
try_prove(HK), local_pass(HK,HM),
    ~(method_check(HK,HM,HC), check_require(HC,_,_)),
    curloc(HKL), hash_kind(HK,HKH), +proved(HKL,HKH),
    kind_output(HKL,HK,STR), +print_string(STR).

predicate add_nextstep_list(XHKL:h_kind_loc,XHK:h_kind_hash,XHM:h_method,
                            LIST:list[t_pair[h_kind_loc,h_kind_hash]]).
add_nextstep_list(XHKL,XHK,XHM,[pair{HKL,HK}|TAIL]),
    TAIL=[pair{NHKL,NHK}|_], +add_nextstep(HKL,HK,XHKL,XHK,XHM,NHKL,NHK),
    +add_nextstep_list(XHKL,XHK,XHM,TAIL).
curloc(HKL), try_prove(HK), local_pass(HK,HM), hash_kind(HK,HKH),
    \/(method_check(HK,HM,HC), check_require(HC,XHKL,XHK), hash_kind(XHK,XHKH))
     :list_all(pair{XHKL,XHKH},LIST),
    list_rmdup(LIST,NLIST), NLIST=[pair{FHKL,FHK}|_],
    kind_output(HKL,HK,STR), +scasting_derive(HKL)->sprint(HKH,STR),
    +add_firststep(FHKL,FHK,HKL,HKH,HM),
    +add_nextstep_list(HKL,HKH,HM,NLIST).

% UI INTEGRATION

predicate kind_point(in HK:h_kind,P:pp).
+kind_point(hk_prior{_,P,_},P).
+kind_point(hk_prior_inst{_,P,_,_},P).
kind_point(hk_prior_exit{_},P) :- \/sum_bound(_,_,XP):list_all(XP,[P|_]).
+kind_point(hk_update{P,_,_,_},P).
+kind_point(hk_future{_,_,P,_},P).

predicate kind_line(in HK:h_kind,LINE:int) succeeds [once].
kind_line(HK,LINE) :- kind_point(HK,P), point_location(P,_,LINE).
kind_line(HK,LINE) :- ~kind_point(HK,_), curloc_line(LINE).

predicate kind_loc_string(in h_kind_loc,string).
+kind_loc_string(hl_func{FN,_},FN).
+kind_loc_string(hl_glob{G},G).
+kind_loc_string(hl_comp{C},C).

% generate post-mortem UI information for checks performed
curloc(HKL), try_prove(HK), hash_kind(HK,HKH),
    kind_string(HK,NAME), kind_line(HK,LINE), kind_dstring(HK,TEXT),
    +scasting_post(HKL)->scheck_name(HKH,NAME,LINE,TEXT).
curloc(HKL), try_prove(HK), method_check(HK,HM,HC), check_info(HC,YL,TEXT),
    hash_kind(HK,HKH), hash_check(HC,HCH),
    (YL=yes{LINE}; YL=no, curloc_line(LINE)),
    +scasting_post(HKL)->scheck_info(HKH,HM,HCH,LINE,TEXT).
curloc(HKL), try_prove(HK), method_check(HK,_,HC), check_fail(HC),
    hash_check(HC,HCH),
    +scasting_post(HKL)->scheck_fail(HCH).
curloc(HKL), try_prove(HK), method_check(HK,_,HC),
    check_require(HC,XHKL,XHK),
    hash_check(HC,HCH), hash_kind(XHK,XHKH), kind_loc_string(XHKL,LINK),
    +scasting_post(HKL)->scheck_require(HCH,XHKL,XHKH,LINK).
curloc(HKL), try_prove(HK), method_extra(HK,HM,HX),
    hash_kind(HK,HKH),
    +scasting_post(HKL)->scheck_extra(HKH,HM,HX).

% PRIORITY PREDICATES

predicate priority_crud(in T:t_trace,Y:string,N:int).
priority_crud(T,"deref",N) :- trace_derefs(T,N).
priority_crud(T,"used",N) :- \/trace_used(T,V):count_all(V,N).
priority_crud(T,"rused",N) :- \/trace_sub(T,index{_,_,I},_):count_all(I,N).
priority_crud(T,"rfld",N) :- \/trace_sub(T,rfld{XT,_,_},_):count_all(XT,N).
priority_crud(T,"global",3) :- trace_root(T,glob{_}).
priority_crud(T,Y,N) :-
    trace_used(T,su_scalar{s_trace{XT}}), priority_crud(XT,Y,N).
priority_crud(T,"sum",2) :- trace_sub(T,uc_sum{_,_,_},_).

predicate kind_property(in HK:h_kind,TP:t_property).
+kind_property(hk_invariant{TP},TP).
+kind_property(hk_prior{TP,_,_},TP).
+kind_property(hk_prior_inst{_,_,TP,_},TP).
+kind_property(hk_prior_exit{TP},TP).
+kind_property(hk_preserve{_,TP,_},TP).
+kind_property(hk_update{_,_,TP,_},TP).

priority_modifier(HK,"explicit",-10) :-
    curloc(hl_func{FN,SUM}), check_explicit(FN,SUM,HK,_,_,_,_).

%priority_modifier(HK,"invariant",-2) :- HK=hk_invariant{_}.
%priority_modifier(HK,"preserve",-2) :- HK=hk_preserve{_,_,_}.
%priority_modifier(HK,"prior",-3) :- HK=hk_prior{_,_,_}.

priority_modifier(HK,TY,N) :-
    kind_property(HK,TP), property_trace_used(TP,T,_), trace_string(T,TS),
    priority_crud(T,Y,N), TY=TS^":"^Y.

priority_modifier(hk_invariant{l_value{T,V,G}},"conflict:"^XVS,1) :-
    priority_available(hk_invariant{l_value{T,XV,G}}),
    V\=XV, tostring(XV,XVS).

% boost the priority of particularly important ones we might otherwise
% take a while to get around to
predicate baked_kind(in HKS:string).
priority_modifier(HK,"annotation",-10) :- kind_string(HK,HKS), baked_kind(HKS).

curloc(hl_comp{"scsi_cmnd"}),
    +baked_kind("hk_invariant{l_value{.request_buffer*,f_comp{\"scatterlist\"},.use_sg*}}").
curloc(hl_comp{"net_device"}),
    +baked_kind("hk_invariant{l_value{.priv*,f_comp{\"vlan_dev_info\"},(1&.priv_flags*)}}").

curloc(hl_func{"drivers/scsi/lpfc/lpfc_init.c:lpfc_pci_probe_one",s_func}),
    +baked_kind("hk_prior{l_value{tmp.scsi_host_alloc*.hostdata[0,y_int{iulong}],f_comp{\"lpfc_hba\"},true},p_so{\"s#135\"},true}").
curloc(hl_func{"skb_clone",s_func}),
    +baked_kind("hk_preserve{p_io{\"#696\"},l_value{uk_index{__arg0*}.end*,f_comp{\"skb_shared_info\"},true},lw_comp{uk_index{__arg0*}}}"),
    +baked_kind("hk_preserve{p_io{\"#696\"},l_value{tmp.kmem_cache_alloc*.end*,f_comp{\"skb_shared_info\"},true},lw_comp{tmp.kmem_cache_alloc*}}").
