% theory of how values may be interpreted as various possibly conflicting types

import "typehdr.clp".
import "../smemory/memory.clp".
import "../smemory/scalar_sat.clp".
import "../smemory/aliasing_full.clp".
import "../daliasing/aliasing_sum.clp".
import "../daliasing/escape.clp".

using set.

% TYPES

% types a trace can possess which we are interested in. these are just
% the names of composite structures
type t_typed_val = string.

% particular properties we're interested in checking
type t_property ::=
     % condition G is sufficient to ensure that T has type consistent with V
    l_value{T:t_trace,V:t_typed_val,G:g_guard}.

% conditions a property might be holding with respect to for a function
type t_property_where ::=
    lw_glob              % instantiated from a global invariant
  | lw_comp{T:t_trace}.  % instantiated from a type invariant on T

type t_update_where ::=
    lu_assign{T:t_trace,V:scalar}
  | lu_argument{I:sum,AT:t_trace,V:scalar}.

% PROOF PRIMITIVES

% func/glob/comp we can infer h_kind checks on
type h_kind_loc ::=
    hl_func{FN:string,SUM:sum}
  | hl_glob{GLOB:string}
  | hl_comp{COMP:string}.

% individual checks we can prove hold for an h_kind_loc
type h_kind ::=
     % hl_glob, hl_comp: TP is invariant over all writes
    hk_invariant{TP:t_property}
     % hl_func: property TP was established by write(s) within this
     % function, its callers or prior callees. TP is relative to
     % the state at function entry
  | hk_prior{TP:t_property,P:pp,GSPLIT:bool}
     % hl_func: property CTP within the isum I/P was established
     % prior to the isum. RECURSE indicates if this is direct recursion
  | hk_prior_inst{I:sum,P:pp,CTP:t_property,RECURSE:bool}
     % hl_func: property TP within this function was established
     % prior to this function exiting. TP is relative to the exit point
  | hk_prior_exit{TP:t_property}
     % hl_glob, hl_func: invariant property TP is preserved within this
     % func/glob. TP is relative to point P
  | hk_preserve{P:pp,TP:t_property,TY:t_property_where}
     % hl_glob, hl_func: possible update HK at P preserves TP wrt TY
  | hk_update{P:pp,UK:t_update_where,TP:t_property,TY:t_property_where}
     % hl_func: all values for PT at P or in the future (up to or past
     % the end of the current frame depending on FRAME) are of type V.
  | hk_future{PT:t_trace,V:t_typed_val,P:pp,FRAME:bool}
  .

% methods for proving an h_kind check holds. each method is separate and if
% any one succeeds then the h_kind is proved
type h_method ::=
    hm_trivial
  | hm_local
  | hm_external
  | hm_callee{CFN:string}
  | hm_glob
  | hm_comp{STR:string}
  | hm_guard{G:string}
  | hm_xmethod{TEXT:string}
  | hm_write_copy
  | hm_write_noimpact
  | hm_write_null
  | hm_write_sync{T:string}
  | hm_preserve.

% extra information that can be associated with each 
type h_method_extra ::=
 %%%%% correctness information
     % this method assumes that assignments to field F of type C
     % are type preserving
    hx_future_preserves{C:string,F:string}
     % this method asserts that values of PT at P are of type V
  | hx_future_assign{FN:string,SUM:sum,P:pp,PT:t_trace,V:t_typed_val}
 %%%%% statistics information
     % this method is dependent on a trace's explicit type
  | hx_explicit_type
     % this method is dependent on an explicit cast of a trace
  | hx_explicit_use
     % this method is dependent on a fixed type for a trace
  | hx_explicit_fixed
     % this method is dependent on a global invariant on G
  | hx_invariant_glob{G:string}
     % this method is dependent on a type invariant on C
  | hx_invariant_type{C:string}
     % this method is dependent on a polymorphic data relationship
  | hx_poly_comp{C:string,RFT:t_trace,T:t_trace}
  | hx_poly_func{FN:string,FT:t_trace,T:t_trace}
  | hx_poly_attr{C:string,F:string,ST:t_trace}
     % this method is dependent on proving properties of future writes
  | hx_future{FRAME:bool}.

% SESSIONS

% explicit accesses to traces performed within each summary
session scasting_explicit(FN:string,SUM:sum)
    containing [sexplicit,sexplicit_fixed].

% there is an explicit use of T with type V if G holds.
% if WRITE is true this is a use inserted because of a field write
predicate sexplicit(P:pp,T:t_trace,V:t_typed_val,G:g_guard,WRITE:bool).

% T is a just-allocated trace which has been assigned explicit type V
predicate sexplicit_fixed(T:t_trace,V:t_typed_val).

% explicit accesses to traces which have been proved safe during previous runs
session scasting_explicit_safe(FN:string,SUM:sum)
    containing [sexplicit].

% explicit accesses to a particular flat field of a structure.
% this could be an array of ints within the structure which is interpreted
% as some other type, or a hidden untyped field at the end of the structure
session scasting_explicit_field(C:string,F:string)
    containing [sexplicit_field].

% this field might be accessed as type V within FN/SUM
predicate sexplicit_field(V:t_typed_val,FN:string,SUM:sum).

% subset of salias_allocate functions which do not generate typed data.
% wfunc_untyped indicates the actual assigned trace in FN we assumed untyped
session scasting_allocate_untyped(FN:string)
    containing [wfunc_allocate,wfunc_untyped].
predicate wfunc_untyped(T:t_trace).

% for each type, there is one wpointer_field indicating (a) whether there are
% any pointer fields transitively in that type, and (b) if there are such
% pointer fields, whether any of them are dereferenced
session scasting_type_pointer(C:string) containing [wpointer_field].
predicate wpointer_field(PTR:bool,DRF:bool).

% keep track of structures PC which C is known to be a subtype of.
% the subtypes we recognize are those where an explicit downcast occurs,
% e.g. pstr *x; str *y = (str*)x;   % str <: pstr
session scasting_subtype(C:string) containing [wsubtype].
predicate wsubtype(PC:string).

% all the impure fields which were assumed to be type-preserving during
% casting proofs, and the casts which were proved safe using this assumption
session scasting_restrict_field(C:string,F:string) containing [wrestrict_field].
predicate wrestrict_field(HKL:h_kind_loc,HKS:string).

% all the locations whose targets were not assumed to receive other values,
% and the casts which were proved safe using this assumption
session scasting_restrict_assign(FN:string,SUM:sum) containing [wrestrict_assign].
predicate wrestrict_assign(P:pp,PT:t_trace,V:t_typed_val,
                           HKL:h_kind_loc,HKS:string).

% all the traces which were assumed during polymorphic propagation to
% have a value of a type V at offset RT from T
session scasting_restrict_poly(FN:string,SUM:sum) containing [wrestrict_poly].
predicate wrestrict_poly(P:pp,T:t_trace,RT:t_trace,V:t_typed_val,
                         HKL:h_kind_loc,HKS:string).

% TODO: incorporate wfunc_untyped, extra correctness info into checking

% GENERAL PREDICATES

% type Y is a subtype of type PY. values of type Y may be used anywhere a value
% of type PY is expected. If Y <: PY and PY <: Y then Y and PY are equivalent.
predicate check_subtype(in Y:t_type,in PY:t_type).

% get explicit typing information from the form of T
predicate explicit_type(in T:t_trace,V:t_typed_val).

% V is a supertype of the explicit XV type of T, using known XV <: V
predicate known_explicit_super_type(in T:t_trace,in V:t_typed_val).

% V and XV are compatible types, using either known V <: XV or XV <: V
predicate known_compatible_types(in V:t_typed_val,in XV:t_typed_val).

% PROPERTY PREDICATES

% remap TP according to a ts_map
predicate map_property(in ts_map,in TP:t_property,RTP:t_property)
    succeeds [once].

% get a trace mentioned anywhere in a property
predicate property_trace_used(in TP:t_property,XT:t_trace,MUST:bool).

% CUSTOMIZATION

% casts from values of type TYPEDEF to Y are ok
predicate check_cast_typedef(in TYPEDEF:string,in Y:t_type).

% STRINGIFYING

predicate typed_val_dstring(in t_typed_val,string) succeeds [once].
predicate typed_when_dstring(in t_typed_val,in g_guard,string) succeeds [once].

predicate property_string(in t_property,string) succeeds [once].
predicate property_dstring(in t_property,string) succeeds [once].

predicate property_where_string(in t_property_where,string) succeeds [once].
predicate property_where_dstring(in t_property_where,string) succeeds [once].

predicate update_where_string(in t_update_where,string) succeeds [once].
predicate update_where_dstring(in t_update_where,string) succeeds [once].

predicate kind_string(in h_kind,string) succeeds [once].
predicate kind_dstring(in h_kind,string) succeeds [once].

% RULES

import "typebaked.clp".

% IMPURE VALUES

predicate impure_value(in T:t_trace,P:pp,PT:t_trace).
impure_value(T,P,PT) :-
    cursum(FN,SUM), ~match_call_exit(FN,SUM,T,_,_,_,_,_),
    T=uc_sum{I,P,PT}, PT=fld{_,F,C},
    ~(isum_callee(FN,SUM,I,CFN),
      salias_write_comp(C,F)->wcomp_init(CFN,_)).

predicate follow_impure_values(in P:pp,in V:scalar,NP:pp,NV:scalar)
    succeeds [many].
follow_impure_values(P,V,P,V) :- V\=s_trace{_}.
follow_impure_values(P,s_trace{T},P,s_trace{T}) :- ~impure_value(T,_,_).
follow_impure_values(_,s_trace{T},NP,NV) :- impure_value(T,XP,PT),
    val(XP,PT,V,G), guard_sat(G), follow_impure_values(XP,V,NP,NV).

% CAST SAFETY

% every type is a subtype of itself
+check_subtype(Y,Y).

% every type is a subtype of void
+check_subtype(_,y_void).

% subtypes can be followed across pointers only if the two are equivalent
check_subtype(y_ptr{Y},y_ptr{PY}) :-
    check_subtype(Y,PY), check_subtype(PY,Y).

% an array of the subtypes can be used in place of the supertype
check_subtype(y_array{Y,_},PY) :- check_subtype(Y,PY).

% the subtype can be used in place of the supertype only if the subtype
% and supertype have the same width.
check_subtype(Y,y_array{PY,_}) :- Y\=y_array{_,_},
    type_width(Y,N), type_width(PY,N), check_subtype(Y,PY).

% a structure is a subtype of whatever its first field is
% (if its first field is a structure, also consider the first field of the
% first field, and so forth)
check_subtype(y_comp{C},Y) :- first_field(C,Y,_,_).

% a structure is a subtype of a structure containing a single field
% if it is a subtype of the type of that field
check_subtype(Y,y_comp{C}) :-
    comp_field(C,0,_,_,Y), ~comp_field(C,1,_,_,_), first_field(C,PY,true,_),
    check_subtype(Y,PY).

% allow subtyping between integer types with the same size
check_subtype(y_int{K0},y_int{K1}) :-
    ikind_bytes(K0,_,BYTES), ikind_bytes(K1,_,BYTES).

% field N of C should be considered padding when comparing to subtype XC
predicate padding_field(in C:string,in N:int,in XC:string).

% padding field which is an array of bytes extending sizeof(C) to sizeof(XC)
padding_field(C,N,XC) :-
    comp_field(C,N,_,_,Y), int_add(N,1,NN), ~comp_field(C,NN,_,_,_),
    Y=y_array{YE,_}, type_width(YE,1),
    type_width(y_comp{C},W), type_width(y_comp{XC},W).

% padding field which is a flexible array of 0 or 1
padding_field(C,N,_) :-
    comp_field(C,N,_,_,Y), int_add(N,1,NN), ~comp_field(C,NN,_,_,_),
    Y=y_array{y_int{_},LEN}, (LEN=0; LEN=1).

% take care of per-field structural subtyping.
% y_comp{C} is a subtype of y_comp{PC} if, for all non-padding fields in PC,
% there is a field in C at the same offset and with the same width which
% has an equivalent type
type t_comp_fields ::= y_comp_fields{F:string,BS:int,Y:t_type}.
predicate fields_subtype(in C:string,in FIELDS:list[t_comp_fields]).
check_subtype(y_comp{C},y_comp{PC}) :- C\=PC,
    \/(comp_field(PC,N,BS,F,Y), ~padding_field(PC,N,C))
     :list_all(y_comp_fields{F,BS,Y},FIELDS),
    FIELDS\=[], fields_subtype(C,FIELDS).

% any type is the subtype of an empty set of fields
+fields_subtype(_,[]).

% for each field make sure there is one of the same offset/width in C
% with an equivalent type
fields_subtype(C,[y_comp_fields{_,BS,Y}|TAIL]) :-
    comp_field(C,_,BS,_,XY), type_width(Y,W), type_width(XY,W),
    check_subtype(XY,Y), check_subtype(Y,XY),
    fields_subtype(C,TAIL).

% extensions for looking at special typedefs used on fields
%fields_subtype(C,[y_comp_fields{F,BS,Y}|TAIL]) :-
%    comp_field(C,_,BS,_,XY), type_width(Y,W), type_width(XY,W),
%    cil_comp(C)->cil_field_name(_,F,TYP,_,_),
%    cil_comp(C)->cil_type_named(TYP,TYPEDEF,_),
%    check_cast_typedef(TYPEDEF,XY),
%    fields_subtype(TAIL).

% EXPLICIT TYPED VALUES

% handle traces T which have an explicit language type.
% don't do this for indexes into pointer buffers.
explicit_type(T,C) :-
    cil_curfn(FN), trace_type(FN,T,y_comp{C}), T\=index{drf{_},_,_},
    ~(T=uc_sum{s_call{I},P,RT}, callret(P,I,RT,_)).
explicit_type(T,C) :-
    T=uc_sum{s_call{I},P,RT}, callret(P,I,RT,_), anycall(I,CFN,_),
    trace_type(CFN,drf{root{return}},y_comp{C}).

+known_compatible_types(V,V).
known_compatible_types(V,XV) :-
    V\=XV, scasting_subtype(V)->wsubtype(XV).
known_compatible_types(V,XV) :-
    V\=XV, scasting_subtype(XV)->wsubtype(V).

known_explicit_super_type(T,V) :-
    explicit_type(T,V).
known_explicit_super_type(T,V) :-
    explicit_type(T,XV), scasting_subtype(XV)->wsubtype(V).

% PROPERTY PREDICATES

map_property(X,l_value{OB,V,OG},l_value{NB,V,NG}) :-
    map_scalar(X,s_trace{OB},s_trace{NB}), map_guard(X,OG,NG).

+property_trace_used(l_value{T,_,_},T,true).
property_trace_used(l_value{_,_,G},T,false) :-
    guard_used(G,su_scalar{s_trace{T}}).
property_trace_used(TP,T,false) :-
    property_trace_used(TP,MT,_), trace_used(MT,su_scalar{s_trace{T}}).

predicate reptrace_property(in OT:t_trace,in NT:t_trace,
                            in TP0:t_property,TP1:t_property).
reptrace_property(OT,NT,TP0,TP1) :-
    map_property(ts_reptrace{OT,s_trace{NT}},TP0,TP1).

% convert a property relative to point P to a property relative to entry
predicate convert_point_property(in P:pp,in TP:t_property,
                                 NTP:t_property,G:g_guard).
convert_point_property(P,l_value{T,V,G},l_value{NT,V,NG},WG) :-
    convert_point_trace(P,T,s_trace{NT},WG0),
    gmap_guard(ts_convert,P,G,NG,WG1), #and(WG0,WG1,WG).

predicate convert_point_property_where(in P:pp,in TY:t_property_where,
                                       NTY:t_property_where,G:g_guard).
#bool_g(true,G), +convert_point_property_where(_,lw_glob,lw_glob,G).
convert_point_property_where(P,lw_comp{STR},lw_comp{NSTR},G) :-
    convert_point_trace(P,STR,s_trace{NSTR},G).

predicate property_remove_uc_sum(in TP:t_property,NTP:t_property).
property_remove_uc_sum(l_value{T,V,G},l_value{NT,V,NG}) :-
    trace_remove_uc_sum(T,NT), map_guard(ts_remove_uc_sum,G,NG).

% PROPERTY UPDATING

predicate property_trace_used_where(in TP:t_property,in TY:t_property_where,
                                    T:t_trace,RT:t_trace).
property_trace_used_where(TP,TY,T,WT) :- TY\=lw_comp{_},
    property_trace_used(TP,XT,_), trace_sub(XT,T,WT).
property_trace_used_where(TP,TY,T,WT) :- TY=lw_comp{STR},
    reptrace_property(STR,empty,TP,RTP),
    property_trace_used(RTP,RXT,_), trace_sub(RXT,RT,WT), RT\=empty,
    map_scalar(ts_reptrace{empty,s_trace{STR}},s_trace{RT},s_trace{T}).

predicate maybe_impact(in UK:t_update_where,
                       in TP:t_property,in TY:t_property_where).
maybe_impact(lu_assign{XT,_},TP,TY) :-
    property_trace_used_where(TP,TY,XT,_).
maybe_impact(lu_assign{_,s_trace{XT}},TP,TY) :-
    property_trace_used_where(TP,TY,XT,RT), RT\=empty.
maybe_impact(lu_argument{_,_,s_trace{XT}},TP,TY) :-
    property_trace_used_where(TP,TY,XT,RT), RT\=empty.

predicate base_update_trace(ST:t_trace,RTNE:bool,P:pp,UK:t_update_where).
(iset(P,_,_); P=p_abort), assign(P,ST,V,_),
    +base_update_trace(drf{ST},false,P,lu_assign{ST,V}),
    V=s_trace{DST}, +base_update_trace(DST,true,P,lu_assign{ST,V}).
icall(P,_,I), argument(P,I,AT,V,_),
    V=s_trace{DST}, +base_update_trace(DST,true,P,lu_argument{s_call{I},AT,V}).

predicate update_property(in TP:t_property,in TY:t_property_where,
                          P:pp,UK:t_update_where).
update_property(TP,TY,P,UK) :-
    property_trace_used_where(TP,TY,XT,RT),
    (RT=empty, base_update_trace(XT,false,P,UK);
     RT\=empty, base_update_trace(XT,_,P,UK)),
    maybe_impact(UK,TP,TY).

% STRINGIFYING

typed_val_dstring(C,STR) :- STR = "struct "^C.

typed_when_dstring(C,W,STR) :-
    #bool_g(true,W), typed_val_dstring(C,STR).
typed_when_dstring(C,W,STR) :-
    ~#bool_g(true,W), typed_val_dstring(C,CS), guard_string(W,BWS),
    str_len(BWS,LEN),
    (int_lt(LEN,200), WS=BWS; int_ge(LEN,200), WS="[...]"),
    STR = CS ^ " when " ^ WS.

property_string(l_value{T,C,G},STR) :-
    trace_string(T,TS), guard_string(G,GS),
    STR = "l_value{"^TS^","^C^","^GS^"}".

property_dstring(l_value{T,C,G},STR) :-
    trace_string(T,TS), typed_when_dstring(C,G,VS),
    STR = TS^" has type "^VS.

+property_where_string(lw_glob,"lw_glob").
property_where_string(lw_comp{T},STR) :-
    trace_string(T,TS), STR = "lw_comp{"^TS^"}".

+property_where_dstring(lw_glob,"global invariant").
property_where_dstring(lw_comp{T},STR) :-
    trace_string(T,TS), STR = "type invariant over "^TS.

update_where_string(lu_assign{T,V},STR) :-
    trace_string(T,TS), scalar_string(V,VS),
    STR = "lu_assign{"^TS^","^VS^"}".
update_where_string(lu_argument{I,AT,V},STR) :-
    tostring(I,IS), trace_string(AT,AS), scalar_string(V,VS),
    STR = "lu_argument{"^IS^","^AS^","^VS^"}".

update_where_dstring(lu_assign{T,V},STR) :-
    trace_string(T,TS), scalar_string(V,VS),
    STR = "write to "^TS^" with "^VS.
update_where_dstring(lu_argument{I,AT,V},STR) :-
    dstring_sum(I,IS), scalar_string(V,VS),
    STR = "passing argument "^VS^" to "^IS.

kind_string(hk_invariant{TP},STR) :-
    property_string(TP,TPS),
    STR = "hk_invariant{"^TPS^"}".
kind_string(hk_prior{TP,P,GSPLIT},STR) :-
    property_string(TP,TPS), tostring(P,PS), tostring(GSPLIT,GSPLITS),
    STR = "hk_prior{"^TPS^","^PS^","^GSPLITS^"}".
kind_string(hk_prior_inst{I,P,CTP,RECURSE},STR) :-
    tostring(I,IS), tostring(P,PS), property_string(CTP,CTPS),
    tostring(RECURSE,RS),
    STR = "hk_prior_inst{"^IS^","^PS^","^CTPS^","^RS^"}".
kind_string(hk_prior_exit{TP},STR) :-
    property_string(TP,TPS),
    STR = "hk_prior_exit{"^TPS^"}".
kind_string(hk_preserve{P,TP,TY},STR) :-
    tostring(P,PS), property_string(TP,TPS), property_where_string(TY,TYS),
    STR = "hk_preserve{"^PS^","^TPS^","^TYS^"}".
kind_string(hk_update{P,UK,TP,TY},STR) :-
    tostring(P,PS), update_where_string(UK,UKS),
    property_string(TP,TPS), property_where_string(TY,TYS),
    STR = "hk_update{"^PS^","^UKS^","^TPS^","^TYS^"}".
kind_string(hk_future{PT,V,P,FRAME},STR) :-
    trace_string(PT,PTS), tostring(V,VS), tostring(P,PS), tostring(FRAME,FS),
    STR = "hk_future{"^PTS^","^VS^","^PS^","^FS^"}".

kind_dstring(hk_invariant{TP},STR) :-
    property_dstring(TP,STR).
kind_dstring(hk_prior{TP,_,_},STR) :-
    property_dstring(TP,TPS),
    STR = TPS^" must hold".
kind_dstring(hk_prior_inst{I,_,CTP,_},STR) :-
    dstring_sum(I,IS), property_dstring(CTP,CTPS),
    STR = "within "^IS^", "^CTPS^" must hold".
kind_dstring(hk_prior_exit{TP},STR) :-
    property_dstring(TP,TPS),
    STR = TPS^" must hold at exit".
kind_dstring(hk_preserve{_,TP,TY},STR) :-
    property_dstring(TP,TPS), property_where_dstring(TY,TYS),
    STR = "all writes and escapes preserve invariant that "^TPS^" is "^TYS.
kind_dstring(hk_update{P,UK,TP,TY},STR) :-
    update_where_dstring(UK,UKS),
    property_dstring(TP,TPS), property_where_dstring(TY,TYS),
    STR = UKS^" preserves invariant that "^TPS^" is "^TYS.
kind_dstring(hk_future{PT,V,P,FRAME},STR) :-
    trace_string(PT,PTS), tostring(V,VS),
    STR = "current and future values of "^PTS^" have type "^VS.
