%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Path-sensitive intraprocedural points-to and value-aliasing information.

import "traces.clp".
import "paths.clp".
import "scalar.clp".

% PREDICATES

% note that all per-pp guards, i.e. val, eval, lval, inst_trace, etc.
% are dependent on P actually being reachable, guard(P,G). to get the conds
% under which both P is reachable *AND* a property val(P,S,T,_) holds, conjoin
% guard(P,GG) and val(P,S,T,G).

% at program point P, the memory location with canonical name S contains value 
% T when G holds (if S is a pointer, then S points to T). note that S must be 
% the canonical name for the location -- this predicate does NOT work for
% arbitrary access paths. for converting access paths to the canonical name,
% use the convert_point_trace predicate
predicate val(in P:pp,in S:t_trace,T:scalar,G:g_guard).

% at program point P, expression E points to T if G holds.
predicate eval(in P:pp,in E:c_exp,T:scalar,G:g_guard).

% at program point P, lval LV refers to trace T if G holds
predicate lval(in P:pp,in LV:c_lval,T:t_trace,G:g_guard).

% static initializer I refers to trace T unconditionally. P should be p_abort
predicate ival(in P:pp,in I:c_init,T:t_trace).

% as interpreted by the call/loop I at point P, trace CT maps to T if G holds
predicate inst_trace(in I:sum,in P:pp,in CT:t_trace,T:scalar,G:g_guard).

% interpret callee bits, scalars, guards at a call/loop I at point P
predicate inst_scalar_bit(in I:sum,in P:pp,in CB:scalar_bit,
                          BG:g_guard,G:g_guard).
predicate inst_scalar(in I:sum,in P:pp,in CI:scalar,I:scalar,G:g_guard).
predicate inst_guard(in I:sum,in P:pp,in CG:g_guard,G:g_guard) succeeds [once].

% convert a trace T which represents an access path relative to point P
% to an equivalent canonical value for use with the val() and related
% predicates, i.e. each drf{ST} within T is converted to
% whatever val(P,ST,_,_) is.
predicate convert_point_trace(in P:pp,in T:t_trace,NV:scalar,G:g_guard).
predicate convert_point_scalar_bit(in P:pp,in B:scalar_bit,
                                   NG:g_guard,G:g_guard).
predicate convert_point_scalar(in P:pp,in V:scalar,NV:scalar,G:g_guard).
predicate convert_point_guard(in P:pp,in G:g_guard,NG:g_guard).

% refine the condition G to EG, according to the guard under which transfer
% occurs from P0 to P1. where guard(P0,G0) and guard(P1,G1), if EG&G1 holds,
% then G&G0 holds and the edge P0->P1 will be taken during function execution
predicate eguard(in P0:pp,in P1:pp,in G:g_guard,EG:g_guard) succeeds [once].

type access_type ::= read | write.

% T is directly accessed at program point P as a read or write (not deepwrite).
% accesses through callees are not included here.
predicate access(P:pp,T:t_trace,access_type).

% assignment at P from V to T under guard G. this may be at an iset or isum.
% in the latter case it overrides the default call behavior for the trace
predicate assign(in P:pp,T:t_trace,V:scalar,G:g_guard).

% arguments to function calls and assembly statements
predicate argument(in P:pp,I:c_instr,AT:t_trace,V:scalar,G:g_guard).
predicate argument_asm_in(in P:pp,I:c_instr,AT:t_trace,V:scalar,G:g_guard).
predicate argument_asm_out(in P:pp,I:c_instr,AT:t_trace,T:t_trace,G:g_guard).

% the return value of instruction I is stored at T if G holds
predicate callret(in P:pp,I:c_instr,T:t_trace,G:g_guard).

% for an indirect call I, function trace T is executed if G holds
predicate callfnptr(in P:pp,I:c_instr,FT:t_trace,G:g_guard).

% ALIASING PREDICATES

% all information about aliasing between input traces and side effects
% of callees is given by the following predicates. these are not defined
% within this file but should be provided by other aliasing files,
% which vary in levels of soundness and precision

% T0 and T1 as passed into FN/SUM might alias if G holds
predicate trace_alias(in FN:string,in SUM:sum,
                      in T0:t_trace,in T1:t_trace,G:g_guard)
    succeeds [zero,once].

% a trace always aliases itself. other rules must be provided by clients
#bool_g(true,G), +trace_alias(_,_,T,T,G).

% the value of T might be clobbered by I/P. P may be remote, in which case
% I refers to an isum in that remote function
predicate trace_clobber(in I:sum,in P:pp,in T:t_trace).

% REMOTE MEMORY

% predicates guard(), val() and inst_trace() (and inst_scalar() etc.) can be
% queried using remote program points and isums - these will perform the
% memory analysis within the specified remote function, and will return
% traces, scalars, and guards with respect to that remote function.
% in order to enable this, cil_sum_body.db must first be generated,
% and memoryremote.clp must be run with the
% appropriate clients for aliasing/clobbering included
% (e.g. aliasing_mid_remote.clp or aliasing_full_remote.clp)

% storage for all predicates which we need to compute remote memory queries
session smemory_remote(FN:string,SUM:sum) containing [
  guard,econd,callret,callfnptr,
  assign,argument,argument_asm_in,argument_asm_out
  ].

% CUSTOMIZATION

% val() will not be computed anywhere - disables memory analysis
predicate disable_val().

% assign() will not be computed through the usual rules
predicate disable_assign().

% prevent lval(P,LV,_,_) from being computed in the standard way,
% allowing for customization of lval(P,LV,_,_)
predicate omit_lval(in P:pp,in LV:c_lval).

% prevent eval(P,E,_,_) from being computed in the standard way,
% allowing for customization of eval(P,E,_,_)
predicate omit_eval(in P:pp,in E:c_exp).

% prevent val(P,T,_,_) from being computed in the standard way,
% allowing for customization of val(P,T,_,_)
predicate omit_val(in P:pp,in T:t_trace).

% modify eval() to ignore casts between or in/out of integer types.
% these are needed to model any sign changes etc. that could occur but also
% mask the flow of pointers that are masquerading as integers
% NOTE: ON BY DEFAULT
%predicate ignore_int_casts().

% RULES

% CIL location offsets

% CIL offset OFF of root trace RT as calculated at P is T if G holds
predicate trace_offset(in P:pp,in RT:t_trace,in OFF:c_offset,
                       T:t_trace,G:g_guard).

% empty offsets preserve the original location
#bool_g(true,G), cil_off_none(OFF), +trace_offset(_,RT,OFF,RT,G).

% field offsets take the specified field of the original location
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,F,_,_,_),
    base_off_type(OFF,y_comp{C}),
    ?trace_offset(P,RT,OFF,_,_), trace_simplify(fld{RT,F,C},FT),
    trace_offset(P,FT,NOFF,T,G), +trace_offset(P,RT,OFF,T,G).

% index offsets take the specified index on the original location
cil_off_index(OFF,E,NOFF), base_off_type(NOFF,EY),
    ?trace_offset(P,RT,OFF,_,_), eval(P,E,I,IG),
    trace_simplify(index{RT,EY,I},IT),
    trace_offset(P,IT,NOFF,T,TG), #and(IG,TG,G), ~#bool_g(false,G),
    +trace_offset(P,RT,OFF,T,G).

% TODO: while manipulating indexes in the following rules we assume they
% are int32 quantities. this needs to be fixed for integer overflow checking

% reverse RT to T such that trace_offset(P,T,OFF,RT,G).
predicate trace_rev_offset(in P:pp,in RT:t_trace,in OFF:c_offset,
                           T:t_trace,G:g_guard).
#bool_g(true,G), cil_off_none(OFF), +trace_rev_offset(_,RT,OFF,RT,G).
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,F,_,_,_),
    base_off_type(OFF,y_comp{C}),
    ?trace_rev_offset(P,RT,OFF,_,_), trace_rev_offset(P,RT,NOFF,MT,G),
    trace_simplify(rfld{MT,F,C},T), +trace_rev_offset(P,RT,OFF,T,G).
cil_off_index(OFF,E,NOFF), base_off_type(NOFF,EY),
    ?trace_rev_offset(P,RT,OFF,_,_), eval(P,E,I,IG),
    simplify_scalar(s_binop{b_mult,I,s_const{-1},true,32},NI),
    trace_rev_offset(P,RT,NOFF,MT,TG), #and(IG,TG,G), ~#bool_g(false,G),
    trace_simplify(index{MT,EY,NI},T), +trace_rev_offset(P,RT,OFF,T,G).

% get a fixed number of bytes represented by an offset.
% TODO: assuming indexes are int32
predicate offset_bytes(in P:pp,in OFF:c_offset,V:scalar,G:g_guard).
cil_off_none(OFF), #bool_g(true,G), +offset_bytes(_,OFF,s_const{0},G).
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,_,_,BBS,_), int_div(BBS,8,BS),
    ?offset_bytes(P,OFF,_,_), offset_bytes(P,NOFF,NV,G),
    simplify_scalar(s_binop{b_plusa,NV,s_const{BS},true,32},V),
    +offset_bytes(P,OFF,V,G).
cil_off_index(OFF,E,NOFF), base_off_type(NOFF,EY),
    ?offset_bytes(P,OFF,_,_), offset_bytes(P,NOFF,NV,G0),
    eval(P,E,IV,G1), type_width(EY,BS),
    simplify_scalar(s_binop{b_mult,IV,s_const{BS},true,32},MV),
    simplify_scalar(s_binop{b_plusa,NV,MV,true,32},V),
    #and(G0,G1,G), +offset_bytes(P,OFF,V,G).

% pointer casts

% ptr->ptr casting T to type Y yields CT. not defined for all T
predicate trace_pointer_cast(in T:t_trace,in Y:t_type,CT:t_trace)
    succeeds [zero,once].

predicate fld_chain(in T:t_trace,in FLIST:list[t_pair[string,string]],
                    NT:t_trace).
+fld_chain(T,[],T).
fld_chain(T,[pair{F,C}|FLIST],NT) :-
    trace_simplify(fld{T,F,C},MT), fld_chain(MT,FLIST,NT).

predicate rfld_chain(in T:t_trace,in FLIST:list[t_pair[string,string]],
                     NT:t_trace).
+rfld_chain(T,[],T).
rfld_chain(T,[pair{F,C}|FLIST],NT) :-
    rfld_chain(T,FLIST,MT), trace_simplify(rfld{MT,F,C},NT).

% look for field upcasts, where a pointer to some comp C is cast to the type of
% one of its first fields, or downcasts, where a pointer to some type Y is
% cast to a type which contains Y as its first field
trace_pointer_cast(T,Y,NT) :-
    Y\=y_int{_}, cil_curfn(FN), trace_type(FN,T,y_comp{C}),
    first_field(C,Y,true,FLIST), fld_chain(T,FLIST,NT).
trace_pointer_cast(T,y_comp{C},NT) :-
    cil_curfn(FN), trace_type(FN,T,Y), Y\=y_int{_},
    first_field(C,Y,true,FLIST), rfld_chain(T,FLIST,NT).

% ptr->ptr casting V to type Y yields CV. defined for all V
predicate scalar_pointer_cast(in V:scalar,in Y:t_type,CV:scalar).

scalar_pointer_cast(s_trace{T},Y,s_trace{CT}) :- trace_pointer_cast(T,Y,CT).
scalar_pointer_cast(s_trace{T},Y,s_trace{T}) :- ~trace_pointer_cast(T,Y,_).
scalar_pointer_cast(V,_,V) :- V\=s_trace{_}.

% lvalue evaluation

% value to use for unconstrained expression
predicate eval_unconstrained(in P:pp,in E:c_exp,T:t_trace).
eval_unconstrained(P,E,uc_exp{P,E,Y}) :- exp_type(E,Y).

predicate eval_trace(in P:pp,in E:c_exp,T:t_trace,G:g_guard).
eval_trace(P,E,T,G) :- eval(P,E,s_trace{T},G).
eval_trace(P,E,T,G) :- eval(P,E,V,G), V\=s_trace{_}, eval_unconstrained(P,E,T).

% compute the trace for variable lvals. this is an offset from the var_trace
% for the variable, which may itself be a recursive cell if the head element
% is a stack- or statically-allocated structure
cil_lval_var(LV,X,OFF), var_trace(X,R), ?lval(P,LV,_,_), ~omit_lval(P,LV),
    trace_simplify(root{R},RT), trace_offset(P,RT,OFF,T,G), +lval(P,LV,T,G).

% compute the trace for memory lvals. evaluate the pointer for its targets,
% then take the offset of each target
cil_lval_mem(LV,ME,OFF), ?lval(P,LV,_,_), ~omit_lval(P,LV),
    eval_trace(P,ME,RT,MG), trace_offset(P,RT,OFF,T,OG),
    #and(MG,OG,G), ~#bool_g(false,G), +lval(P,LV,T,G).

% static initializer evaluation

cil_init(G), cil_var_init(_,INIT), +ival(_,INIT,root{glob{G}}).
cil_init_cmpnd_field(INIT,FX,SINIT), ?ival(P,SINIT,_), ival(P,INIT,T),
    cil_field_name(FX,F,_,_,_), cil_init_cmpnd_type(INIT,TYP),
    cil_type_comp(TYP,C), trace_simplify(fld{T,F,C},FT), +ival(P,SINIT,FT).
cil_init_cmpnd_index(INIT,E,SINIT), ?ival(P,SINIT,_), ival(P,INIT,T),
    trace_type("",T,y_array{Y,N}),
    ~trace_type("",T,y_array{y_array{Y,N},_}),  % multi-dimensional case
    eval(P,E,I,_), trace_simplify(index{T,Y,I},IT), +ival(P,SINIT,IT).

% expression evaluation

% integer constant expressions
(cil_exp_const(E,_);
 cil_exp_x_sizeof_alignof(E);
 cil_exp_binop(E,BOP,_,_,_), arith_op(BOP);
 cil_exp_unop(E,UOP,_,_), UOP\=u_lnot), cil_exp_x_intval(E,N),
    ?eval(P,E,_,_), ~omit_eval(P,E), #bool_g(true,G), +eval(P,E,s_const{N},G).

% get the (constant or non-constant) offset of an &(0->field)c
predicate offset_address(E:c_exp).
cil_exp_x_intval(MCE,0), cil_exp_cast(ME,MCE,_), cil_lval_mem(LV,ME,OFF),
    cil_exp_addr(E,LV), +offset_address(E),
    ?eval(P,E,_,_), ~omit_eval(P,E),
    offset_bytes(P,OFF,V,G), +eval(P,E,V,G).

% float constant expressions
cil_exp_const(E,C), cil_const_real(C,_,_),
    ?eval(P,E,_,_), ~omit_eval(P,E), eval_unconstrained(P,E,T),
    #bool_g(true,G), +eval(P,E,s_trace{T},G).

% string constant expressions
cil_const_str(C,STR), cil_exp_const(E,C), ?eval(P,E,_,_), ~omit_eval(P,E),
    #bool_g(true,G), +eval(P,E,s_trace{root{cstr{STR}}},G).

% evaluate lval expressions by getting val() for the referenced trace S.
% note that all variables are evaluated here, not just pointers.
% an 'abstract target' will then be generated for integer variables which
% are evaluated, i.e. int trace T will initially point to drf{T}
cil_exp_lval(E,LV), ?eval(P,E,_,_), ~omit_eval(P,E),
    lval(P,LV,S,SG), val(P,S,T,TG),
    #and(SG,TG,G), ~#bool_g(false,G), +eval(P,E,T,G).

% address-of lvalue expressions
cil_exp_addr(E,LV), ~offset_address(E), ?eval(P,E,_,_), ~omit_eval(P,E),
    lval(P,LV,T,G), +eval(P,E,s_trace{T},G).

% evaluate unary arithmetic operations (-,~)
cil_exp_unop(E,OP,RE,_), OP\=u_lnot,
    exp_type(E,EY), type_bits(EY,S,BITS),
    ~cil_exp_x_intval(E,_), ?eval(P,E,_,_), ~omit_eval(P,E), eval(P,RE,RV,G),
    simplify_scalar(s_unop{OP,RV,S,BITS},V), +eval(P,E,V,G).

% evaluate binary arithmetic operations (+,-,*,etc.)
arith_op(OP), cil_exp_binop(E,OP,LE,RE,_),
    exp_type(E,EY), type_bits(EY,S,BITS),
    ~cil_exp_x_intval(E,_), ?eval(P,E,_,_), ~omit_eval(P,E),
    eval(P,LE,LV,LG), eval(P,RE,RV,RG),
    #and(LG,RG,G), ~#bool_g(false,G),
    simplify_scalar(s_binop{OP,LV,RV,S,BITS},V), +eval(P,E,V,G).

% evaluate pointer arithmetic operations (+,- on a pointer)
pointer_op(OP), cil_exp_binop(E,OP,LE,RE,_), ?eval(P,E,_,_), ~omit_eval(P,E),
    exp_type(E,y_ptr{DY}), eval_trace(P,LE,T,LG), eval(P,RE,RI,RG),
    (OP=b_pluspi, NRI=RI;
     OP=b_minuspi,
         simplify_scalar(s_binop{b_minusa,s_const{0},RI,true,32},NRI)),
    #and(LG,RG,G), ~#bool_g(false,G),
    trace_simplify(index{T,DY,NRI},NT), +eval(P,E,s_trace{NT},G).

% get an unconstrained value for pointer difference
pointer_int_op(OP), cil_exp_binop(E,OP,_,_,_),
    ?eval(P,E,_,_), ~omit_eval(P,E), eval_unconstrained(P,E,T),
    #bool_g(true,G), +eval(P,E,s_trace{T},G).

% add the cast for pointer to pointer cast expressions
cil_exp_cast(E,CE,TYP), exp_tptr(E), exp_tptr(CE),
    convert_type(TYP,y_ptr{DY}),
    ?eval(P,E,_,_), ~omit_eval(P,E), eval(P,CE,V,G),
    scalar_pointer_cast(V,DY,CV), +eval(P,E,CV,G).

% ignore * -> int and int -> ptr casts
cil_exp_cast(E,CE,_), (~exp_tptr(E); ~exp_tptr(CE)),
    ?eval(P,E,_,_), ~omit_eval(P,E), eval(P,CE,V,G), +eval(P,E,V,G).

% evaluate boolean expressions by lifting with s_guard
exp_boolean(E), ?eval(P,E,_,_), ~omit_eval(P,E),
    beval(P,E,VG,G), +eval(P,E,s_guard{VG},G).

% boolean expression evaluation

% evaluate comparisons directly. exp_boolean is defined for these
compare_op(OP), cil_exp_binop(E,OP,LE,RE,_),
    exp_type(LE,LY), type_bits(LY,S,BITS),  % RE should be same S/BITS
    ?beval(P,E,_,_), ~omit_beval(P,E),
    eval(P,LE,LV,LG), eval(P,RE,RV,RG),
    #and(LG,RG,G), ~#bool_g(false,G),
    #id_g(sc_cmp{OP,LV,RV,S,BITS},VG), +beval(P,E,VG,G).

% evaluate non-boolean expressions by lifting to guards
?beval(P,E,_,_), ~omit_beval(P,E), ~exp_boolean(E), eval(P,E,V,G),
    exp_type(E,Y), type_bits(Y,S,BITS),
    scalar_guard(V,S,BITS,VG), +beval(P,E,VG,G).

% instruction trace evaluation

% map arguments out of a call or asm
inst_trace(s_call{I},P,drf{AT},V,G) :- argument(P,I,AT,V,G).
inst_trace(s_asm{I},P,drf{AT},V,G) :- argument_asm_in(P,I,AT,V,G).
inst_trace(s_asm{I},P,drf{AT},s_trace{T},G) :- argument_asm_out(P,I,AT,T,G).

% map the return value out of a call
inst_trace(s_call{I},P,root{return},s_trace{T},G) :- callret(P,I,T,G).

% map arguments and locals out of a loop. just preserve the trace
inst_trace(s_loop{_},P,root{R},s_trace{root{R}},G) :-
    (R=arg{_}; R=local{_}), #bool_g(true,G).

% map globals and constant strings out of a call/loop. just preserve the trace
inst_trace(I,P,root{R},s_trace{root{R}},GG) :-
    (R=glob{_}; R=cstr{_}), #bool_g(true,GG).

% map drf{T} out of a call/loop. since drf{T} means the target at entry to the
% call/loop, evaluate T for its location and take its target at P.
inst_trace(I,P,drf{CT},V,G) :-
    inst_trace(I,P,CT,s_trace{T},PG), val(P,T,V,TG),
    #and(PG,TG,G), ~#bool_g(false,G).

% map fields out of a call/loop, just map out the parent trace
inst_trace(I,P,fld{CT,F,C},s_trace{T},G) :-
    inst_trace(I,P,CT,s_trace{FT},G), trace_simplify(fld{FT,F,C},T).
inst_trace(I,P,rfld{CT,F,C},s_trace{T},G) :-
    inst_trace(I,P,CT,s_trace{FT},G), trace_simplify(rfld{FT,F,C},T).

% map out array indexes using inst_scalar
inst_trace(I,P,index{CT,Y,CI},s_trace{NT},G) :-
    inst_trace(I,P,CT,s_trace{T},TG), inst_scalar(I,P,CI,XI,IG),
    #and(TG,IG,G), ~#bool_g(false,G), trace_simplify(index{T,Y,XI},NT).

% fill in try_gmap_trace/try_gmap_bit from scalar.clp for inst_trace
try_gmap_trace(ts_inst{I},P,T,V,G) :- inst_trace(I,P,T,V,G).
try_gmap_bit(ts_inst{I},P,B,NG,G) :- inst_scalar_bit(I,P,B,NG,G).

% compute inst_scalar/inst_guard using gmap_scalar/gmap_guard
inst_scalar(I,P,CS,S,G) :- gmap_scalar(ts_inst{I},P,CS,S,G).
inst_guard(I,P,CG,G) :-
    \/(gmap_guard(ts_inst{I},P,CG,XG0,XG1), #and(XG0,XG1,XG)):#and_all(XG,G).

% trace conversion

convert_point_trace(_,root{R},s_trace{root{R}},G) :- #bool_g(true,G).
convert_point_trace(P,drf{T},V,G) :-
    convert_point_trace(P,T,s_trace{NT},G0), val(P,NT,V,G1), #and(G0,G1,G).
convert_point_trace(P,fld{T,F,C},s_trace{NFT},G) :-
    convert_point_trace(P,T,s_trace{NT},G), trace_simplify(fld{NT,F,C},NFT).
convert_point_trace(P,rfld{T,F,C},s_trace{NFT},G) :-
    convert_point_trace(P,T,s_trace{NT},G), trace_simplify(rfld{NT,F,C},NFT).
convert_point_trace(P,index{T,Y,I},s_trace{NIT},G) :-
    convert_point_trace(P,T,s_trace{NT},G0), convert_point_scalar(P,I,NI,G1),
    #and(G0,G1,G), trace_simplify(index{NT,Y,NI},NIT).

% fill in try_gmap_trace/try_gmap_bit from scalar.clp for convert_point_trace
try_gmap_trace(ts_convert,P,T,V,G) :- convert_point_trace(P,T,V,G).
try_gmap_bit(ts_convert,P,B,NG,G) :- convert_point_scalar_bit(P,B,NG,G).

% compute convert_point_scalar/compute_point_guard using gmap_scalar/gmap_guard
convert_point_scalar(P,S,NS,G) :- gmap_scalar(ts_convert,P,S,NS,G).
convert_point_guard(P,G,NG) :-
    \/(gmap_guard(ts_convert,P,G,XG0,XG1), #and(XG0,XG1,XG)):#and_all(XG,NG).

predicate trace_compose_point(in P:pp,in T:t_trace,in RT:t_trace,
                              V:scalar,G:g_guard).
trace_compose_point(_,T,empty,s_trace{T},G) :- #bool_g(true,G).
trace_compose_point(P,T0,drf{PT1},V,G) :-
    trace_compose_point(P,T0,PT1,s_trace{PTR},G0),
    val(P,PTR,V,G1), #and(G0,G1,G).
trace_compose_point(P,T0,fld{FT1,F,C},s_trace{TR},G) :-
    trace_compose_point(P,T0,FT1,s_trace{FTR},G),
    trace_simplify(fld{FTR,F,C},TR).
trace_compose_point(P,T0,rfld{FT1,F,C},s_trace{TR},G) :-
    trace_compose_point(P,T0,FT1,s_trace{FTR},G),
    trace_simplify(rfld{FTR,F,C},TR).
trace_compose_point(P,T0,index{IT1,Y,I},s_trace{TR},G) :-
    trace_compose_point(P,T0,IT1,s_trace{ITR},G0),
    convert_point_scalar(P,I,NI,G1),
    trace_simplify(index{ITR,Y,NI},TR), #and(G0,G1,G).

predicate uc_trace_sub(in T:t_trace,TS:t_trace,TR:t_trace) succeeds [many].
+uc_trace_sub(T,T,empty).
uc_trace_sub(drf{T},TS,drf{TR}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(fld{T,F,C},TS,fld{TR,F,C}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(rfld{T,F,C},TS,rfld{TR,F,C}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(index{T,Y,I},TS,index{TR,Y,I}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(uk_index{T,Y},TS,uk_index{TR,Y}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(uk_recurse{T,RL},TS,uk_recurse{TR,RL}) :- uc_trace_sub(T,TS,TR).
uc_trace_sub(uc_sum{_,_,T},TS,drf{TR}) :- uc_trace_sub(T,TS,TR).

% edge guards

% compute guard for remote points
guard(p_remote{FN,SUM,P},G) :- smemory_remote(FN,SUM)->guard(P,G).

% compute econd for remote points
predicate cfg_econd(in P0:pp,in P1:pp,ECG:g_guard) order [P0,P1].
cfg_econd(P0,P1,ECG) :- econd(P0,P1,ECG).
cfg_econd(p_remote{FN,SUM,P0},p_remote{FN,SUM,P1},ECG) :-
    smemory_remote(FN,SUM)->econd(P0,P1,ECG).

% P has more than one incoming edge. P may be remote
predicate cfg_join(in P:pp).
join(P), +cfg_join(P).
cfg_join(p_remote{FN,SUM,P}) :- cil_sum_body(FN,SUM)->join(P).

% if P1 is a non-join then the guard at P1 implies the guard at P0 and
% the state is left unchanged along the edge
eguard(P0,P1,G,G) :- ~cfg_join(P1).

% if P1 is a join point then generate the edge guards from the edge cond ECG.
% the distinguishing guard EG is (G1=>G0&ECG)
eguard(P0,P1,G,EG) :- cfg_join(P1),
    guard(P0,G0), guard(P1,G1), cfg_econd(P0,P1,ECG),
    #not(G1,NG1), #and(G0,ECG,GAND), #or(NG1,GAND,GIMP), #and(G,GIMP,EG).

% compute assignment and argument values

assign(p_remote{FN,SUM,P},T,V,G) :-
    smemory_remote(FN,SUM)->assign(P,T,V,G).
argument(p_remote{FN,SUM,P},I,AT,V,G) :-
    smemory_remote(FN,SUM)->argument(P,I,AT,V,G).
argument_asm_in(p_remote{FN,SUM,P},I,AT,V,G) :-
    smemory_remote(FN,SUM)->argument_asm_in(P,I,AT,V,G).
argument_asm_out(p_remote{FN,SUM,P},I,AT,T,G) :-
    smemory_remote(FN,SUM)->argument_asm_out(P,I,AT,T,G).
callret(p_remote{FN,SUM,P},I,T,G) :-
    smemory_remote(FN,SUM)->callret(P,I,T,G).
callfnptr(p_remote{FN,SUM,P},I,T,G) :-
    smemory_remote(FN,SUM)->callfnptr(P,I,T,G).

% find the direct reads/writes to traces
cil_exp_lval(E,LV), exp_point(E,P), lval(P,LV,T,_), +access(P,T,read).
iset(P,_,_), assign(P,T,_,_), +access(P,T,write).

predicate fill_assign(P:pp,T:t_trace,V:scalar,Y:t_type,G:g_guard).
fill_assign(P,T,V,Y,G), Y\=y_comp{_}, +assign(P,T,V,G).
fill_assign(P,T,s_trace{drf{VT}},y_comp{C},G), comp_field(C,_,_,F,Y),
    +fill_assign(P,fld{T,F,C},s_trace{drf{fld{VT,F,C}}},Y,G).
fill_assign(P,T,s_trace{uc_sum{_,_,VT}},y_comp{C},G), comp_field(C,_,_,F,Y),
    +fill_assign(P,fld{T,F,C},s_trace{drf{fld{VT,F,C}}},Y,G).
fill_assign(P,T,s_const{N},y_comp{C},G), comp_field(C,_,_,F,Y),
    +fill_assign(P,fld{T,F,C},s_const{N},Y,G).
~disable_assign(), iset(P,_,I),
    cil_instr_set(I,LV,E), lval(P,LV,T,LG), eval(P,E,V,RG),
    #and(LG,RG,G), ~#bool_g(false,G), exp_type(E,Y),
    +fill_assign(P,T,V,Y,G).
~disable_assign(),
    cil_init_single(I,E), ival(p_abort,I,LT), eval(p_abort,E,RV,G),
    exp_type(E,Y), +fill_assign(p_abort,LT,RV,Y,G).

predicate fill_argument(P:pp,I:c_instr,AT:t_trace,V:scalar,Y:t_type,G:g_guard).
fill_argument(P,I,AT,V,Y,G), Y\=y_comp{_}, +argument(P,I,AT,V,G).
fill_argument(P,I,AT,s_trace{drf{T}},y_comp{C},G), comp_field(C,_,_,F,Y),
    +fill_argument(P,I,fld{AT,F,C},s_trace{drf{fld{T,F,C}}},Y,G).
fill_argument(P,I,AT,s_trace{uc_sum{_,_,T}},y_comp{C},G),
    +fill_argument(P,I,AT,s_trace{drf{T}},y_comp{C},G).
~disable_assign(), icall(P,_,I), cil_instr_call_arg(I,A,E), eval(P,E,V,G),
    exp_type(E,Y), +fill_argument(P,I,root{arg{A}},V,Y,G).

predicate fill_argument_asm_in(P:pp,I:c_instr,AT:t_trace,
                               V:scalar,Y:t_type,G:g_guard).
fill_argument_asm_in(P,I,AT,V,Y,G),
    Y\=y_comp{_}, +argument_asm_in(P,I,AT,V,G).
fill_argument_asm_in(P,I,AT,s_trace{drf{T}},y_comp{C},G),
    comp_field(C,_,_,F,Y),
    +fill_argument_asm_in(P,I,fld{AT,F,C},s_trace{drf{fld{T,F,C}}},Y,G).
fill_argument_asm_in(P,I,AT,s_trace{uc_sum{_,_,T}},y_comp{C},G),
    +fill_argument_asm_in(P,I,AT,s_trace{drf{T}},y_comp{C},G).
~disable_assign(), iasm(P,_,I), cil_instr_asm_in(I,A,_,_,E), eval(P,E,V,G),
    exp_type(E,Y), +fill_argument_asm_in(P,I,root{asm_in{A}},V,Y,G).

predicate fill_argument_asm_out(P:pp,I:c_instr,AT:t_trace,
                                T:t_trace,Y:t_type,G:g_guard).
fill_argument_asm_out(P,I,AT,T,Y,G),
    Y\=y_comp{_}, +argument_asm_out(P,I,AT,T,G).
fill_argument_asm_out(P,I,AT,T,y_comp{C},G), comp_field(C,_,_,F,Y),
    +fill_argument_asm_out(P,I,fld{AT,F,C},fld{T,F,C},Y,G).
~disable_assign(), iasm(P,_,I), cil_instr_asm_out(I,A,_,_,LV), lval(P,LV,T,G),
    lval_type(LV,Y), +fill_argument_asm_out(P,I,root{asm_out{A}},T,Y,G).

% evaluate and store call site return lvals
~disable_assign(), icall(P,_,I), cil_instr_call_ret(I,LV), lval(P,LV,T,G),
    +callret(P,I,T,G).

% evaluate and store indirect call site targets
~disable_assign(), icall(P,_,I), ~dircall(I,_), cil_instr_call(I,E),
    (eval(P,E,s_trace{drf{T}},G); eval(P,E,s_trace{uc_sum{_,_,T}},G)),
    +callfnptr(P,I,T,G).

% val transfer function

% adding will merge G into the condition under which S points to T at P
predicate vmerge(in P:pp,in T:t_trace,V:scalar,G:g_guard).

% there are an excessive number of incoming vmerge edges for T at P.
% use this to throttle val() and prevent exponential blowup when analyzing
% code which does complex transformations on integers (which we generally can't
% model precisely in the first place due to non-linear arithmetic etc.)
predicate excess_vmerge(in P:pp,in T:t_trace).
excess_vmerge(P,T) :-
    \/(vmerge(P,T,V,_), V\=s_trace{_}):count_all(V,COUNT), int_ge(COUNT,5).

% compute val() by doing a lossless merge of all incoming vmerge edges,
% unless there are an excessive number
~disable_val(), ?val(P,T,_,_), ~omit_val(P,T), ~excess_vmerge(P,T),
    vmerge(P,T,V,_), \/vmerge(P,T,V,G):#or_all(G,MG), #simplify(MG,NMG),
    ~#bool_g(false,NMG), +val(P,T,V,NMG).

% compute an unconstrained value for val() if there are too many vmerge edges
val(P,T,V,G) :- excess_vmerge(P,T),
    break_point(P,_,_,XP), V=s_trace{uc_val{XP,T}}, #bool_g(true,G).

% P is a CFG entry point. P may be remote
predicate cfg_entry(in P:pp).
cfg_entry(P) :- sum_bound(_,P,_).
cfg_entry(p_remote{FN,SUM,P}) :- cil_sum_body(FN,SUM)->sum_bound(_,P,_).

% resolve vmerge at summary entry points, each T -> drf{T}
vmerge(P,T,s_trace{DT},G) :- cfg_entry(P),
    trace_simplify(drf{T},DT), #bool_g(true,G).

% P0 is a predecessor of P1 in the CFG. P1/P0 may be remote
predicate cfg_predecessor(in P1:pp,P0:pp) order [P0,P1].
cfge_final(P0,P1,_), +cfg_predecessor(P1,P0).
cfg_predecessor(p_remote{FN,SUM,P1},p_remote{FN,SUM,P0}) :-
    cil_sum_body(FN,SUM)->cfge_final(P0,P1,_).

% T is written with the value V at P if G holds. covers aliasing between traces
% P may be remote
predicate cfg_assign_write(in P:pp,in T:t_trace,V:scalar,G:g_guard).
cfg_assign_write(P,T,V,G) :- assign(P,XT,V,XG),
    break_point(P,FN,SUM,_), trace_alias(FN,SUM,T,XT,AG), #and(AG,XG,G).

% if T is assigned V along a CFG edge, merge V with its targets after the set
vmerge(P1,T,V,EG) :- cfg_predecessor(P1,P0),
    cfg_assign_write(P0,T,V,G), eguard(P0,P1,G,EG).

% I is a summary edge starting at P. P/I may be remote
predicate cfg_isum(in P:pp,I:sum).
isum(P,_,I), +cfg_isum(P,I).
cfg_isum(p_remote{FN,SUM,P},I) :-
    cil_sum_body(FN,SUM)->isum(P,_,I).

% base value of T after the CFG edge at P, provided T is not directly assigned
% anything by the edge. this will be either the value of T at P, or a
% uc_sum value introduced by a potentially-clobbering isum at P
predicate base_val(in P:pp,in T:t_trace,V:scalar,G:g_guard).
base_val(P,T,V,G) :- ~cfg_isum(P,_), val(P,T,V,G).
base_val(P,T,V,G) :- cfg_isum(P,I), ~trace_clobber(I,P,T), val(P,T,V,G).
base_val(P,T,V,G) :- cfg_isum(P,I), trace_clobber(I,P,T),
    break_point(P,_,_,XP), V=s_trace{uc_sum{I,XP,T}}, #bool_g(true,G).

% if T might not be assigned to by a CFG edge, merge its base value before the
% edge with its value after the edge
vmerge(P1,T,V,EG) :- cfg_predecessor(P1,P0),
    \/cfg_assign_write(P0,T,_,AG):#or_all(AG,MAG), #not(MAG,NMAG),
    base_val(P0,T,V,VG), #and(NMAG,VG,G), eguard(P0,P1,G,EG).

% baked in behaviors

import "memorybaked.clp".
