import "traces.clp".
import "paths.clp".   % for g_guard
using map.

% ----------------------------------------------------------------------------

type scalar.
type scalar_cmp.
type scalar_bit.

% g_guard = bval[g_bit]
type g_bit = scalar_cmp.

% scalar values. the sign/length of a scalar is fixed by the binops and other
% operations that it is used in
type scalar ::=
    s_const{int}                         % leaf constant
  | s_trace{t_trace}                     % leaf unconstrained trace
  | s_cast{scalar,XS:bool,XBITS:int}     % cast from an old sign/len
  | s_unop{unop,scalar,S:bool,BITS:int}           % unary exp over a sign/len
  | s_binop{binop,scalar,scalar,S:bool,BITS:int}  % binary exp over a sign/len
  | s_guard{g_guard}.                    % guard lifted to 0/1

type scalar_cmp ::=
    sc_cmp{binop,scalar,scalar,S:bool,BITS:int}  % binary cmp over a sign/len
  | sc_eqz{scalar,S:bool,BITS:int}               % shorthand for 0 == ...
  | sc_bit{scalar_bit}.                          % extra bit

% client-defined unconstrained bit extension
type t_xbit.

type scalar_bit ::=
    sb_entry{P:pp}                   % non-deterministic entry bits
  | sb_path{P0:pp,P1:pp}             % non-deterministic branch bits
  | sb_alias{T0:t_trace,T1:t_trace}  % aliasing bits between T0/T1
  | sb_xbit{t_xbit}.

% ----------------------------------------------------------------------------

% convert a scalar to a guard representing whether that scalar is zero or not.
% the scalar should be interpreted with S/BITS sign/length
predicate scalar_guard(in V:scalar,in S:bool,in BITS:int,G:g_guard)
    succeeds [once].

scalar_guard(s_const{0},_,_,G) :- #bool_g(false,G).
scalar_guard(s_const{N},_,_,G) :- N\=0, #bool_g(true,G).
scalar_guard(s_guard{G},_,_,G) :- .
scalar_guard(s_trace{root{_}},_,_,G) :- #bool_g(true,G).
scalar_guard(S,SIGN,LEN,G) :-
    S\=s_const{_}, S\=s_guard{_}, S\=s_trace{root{_}},
    #id_g(sc_eqz{S,SIGN,LEN},NG), #not(NG,G).

entry_bit(P,G) :- #id_g(sc_bit{sb_entry{P}},G).
path_bit(P0,P1,G) :- #id_g(sc_bit{sb_path{P0,P1}},G).

% ----------------------------------------------------------------------------

predicate guard_string(in g_guard, out string) succeeds [once].
predicate scalar_string(in scalar, out string) succeeds [once].

% versions which might fail if the guard is too large
predicate guard_string_x(in g_guard, out string) succeeds [zero,once].
predicate scalar_string_x(in scalar, out string) succeeds [zero,once].
predicate scalar_cmp_string_x(in scalar_cmp, out string) succeeds [zero,once].

guard_string(G,S) :- guard_string_x(G,S).
guard_string(G,S) :- ~guard_string_x(G,_), S="???".

scalar_string(V,S) :- scalar_string_x(V,S).
scalar_string(V,S) :- ~scalar_string_x(V,_), S="???".

predicate guard_bop(in g_guard,out g_guard,out g_guard,out int,out string).
guard_bop(G,G0,G1,0," && ") :- #and(G0,G1,G).
guard_bop(G,G0,G1,1," || ") :- #or(G0,G1,G).

predicate guard_string_x_paren(in g_guard,in int,out string).
guard_string_x_paren(G,N,STR) :- ~guard_bop(G,_,_,_,_), guard_string_x(G,STR).
guard_string_x_paren(G,N,STR) :- guard_bop(G,_,_,N,_), guard_string_x(G,STR).
guard_string_x_paren(G,N,STR) :-
    guard_bop(G,_,_,XN,_), N\=XN, guard_string_x(G,BSTR),
    str_cat_list_chk(["(",BSTR,")"],STR).

guard_string_x(G,STR) :- #bool_g(B,G), tostring(B,STR).
guard_string_x(G,STR) :- #id_g(SC,G), scalar_cmp_string_x(SC,STR).
guard_string_x(G,STR) :- guard_bop(G,G0,G1,N,SEP),
    guard_string_x_paren(G0,N,G0S), guard_string_x_paren(G1,N,G1S),
    str_cat_list_chk([G0S,SEP,G1S],STR).
guard_string_x(G,STR) :- #not(G0,G), #id_g(sc_eqz{LS,_,_},G0),
    scalar_string_x(LS,STR).
guard_string_x(G,STR) :- #not(G0,G), ~#id_g(sc_eqz{_,_,_},G0),
    guard_string_x_paren(G0,2,G0S), str_cat_chk("!",G0S,STR).

scalar_cmp_string_x(sc_eqz{LS,_,_},STR) :-
    scalar_string_x(LS,LSS), str_cat_chk("!",LSS,STR).
scalar_cmp_string_x(sc_cmp{OP,LS,RS,_,_},STR) :-
    scalar_string_x(LS,LSS), scalar_string_x(RS,RSS), binop_str(OP,OPS),
    str_cat_list_chk([LSS,OPS,RSS],STR).
scalar_cmp_string_x(sc_bit{B},STR) :- tostring(B,STR).

scalar_string_x(s_const{N},STR) :- tostring(N,STR).
scalar_string_x(s_trace{T},STR) :- trace_string_x(T,STR).
scalar_string_x(s_cast{RS,_,_},STR) :-
    scalar_string_x(RS,RSS),
    str_cat_list_chk(["(int??)(",RSS,")"],STR).
scalar_string_x(s_unop{OP,RS,_,_},STR) :-
    scalar_string_x(RS,RSS), unop_str(OP,OPS),
    str_cat_list_chk([OPS,"(",RSS,")"],STR).
scalar_string_x(s_binop{OP,LS,RS,_,_},STR) :-
    scalar_string_x(LS,LSS), scalar_string_x(RS,RSS), binop_str(OP,OPS),
    str_cat_list_chk(["(",LSS,OPS,RSS,")"],STR).
scalar_string_x(s_guard{G},STR) :-
    guard_string_x(G,GS),
    str_cat_list_chk(["(",GS,")"],STR).

% ----------------------------------------------------------------------------

type s_used ::=
    su_scalar{scalar}
  | su_bit{scalar_bit}.

predicate scalar_used(in scalar,out s_used).
predicate guard_used(in g_guard,out s_used).
predicate trace_used(in t_trace,out s_used).

+scalar_used(S,su_scalar{S}).
scalar_used(s_cast{IS,_,_},XS) :- scalar_used(IS,XS).
scalar_used(s_unop{_,IS,_,_},XS) :- scalar_used(IS,XS).
scalar_used(s_binop{_,IS,_,_,_},XS) :- scalar_used(IS,XS).
scalar_used(s_binop{_,_,IS,_,_},XS) :- scalar_used(IS,XS).
scalar_used(s_guard{IG},XS) :- guard_used(IG,XS).
scalar_used(S,XS) :- scalar_used(S,su_scalar{s_trace{T}}), trace_used(T,XS).

guard_used(G,XS) :- #g_id(G,sc_eqz{IS,_,_}), scalar_used(IS,XS).
guard_used(G,XS) :- #g_id(G,sc_cmp{_,IS,_,_,_}), scalar_used(IS,XS).
guard_used(G,XS) :- #g_id(G,sc_cmp{_,_,IS,_,_}), scalar_used(IS,XS).
guard_used(G,XS) :- #g_id(G,sc_bit{B}), XS=su_bit{B}.
guard_used(G,XS) :- guard_used(G,su_scalar{s_trace{T}}), trace_used(T,XS).

trace_used(T,XS) :- trace_sub(T,index{_,_,IS},_), scalar_used(IS,XS).

% ----------------------------------------------------------------------------

predicate simplify_scalar(in scalar, out scalar) succeeds [once].
predicate simplify_scalar_cmp(in scalar_cmp, out g_guard) succeeds [once].

predicate try_simplify_scalar(in scalar, out scalar).
simplify_scalar(S,S) :- ~try_simplify_scalar(S,_).
simplify_scalar(S,NS) :- \/try_simplify_scalar(S,XS):list_all(XS,[MS|_]),
    simplify_scalar(MS,NS).

predicate try_simplify_scalar_cmp(in scalar_cmp, out g_guard).
simplify_scalar_cmp(S,G) :- ~try_simplify_scalar_cmp(S,_), #id_g(S,G).
simplify_scalar_cmp(S,G) :- \/try_simplify_scalar_cmp(S,XG):list_all(XG,[G|_]).

% collapse index traces on identical types in trace_simplify.
% TODO: assuming int32 for the scalars used in indexes. needs to be fixed
% for doing integer overflow checking
trace_one_simplify(index{index{T,Y,I0},Y,I1},index{T,Y,NI}) :-
    simplify_scalar(s_binop{b_plusa,I0,I1,true,32},NI).

% compute the signed modulus of BN according to SIGN/LEN
predicate signed_modulus(in BN:int, in SIGN:bool, in LEN:int, out N:int).
signed_modulus(BN,false,LEN,N) :-
    int_exp(2,LEN,EXP),
    int_mod(BN,EXP,N).
signed_modulus(BN,true,LEN,N) :-
    int_exp(2,LEN,EXP),
    int_sub(LEN,1,MLEN), int_exp(2,MLEN,MEXP),
    int_mod(BN,EXP,BMN),
    (int_lt(BMN,MEXP), N=BMN;
     int_ge(BMN,MEXP), int_sub(BMN,EXP,N)).

% const-fold casts
try_simplify_scalar(s_cast{CS,SIGN,LEN},S) :-
    CS = s_const{CN},
    signed_modulus(CN,SIGN,LEN,NN),
    S = s_const{NN}.

% const-fold unops
try_simplify_scalar(s_unop{u_neg,LS,SIGN,LEN},S) :-
    LS = s_const{LN},
    int_sub(0,LN,BN), signed_modulus(BN,SIGN,LEN,NN),
    S = s_const{NN}.

predicate const_fold_binop(in binop, in LN:int, in RN:int, out N:int).

const_fold_binop(b_plusa,   LN,RN,N) :- int_add(LN,RN,N).
const_fold_binop(b_band,    LN,RN,N) :- int_ge(LN,0), int_ge(RN,0), int_band(LN,RN,N).
const_fold_binop(b_bxor,    LN,RN,N) :- int_ge(LN,0), int_ge(RN,0), int_bxor(LN,RN,N).
const_fold_binop(b_bor,     LN,RN,N) :- int_ge(LN,0), int_ge(RN,0), int_bor(LN,RN,N).
const_fold_binop(b_mult,    LN,RN,N) :- int_mul(LN,RN,N).
const_fold_binop(b_minusa,  LN,RN,N) :- int_sub(LN,RN,N).
const_fold_binop(b_div,     LN,RN,N) :- RN\=0, int_div(LN,RN,N).
const_fold_binop(b_mod,     LN,RN,N) :- RN\=0, int_mod(LN,RN,N).
const_fold_binop(b_shiftlt, LN,RN,N) :- int_ge(RN,0), int_exp(2,RN,EXP), int_mul(LN,EXP,N).
const_fold_binop(b_shiftrt, LN,RN,N) :- int_ge(RN,0), int_exp(2,RN,EXP), int_div(LN,EXP,N).

% const-fold binary operations. TODO: should check for mismatched scalar kinds
try_simplify_scalar(s_binop{OP,LS,RS,SIGN,LEN},S) :-
    LS = s_const{LN}, RS = s_const{RN},
    const_fold_binop(OP,LN,RN,BN),
    signed_modulus(BN,SIGN,LEN,NN),
    S = s_const{NN}.

% arithmetic axiom: 0+x == x
+try_simplify_scalar(s_binop{b_plusa,s_const{0},RS,_,_},RS).
+try_simplify_scalar(s_binop{b_plusa,LS,s_const{0},_,_},LS).

% arithmetic axiom: x-x == 0
try_simplify_scalar(s_binop{b_minusa,LS,LS,_,_},ZS) :-
    ZS = s_const{0}.

% arithmetic axiom: x+(-y) == x-y
try_simplify_scalar(s_binop{b_plusa,LS,s_unop{u_neg,RS,_,_},SIGN,LEN},S) :-
    S = s_binop{b_minusa,LS,RS,SIGN,LEN}.
try_simplify_scalar(s_binop{b_plusa,s_unop{u_neg,LS,_,_},RS,SIGN,LEN},S) :-
    S = s_binop{b_minusa,RS,LS,SIGN,LEN}.

predicate commutative_binop(out binop).
commutative_binop(b_plusa) :- .
commutative_binop(b_band) :- .
commutative_binop(b_bxor) :- .
commutative_binop(b_bor) :- .
commutative_binop(b_mult) :- .

% canonicalize commutative binary operations
try_simplify_scalar(s_binop{OP,LS,RS,SIGN,LEN},s_binop{OP,RS,LS,SIGN,LEN}) :-
    commutative_binop(OP), lessthan(RS,LS).

predicate const_fold_cmp(in binop, in LN:int, in RN:int).

const_fold_cmp(b_eq, LN,RN) :- int_eq(LN,RN).
const_fold_cmp(b_ne, LN,RN) :- int_ne(LN,RN).
const_fold_cmp(b_lt, LN,RN) :- int_lt(LN,RN).
const_fold_cmp(b_le, LN,RN) :- int_le(LN,RN).
const_fold_cmp(b_gt, LN,RN) :- int_gt(LN,RN).
const_fold_cmp(b_ge, LN,RN) :- int_ge(LN,RN).

% const-fold comparisons
try_simplify_scalar_cmp(sc_cmp{OP,LS,RS,SIGN,LEN},G) :-
    LS = s_const{LN}, RS = s_const{RN},
    signed_modulus(LN,SIGN,LEN,NLN),
    signed_modulus(RN,SIGN,LEN,NRN),
    (const_fold_cmp(OP,NLN,NRN), #bool_g(true,G);
     ~const_fold_cmp(OP,NLN,NRN), #bool_g(false,G)).
try_simplify_scalar_cmp(sc_eqz{S,SIGN,LEN},G) :-
    S = s_const{N},
    signed_modulus(N,SIGN,LEN,NN),
    (int_eq(NN,0), #bool_g(true,G);
     int_ne(NN,0), #bool_g(false,G)).

try_simplify_scalar_cmp(sc_cmp{b_eq,LS,RS,SIGN,LEN},G) :-
    (LS = s_const{0}, NS = RS;
     RS = s_const{0}, NS = LS),
    NS \= s_const{_},
    #id_g(sc_eqz{NS,SIGN,LEN},G).
try_simplify_scalar_cmp(sc_cmp{b_ne,LS,RS,SIGN,LEN},G) :-
    (LS = s_const{0}, NS = RS;
     RS = s_const{0}, NS = LS),
    NS \= s_const{_},
    #id_g(sc_eqz{NS,SIGN,LEN},NG), #not(NG, G).

% ----------------------------------------------------------------------------

% maps for applying simple substitution transformations on a guard/scalar
type ts_xmap.
type ts_map ::=
      % map to replace all occurrences of RT with NV
    ts_reptrace{RT:t_trace,NV:scalar}
  | ts_remove_uc_sum
      % map to replace all occurrences of any value in MAP/BMAP with
      % the corresponding target
  | ts_repmap{MAP:map[t_trace,scalar],BMAP:map[scalar_bit,g_guard]}
      % client-extensible mapping
  | ts_extra{ts_xmap}.

predicate try_map_guard(in ts_map,in RG:g_guard,NG:g_guard).
predicate try_map_scalar(in ts_map,in RS:scalar,NS:scalar).

predicate base_map_guard(in ts_map,in RG:g_guard).
predicate base_map_scalar(in ts_map,in RS:scalar).

predicate map_guard(in ts_map,in RG:g_guard,NG:g_guard) succeeds [once].
predicate map_scalar(in ts_map,in RS:scalar,NS:scalar) succeeds [once].

% preserves form of trace, just runs map_scalar over any array offset indexes
predicate map_trace(in ts_map,in RT:t_trace,NT:t_trace) succeeds [once].

map_guard(X,RG,NG) :- try_map_guard(X,RG,NG).
map_scalar(X,RS,NS) :- try_map_scalar(X,RS,NS).

map_guard(X,RG,RG) :- base_map_guard(X,RG),
    #bool_g(_,RG).
map_guard(X,RG,NG) :- base_map_guard(X,RG),
    #and(RG0,RG1,RG),
    map_guard(X,RG0,NG0),
    map_guard(X,RG1,NG1),
    #and(NG0,NG1,NG).
map_guard(X,RG,NG) :- base_map_guard(X,RG),
    #or(RG0,RG1,RG),
    map_guard(X,RG0,NG0),
    map_guard(X,RG1,NG1),
    #or(NG0,NG1,NG).
map_guard(X,RG,NG) :- base_map_guard(X,RG),
    #not(RG0,RG),
    map_guard(X,RG0,NG0),
    #not(NG0,NG).
map_guard(X,RG,NG) :- base_map_guard(X,RG),
    #id_g(sc_cmp{B,RS0,RS1,SIGN,LEN},RG),
    map_scalar(X,RS0,NS0),
    map_scalar(X,RS1,NS1),
    simplify_scalar_cmp(sc_cmp{B,NS0,NS1,SIGN,LEN},NG).
map_guard(X,RG,NG) :- base_map_guard(X,RG),
    #id_g(sc_eqz{RS,SIGN,LEN},RG),
    map_scalar(X,RS,NS),
    simplify_scalar_cmp(sc_eqz{NS,SIGN,LEN},NG).
map_guard(X,RG,RG) :- base_map_guard(X,RG),
    #id_g(sc_bit{_},RG).

map_scalar(X,RS,RS) :- base_map_scalar(X,RS),
    RS=s_const{_}.
map_scalar(X,RS,NS) :- base_map_scalar(X,RS),
    RS=s_trace{RT},
    map_trace(X,RT,NT),
    NS=s_trace{NT}.
map_scalar(X,RS,NS) :- base_map_scalar(X,RS),
    RS=s_cast{RS0,SIGN,LEN},
    map_scalar(X,RS0,NS0),
    simplify_scalar(s_cast{NS0,SIGN,LEN},NS).
map_scalar(X,RS,NS) :- base_map_scalar(X,RS),
    RS=s_unop{OP,RS0,SIGN,LEN},
    map_scalar(X,RS0,NS0),
    simplify_scalar(s_unop{OP,NS0,SIGN,LEN},NS).
map_scalar(X,RS,NS) :- base_map_scalar(X,RS),
    RS=s_binop{OP,RS0,RS1,SIGN,LEN},
    map_scalar(X,RS0,NS0),
    map_scalar(X,RS1,NS1),
    simplify_scalar(s_binop{OP,NS0,NS1,SIGN,LEN},NS).
map_scalar(X,RS,NS) :- base_map_scalar(X,RS),
    RS=s_guard{RG},
    map_guard(X,RG,NG),
    NS=s_guard{NG}.

map_trace(X,drf{RT},drf{NT}) :- map_trace(X,RT,NT).
map_trace(X,fld{RT,F,C},fld{NT,F,C}) :- map_trace(X,RT,NT).
map_trace(X,rfld{RT,F,C},rfld{NT,F,C}) :- map_trace(X,RT,NT).
map_trace(X,index{RT,Y,RS},index{NT,Y,NS}) :-
    map_trace(X,RT,NT), map_scalar(X,RS,NS).

map_trace(_,T,T) :- T\=drf{_}, T\=fld{_,_,_}, T\=rfld{_,_,_}, T\=index{_,_,_}.

% ----------------------------------------------------------------------------

% maps for applying a per-point trace->(scalar,guard) transformation
% on a guard/scalar. all combinations of scalar/guard pairs for each mentioned
% trace are collected into a set of maps, which are then applied with ts_repmap
type ts_xgmap.
type ts_gmap ::=
      % apply inst_trace to each trace in the guard/scalar
    ts_inst{I:sum}
      % apply convert_point_trace to each trace in the guard/scalar
  | ts_convert
      % client-extensible mapping
  | ts_gextra{ts_xgmap}.

predicate try_gmap_trace(in ts_gmap,in pp,in T:t_trace,NV:scalar,G:g_guard).
predicate try_gmap_bit(in ts_gmap,in pp,in B:scalar_bit,NG:g_guard,G:g_guard).

% collect all maps for a list of traces
predicate gmap_trace_list(in ts_gmap,in pp,in VLIST:list[t_trace],
                          VMAP:map[t_trace,scalar],G:g_guard).
gmap_trace_list(_,_,[],MAP,TG) :-
    map_empty(MAP), #bool_g(true,TG).
gmap_trace_list(X,P,[T|TAIL],NMAP,NG) :-
    gmap_trace_list(X,P,TAIL,MAP,G0),
    try_gmap_trace(X,P,T,V,G1),
    map_insert(T,V,MAP,NMAP), #and(G0,G1,NG), ~#bool_g(false,NG).

% collect all maps for a list of bits
predicate gmap_bit_list(in ts_gmap,in pp,in BLIST:list[scalar_bit],
                        BMAP:map[scalar_bit,g_guard],G:g_guard).
gmap_bit_list(_,_,[],MAP,TG) :-
    map_empty(MAP), #bool_g(true,TG).
gmap_bit_list(X,P,[B|TAIL],NMAP,NG) :-
    gmap_bit_list(X,P,TAIL,MAP,G0),
    try_gmap_bit(X,P,B,BG,G1),
    map_insert(B,BG,MAP,NMAP), #and(G0,G1,NG), ~#bool_g(false,NG).

predicate gmap_scalar(in ts_gmap,in pp,in V:scalar,NV:scalar,WG:g_guard).
gmap_scalar(X,P,V,NV,WG) :-
    \/scalar_used(V,su_scalar{s_trace{T}}):list_all(T,VLIST),
    \/scalar_used(V,su_bit{B}):list_all(B,BLIST),
    gmap_trace_list(X,P,VLIST,VMAP,G0), gmap_bit_list(X,P,BLIST,BMAP,G1),
    map_scalar(ts_repmap{VMAP,BMAP},V,NV), #and(G0,G1,WG).

predicate gmap_guard(in ts_gmap,in pp,in G:g_guard,NG:g_guard,WG:g_guard).
gmap_guard(X,P,G,NG,WG) :-
    \/guard_used(G,su_scalar{s_trace{T}}):list_all(T,VLIST),
    \/guard_used(G,su_bit{B}):list_all(B,BLIST),
    gmap_trace_list(X,P,VLIST,VMAP,G0), gmap_bit_list(X,P,BLIST,BMAP,G1),
    map_guard(ts_repmap{VMAP,BMAP},G,NG), #and(G0,G1,WG).

% -----------------------------------------------------------------------------

predicate reptrace(in RT:t_trace,in NV:scalar,in RBT:t_trace,NBV:scalar).
+reptrace(RT,NV,RT,NV).
reptrace(RT,s_trace{NT},RBT,s_trace{NBT}) :-
    trace_sub(RBT,RT,XT), trace_compose(NT,XT,NBT).

try_map_scalar(ts_reptrace{RT,NV},s_trace{RBT},NBV) :-
    map_trace(ts_reptrace{RT,NV},RBT,XBT),
    (reptrace(RT,NV,XBT,NBV);
     ~reptrace(RT,NV,XBT,_), NBV=s_trace{XBT}).
base_map_scalar(ts_reptrace{_,_},S) :- S\=s_trace{_}.
base_map_guard(ts_reptrace{_,_},_) :- .

try_map_scalar(ts_remove_uc_sum,s_trace{BT},s_trace{NT}) :-
    trace_remove_uc_sum(BT,NT).
base_map_scalar(ts_remove_uc_sum,S) :- S\=s_trace{_}.
base_map_guard(ts_remove_uc_sum,_) :- .

try_map_scalar(ts_repmap{MAP,_},s_trace{RT},NS) :- map_search(MAP,RT,NS).
base_map_scalar(ts_repmap{MAP,_},s_trace{XT}) :- ~map_search(MAP,XT,_).
base_map_scalar(ts_repmap{_,_},S) :- S\=s_trace{_}.

try_map_guard(ts_repmap{_,MAP},RG,NG) :- #id_g(sc_bit{B},RG), map_search(MAP,B,NG).
base_map_guard(ts_repmap{_,MAP},RG) :- #id_g(sc_bit{B},RG), ~map_search(MAP,B,_).
base_map_guard(ts_repmap{_,_},G) :- ~#id_g(sc_bit{_},G).
