% ----------------------------------------------------------------------------
% SAT-based implementation of the symbol integer operations
% Integer operations are modelled as boolean formulae. Floats are modelled as
% uninterpreted functions.
% ----------------------------------------------------------------------------

import "scalar.clp".
using solve_sat.
using solve_bitvector.

% ----------------------------------------------------------------------------
% Integer expressions, bit vector and bit formulae

type g_vecrep ::=
    vr_t{t_trace}
  | vr_u{binop,scalar,scalar}.
type g_arep ::=
    va_b{scalar_bit}.

type g_vec = bvec[g_vecrep, g_arep].
type g_guard_sat = bval[bvecbit[g_vecrep, g_arep]].
type g_guard_asn = bval_asn[bvecbit[g_vecrep, g_arep]].

% ----------------------------------------------------------------------------

predicate scalar_bvec(in scalar, in S:bool, in BITS:int, out g_vec) succeeds [once].
predicate scalar_cmp_bval(in scalar_cmp, out g_guard_sat) succeeds [once].
predicate guard_bval(in g_guard, out g_guard_sat) succeeds [once].

% ----------------------------------------------------------------------------
% Convert a integer expression to a bvec

scalar_bvec(s_const{N},SIGN,BITS,V) :-
    #int_bv(N,SIGN,BITS,V).
scalar_bvec(s_trace{T},SIGN,BITS,V) :-
    #id_bv(vr_t{T},SIGN,false,BITS,V).
scalar_bvec(s_cast{IE0,XSIGN,XBITS},SIGN,BITS,V) :-
    scalar_bvec(IE0,XSIGN,XBITS,V0),
    #bv_cast(V0,SIGN,BITS,V).

predicate scalar_unop(in U:unop,in V0:g_vec,V:g_vec) succeeds [once].
scalar_unop(u_neg,V0,V) :- #bv_neg(V0,V).
scalar_unop(u_bnot,V0,V) :- #bv_bnot(V0,V).

scalar_bvec(s_unop{U,IE0,SIGN,BITS},_,_,V) :- 
    scalar_bvec(IE0,SIGN,BITS,V0), scalar_unop(U,V0,V).

predicate scalar_binop(in B:binop,in V0:g_vec,in V1:g_vec,V:g_vec)
    succeeds [zero,once].
scalar_binop(b_plusa,V0,V1,V) :- #bv_add(V0,V1,V).
scalar_binop(b_minusa,V0,V1,V) :- #bv_sub(V0,V1,V).
scalar_binop(b_band,V0,V1,V) :- #bv_band(V0,V1,V).
scalar_binop(b_bor,V0,V1,V) :- #bv_bor(V0,V1,V).
scalar_binop(b_bxor,V0,V1,V) :- #bv_bxor(V0,V1,V).
scalar_binop(b_mult,V0,V1,V) :- #bv_int(V0,N0), #bv_mul(V1,N0,V).
scalar_binop(b_mult,V0,V1,V) :- ~#bv_int(V0,_), #bv_int(V1,N1), #bv_mul(V0,N1,V).
scalar_binop(b_shiftlt,V0,V1,V) :- #bv_int(V1,N1), #bv_shl(V0,N1,V).
scalar_binop(b_shiftrt,V0,V1,V) :- #bv_int(V1,N1), #bv_shr(V0,N1,V).

scalar_bvec(s_binop{B,IE0,IE1,SIGN,BITS},_,_,V) :-
    scalar_bvec(IE0,SIGN,BITS,V0), scalar_bvec(IE1,SIGN,BITS,V1),
    scalar_binop(B,V0,V1,V).
scalar_bvec(s_binop{B,IE0,IE1,SIGN,BITS},_,_,V) :-
    scalar_bvec(IE0,SIGN,BITS,V0), scalar_bvec(IE1,SIGN,BITS,V1),
    ~scalar_binop(B,V0,V1,_),
    #id_bv(vr_u{B,IE0,IE1},SIGN,false,BITS,V).
scalar_bvec(s_guard{G},SIGN,BITS,V) :-
    guard_bval(G,B), #g_bv(B,SIGN,BITS,V).

% ----------------------------------------------------------------------------
% Guard to SAT formula conversion

guard_bval(G,B) :- #bool_g(true,G), #bool_g(true,B).
guard_bval(G,B) :- #bool_g(false,G), #bool_g(false,B).
guard_bval(RG,NG) :- #id_g(RC,RG), scalar_cmp_bval(RC,NG).

guard_bval(RG,NG) :-
    #and(RG0,RG1,RG),
    guard_bval(RG0,NG0),
    guard_bval(RG1,NG1),
    #and(NG0,NG1,NG).

guard_bval(RG,NG) :-
    #or(RG0,RG1,RG),
    guard_bval(RG0,NG0),
    guard_bval(RG1,NG1),
    #or(NG0,NG1,NG).

guard_bval(RG,NG) :-
    #not(RG0,RG),
    guard_bval(RG0,NG0),
    #not(NG0,NG).

% ----------------------------------------------------------------------------
% Comparison operation to SAT formula conversion

predicate scalar_cmp_binop(in B:binop,in V0:g_vec,in V1:g_vec,G:g_guard_sat)
    succeeds [once].
scalar_cmp_binop(b_eq,V0,V1,G) :- #bv_eq(V0,V1,G).
scalar_cmp_binop(b_ne,V0,V1,G) :- #bv_ne(V0,V1,G).
scalar_cmp_binop(b_lt,V0,V1,G) :- #bv_lt(V0,V1,G).
scalar_cmp_binop(b_le,V0,V1,G) :- #bv_le(V0,V1,G).
scalar_cmp_binop(b_gt,V0,V1,G) :- #bv_gt(V0,V1,G).
scalar_cmp_binop(b_ge,V0,V1,G) :- #bv_ge(V0,V1,G).

scalar_cmp_bval(sc_cmp{B,IE0,IE1,SIGN,BITS},G) :-
    scalar_bvec(IE0,SIGN,BITS,V0), scalar_bvec(IE1,SIGN,BITS,V1),
    scalar_cmp_binop(B,V0,V1,G).
scalar_cmp_bval(sc_eqz{IE,SIGN,BITS},G) :-
    scalar_bvec(IE,SIGN,BITS,V), #bv_eqz(V,G).
scalar_cmp_bval(sc_bit{B},G) :-
    #id_g(b_abit{va_b{B}},G).

% ----------------------------------------------------------------------------
% Satisfiability predicates

guard_sat(G) :- guard_bval(G,B), #sat(B).
guard_valid(G) :- #not(G,NG), ~guard_sat(NG).
guard_implies(G0,G1) :- #not(G0,NG0), #or(NG0,G1,IMP), guard_valid(IMP).
guard_equivalent(G0,G1) :- guard_implies(G0,G1), guard_implies(G1,G0).
guard_disjoint(G0,G1) :- #and(G0,G1,BOTH), ~guard_sat(BOTH).
