%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% canonical memory representation and interprocedural points-to information

% location traces are canonical representations of memory locations accessed by
% a function or loop, to be used when summarizing the function/loop etc.
% each trace is either a root R, typically an argument or global, or
% the offset of a trace drf{T} (initial target of the location at T)
% or fld{T,F} (field F of the location at T)

import "../base/cilbase.clp".
import "../base/utility.clp".
using intops.

type pp.
type t_trace.
type scalar.

% root stack/static locations
type t_root ::=
    arg{A:int}       % argument number A
  | glob{G:string}   % global variable G
  | local{L:string}  % local variable L (used intraprocedurally)
  | temp{TMP:string,WHERE:string,Y:t_type}
                     % temporary TMP with description WHERE and type Y
  | return           % the return value
  | asm_in{A:int}    % numbered input argument to inline assembly
  | asm_out{A:int}   % numbered output argument to inline assembly
  | cstr{string}     % a constant string (treated as read-only global data)
  .

% client-defined trace extension
type t_xtrace.

% canonical representations of locations used by the program
type t_trace ::=
    root{R:t_root}  % root trace for the stack/static location of R
  | drf{T:t_trace}  % the initial result of dereferencing T
  | fld{T:t_trace,F:string,C:string}    % field F of T
  | rfld{T:t_trace,F:string,C:string}   % reverse field F of T
  | index{T:t_trace,Y:t_type,I:scalar}  % array index I by type Y of T
    % skolem traces
  | uk_index{T:t_trace,Y:t_type}           % some array element of T
  | uk_recurse{T:t_trace,RL:list[t_trace]} % some element of recursive data
                                           % structure reachable from T via RL
    % the following traces are NOT rooted
  | empty                         % empty trace, for pure location offsets
  | uc_exp{P:pp,E:c_exp,Y:t_type} % result of evaluating E at P, E has type Y
  | uc_val{P:pp,T:t_trace}        % value of T at P
  | uc_sum{I:sum,P:pp,T:t_trace}  % value of T after being clobbered by I at P
  .

% TRACE PREDICATES

% CIL variable X has root R
predicate var_trace(X:c_var,R:t_root).

% NT is a simplified version of OT, removing any redundancy
predicate trace_simplify(in OT:t_trace,NT:t_trace) succeeds [once].

% trace T is rooted at R. fails on traces containing 'empty' or any 'uc_' trace
predicate trace_root(in T:t_trace,R:t_root) succeeds [zero,once].

% trace T is relative and contains 'empty' instead of a root
predicate trace_relative(in T:t_trace).

% trace T is unconstrained and was derived from a 'uc_' trace
predicate trace_unconstrained(in T:t_trace).

% trace T includes N dereferences. not defined for unconstrained traces
predicate trace_derefs(in T:t_trace,N:int) succeeds [zero,once].

% kinds of ways in which traces can be allocated
type t_alloc_kind ::=
    ak_external     % allocated externally from SUM 
  | ak_stack_local  % stack-allocated within the body of SUM
  | ak_heap_local.  % heap-allocated by a call within the body of SUM

% get the allocation kind of a trace, or none if unknown
predicate trace_alloc(in SUM:sum,in T:t_trace,K:t_alloc_kind)
    succeeds [zero,once].

% T0 and T1 compose into trace TR. T1 must have an empty root
predicate trace_compose(in T0:t_trace,in T1:t_trace,TR:t_trace)
    succeeds [once].

% get TR such that TS composes with TR into T
predicate trace_sub(in T:t_trace,TS:t_trace,TR:t_trace) succeeds [many].

% TS/TR form a trace_sub of T within the same structure (i.e. no dereferences)
predicate trace_fld_sub(in T:t_trace,TS:t_trace,TR:t_trace) succeeds [many].

% strip all outer fld constructors from T
predicate trace_strip_flds(in T:t_trace,BT:t_trace) succeeds [once].

% remov all uc_sum{_,_,PT} from within T and replace with drf{PT}
predicate trace_remove_uc_sum(in T:t_trace,NT:t_trace) succeeds [once].

% as interpreted by function FN, the memory location referred to by T has
% type TYP. FN is ignored for traces rooted at glob{}, which have the same
% type across all functions
predicate trace_type(in FN:string,in T:t_trace,Y:t_type).

% TRACE PRINTING PREDICATES

% convert traces to a more readable string representation
predicate trace_string(in t_trace,string) succeeds [once].

% convert a summary to a UI string
predicate dstring_sum(in sum,string) succeeds [once].

% GENERAL SUMMARY INFO

% convert a trace T (non-empty root) to another trace RT relative to STR
% (with a possibly empty root), if possible
predicate comp_to_relative(in STR:t_trace,in T:t_trace,RT:t_trace).

% convert a relative trace RT (possibly-empty root) in a composite invariant
% on STR to a trace T (non-empty root)
predicate comp_from_relative(in STR:t_trace,in RT:t_trace,T:t_trace).

% SCALARS

import "scalar.clp".

% RULES

% trace inspection/manipulation

+trace_compose(T,empty,T).
trace_compose(T0,drf{PT1},TR) :-
    trace_compose(T0,PT1,PTR), trace_simplify(drf{PTR},TR).
trace_compose(T0,fld{FT1,F,C},TR) :-
    trace_compose(T0,FT1,FTR), trace_simplify(fld{FTR,F,C},TR).
trace_compose(T0,rfld{FT1,F,C},TR) :-
    trace_compose(T0,FT1,FTR), trace_simplify(rfld{FTR,F,C},TR).
trace_compose(T0,index{IT1,Y,I},TR) :-
    trace_compose(T0,IT1,ITR), trace_simplify(index{ITR,Y,I},TR).
trace_compose(T0,uk_index{T1,Y},TR) :-
    trace_compose(T0,T1,KTR), trace_simplify(uk_index{KTR,Y},TR).
trace_compose(T0,uk_recurse{T1,RL},TR) :-
    trace_compose(T0,T1,KTR), trace_simplify(uk_recurse{KTR,RL},TR).

+trace_sub(T,T,empty).
trace_sub(drf{T},TS,drf{TR}) :- trace_sub(T,TS,TR).
trace_sub(fld{T,F,C},TS,fld{TR,F,C}) :- trace_sub(T,TS,TR).
trace_sub(rfld{T,F,C},TS,rfld{TR,F,C}) :- trace_sub(T,TS,TR).
trace_sub(index{T,Y,I},TS,index{TR,Y,I}) :- trace_sub(T,TS,TR).
trace_sub(uk_index{T,Y},TS,uk_index{TR,Y}) :- trace_sub(T,TS,TR).
trace_sub(uk_recurse{T,RL},TS,uk_recurse{TR,RL}) :- trace_sub(T,TS,TR).

% get trace derefs
+trace_derefs(root{_},0), +trace_derefs(empty,0).
trace_derefs(drf{T},N) :- trace_derefs(T,SN), int_add(SN,1,N).
trace_derefs(fld{T,_,_},N) :- trace_derefs(T,N).
trace_derefs(rfld{T,_,_},N) :- trace_derefs(T,N).
trace_derefs(index{T,_,_},N) :- trace_derefs(T,N).
trace_derefs(uk_index{T,_},N) :- trace_derefs(T,N).
trace_derefs(uk_recurse{T,_},N) :- trace_derefs(T,N),
    trace_string(T,TS), +warning("trace_derefs() applied to uk_recurse trace",TS).

trace_fld_sub(T0,T1,TR) :- trace_sub(T0,T1,TR), trace_derefs(TR,0).

% variable traces

cil_fundec_formal(_,A,X), +var_trace(X,arg{A}).
cil_var_global(X), cil_var_name(X,G,_), +var_trace(X,glob{G}).
cil_var_static(X), cil_var_name(X,G,_), +var_trace(X,glob{G}).
cil_var_local(X),  cil_var_name(X,L,_), +var_trace(X,local{L}).
cil_var_return(X), +var_trace(X,return).

% get the terms for each temporary. there are three cases to handle.
% - if call return value, use WHERE to indicate the call
% - if return value for multiple calls, use "mcall" WHERE
% - other temporary values, use "op" WHERE
cil_var_temporary(X), callretvar(I,X), ~(callretvar(XI,X), XI\=I),
    cil_instr_call(I,E), exp_string(E,ES),
    cil_var_name(X,T,TYP), convert_type(TYP,Y),
    +var_trace(X,temp{T,ES,Y}).
cil_var_temporary(X), callretvar(I,X), callretvar(XI,X), XI\=I,
    cil_var_name(X,T,TYP), convert_type(TYP,Y),
    +var_trace(X,temp{T,"multicall",Y}).
cil_var_temporary(X), ~callretvar(_,X),
    cil_var_name(X,T,TYP), convert_type(TYP,Y),
    +var_trace(X,temp{T,"op",Y}).

% trace types

% argument types
trace_type(FN,root{arg{AN}},Y) :-
    cil_body(FN)->cil_fundec_formal(_,AN,X),
    cil_body(FN)->cil_var_name(X,_,TYP), convert_func_type(FN,TYP,Y).

% return value types
trace_type(FN,root{return},Y) :-
    cil_body(FN)->cil_fundec_name(_,X,_),
    cil_body(FN)->cil_var_name(X,_,FNTYP),
    cil_body(FN)->cil_type_func(FNTYP,RTYP,_),
    convert_func_type(FN,RTYP,Y).

% local/temporary types
trace_type(FN,root{local{L}},Y) :-
    cil_body(FN)->cil_var_name(_,L,TYP), convert_func_type(FN,TYP,Y).
+trace_type(_,root{temp{_,_,Y}},Y).

% global types. this includes both global variables and function symbols,
% so we will query both cil_glob and cil_body
trace_type(_,root{glob{G}},Y) :-
    cil_glob(G)->cil_var_name(_,G,TYP), convert_glob_type(G,TYP,Y).
trace_type(_,root{glob{G}},Y) :-
    cil_body(G)->cil_var_name(_,G,TYP), convert_func_type(G,TYP,Y).

% constant strings are character arrays
trace_type(_,root{cstr{STR}},y_array{y_int{ichar},LEN}) :-
    str_len(STR,ZPOS), int_add(ZPOS,1,LEN).

% compute drf/fld/rfld trace types, loading composite info if required
trace_type(FN,drf{T},DY) :- trace_type(FN,T,y_ptr{DY}).
trace_type(FN,fld{T,F,C},FY) :-
    cil_comp(C)->cil_field_name(FLD,F,FTYP,_,_),
    cil_comp(C)->cil_comp_field(COMP,_,FLD),
    cil_comp(C)->cil_comp_name(COMP,C,_),
    convert_comp_type(C,FTYP,FY).
+trace_type(_,rfld{_,_,C},y_comp{C}).

% indexes change to the element type
+trace_type(_,index{_,Y,_},Y).

% trace types for skolemized traces
+trace_type(_,uk_index{_,Y},Y).
trace_type(FN,uk_recurse{T,_},Y) :- trace_type(FN,T,Y).

% unconstrained expression/clobbered value types
+trace_type(_,uc_exp{_,_,y_ptr{Y}},Y).
trace_type(FN,uc_val{_,T},Y) :-
    trace_type(FN,drf{T},Y).
trace_type(FN,uc_sum{_,_,T},Y) :-
    trace_type(FN,drf{T},Y).

% generate both the array and elem type for arrays. TODO: revisit?
trace_type(FN,T,EY) :- trace_type(FN,T,y_array{EY,_}).

% trace simplification

% one simplification step simplifies OT to NT.
% other simplifiers may be defined outside this file
predicate trace_one_simplify(in OT:t_trace,NT:t_trace).

% simplify fld{rfld{T,F},F} and rfld{fld{T,F},F} to T
+trace_one_simplify(fld{rfld{T,F,C},F,C},T).
+trace_one_simplify(rfld{fld{T,F,C},F,C},T).

+trace_one_simplify(uk_index{uk_index{T,Y},Y},uk_index{T,Y}).
+trace_one_simplify(uk_recurse{uk_recurse{T,RL},RL},uk_recurse{T,RL}).

% TODO: uk_index and uk_recurse should be folded together
+trace_one_simplify(uk_index{uk_recurse{T,RL},Y},uk_recurse{uk_index{T,Y},RL}).

% squish together indexes of same width. sometimes arrays are indexed
% both as ints and uints, etc.
trace_one_simplify(uk_index{uk_index{T,Y0},Y1},uk_index{T,Y0}) :-
    type_width(Y0,WIDTH), type_width(Y1,WIDTH).

% simplify traces by (recursively) applying simplifications
trace_simplify(OT,NT) :-
    trace_one_simplify(OT,T),
    ~(trace_one_simplify(OT,XT), XT\=T, trace_sub(T,XT,_)),
    trace_simplify(T,NT).
trace_simplify(OT,OT) :- ~trace_one_simplify(OT,_).

% trace roots and kinds

% get trace roots
+trace_root(root{R},R).
trace_root(drf{T},R) :- trace_root(T,R).
trace_root(fld{T,_,_},R) :- trace_root(T,R).
trace_root(rfld{T,_,_},R) :- trace_root(T,R).
trace_root(index{T,_,_},R) :- trace_root(T,R).
trace_root(uk_index{T,_},R) :- trace_root(T,R).
trace_root(uk_recurse{T,_},R) :- trace_root(T,R).

% get relative/unconstrained traces
trace_relative(T) :- ~trace_root(T,_), trace_sub(T,empty,_).
trace_unconstrained(T) :- ~trace_root(T,_), ~trace_sub(T,empty,_).

predicate trace_alloc_root(in SUM:sum,in R:t_root,
                           NODRF_KIND:t_alloc_kind,DRF_KIND:t_alloc_kind).
trace_alloc(SUM,T,K) :-
    SUM\=s_call{_}, trace_root(T,R), trace_derefs(T,N),
    (N=0, trace_alloc_root(SUM,R,K,_);
     N\=0, trace_alloc_root(SUM,R,_,K)).
trace_alloc(s_call{_},T,K) :- trace_alloc(s_func,T,K).

% data reachable from global variables were always allocated externally
+trace_alloc_root(_,glob{_},ak_external,ak_external).

% constant strings are always visible
+trace_alloc_root(_,cstr{_},ak_external,ak_external).

% data originating from locals and stack arguments were allocated external
% to a loop invocation
+trace_alloc_root(s_loop{_},arg{_},ak_external,ak_external).
+trace_alloc_root(s_loop{_},local{_},ak_external,ak_external).

% as arguments are passed by value, only non-root data originating from
% arguments were allocated external to the function itself
+trace_alloc_root(s_func,arg{_},ak_stack_local,ak_external).
+trace_alloc_root(s_func,local{_},ak_stack_local,ak_heap_local).

% consider the return value as passed by reference
+trace_alloc_root(_,return,ak_external,ak_external).

% temporaries are always stack local
+trace_alloc_root(_,temp{_,_,_},ak_stack_local,ak_heap_local).

% input/output arguments to assembly are always treated as visible
+trace_alloc_root(s_asm{_},asm_in{_},ak_external,ak_external).
+trace_alloc_root(s_asm{_},asm_out{_},ak_external,ak_external).

% no allocation kinds for uc_* traces. we don't know where these came from

% removing outer fields

trace_strip_flds(fld{T1,_,_},T2) :- trace_strip_flds(T1,T2).
trace_strip_flds(T,T) :- T\=fld{_,_,_}.

% removing uc_sum

+trace_remove_uc_sum(root{R},root{R}).
+trace_remove_uc_sum(uc_exp{P,E,Y},uc_exp{P,E,Y}).
+trace_remove_uc_sum(uc_val{P,T},uc_val{P,T}).
trace_remove_uc_sum(drf{T},drf{NT}) :- trace_remove_uc_sum(T,NT).
trace_remove_uc_sum(fld{T,F,C},fld{NT,F,C}) :- trace_remove_uc_sum(T,NT).
trace_remove_uc_sum(rfld{T,F,C},rfld{NT,F,C}) :- trace_remove_uc_sum(T,NT).
trace_remove_uc_sum(index{T,Y,I},index{NT,Y,I}) :- trace_remove_uc_sum(T,NT).
trace_remove_uc_sum(uc_sum{_,_,T},drf{NT}) :- trace_remove_uc_sum(T,NT).

trace_remove_uc_sum(uk_index{T,Y},uk_index{NT,Y}) :- trace_remove_uc_sum(T,NT).
trace_remove_uc_sum(uk_recurse{T,R},uk_recurse{NT,R}) :- trace_remove_uc_sum(T,NT).

% stringifying

% version which might fail if the trace is too large due to inner index
predicate trace_string_x(in T:t_trace,STR:string) succeeds [zero,once].
trace_string(T,STR) :- trace_string_x(T,STR).
trace_string(T,STR) :- ~trace_string_x(T,_), STR="???".

% stringify root traces
trace_string_x(root{arg{A}},S) :- tostring(A,AS), str_cat("__arg",AS,S).
trace_string_x(root{asm_in{A}},S) :- tostring(A,AS), str_cat("__asmin",AS,S).
trace_string_x(root{asm_out{A}},S) :- tostring(A,AS), str_cat("__asmout",AS,S).
+trace_string_x(root{glob{G}},G).
+trace_string_x(root{local{L}},L).
trace_string_x(root{cstr{S}},SV) :- str_cat3("\"",S,"\"",SV).
trace_string_x(root{temp{T,W,_}},S) :- str_cat("tmp.",W,S).
+trace_string_x(root{return},"return").

% stringify offset traces
trace_string_x(drf{T},S) :- trace_string_x(T,TS), str_cat(TS,"*",S).
trace_string_x(fld{T,F,_},S) :- trace_string_x(T,TS), str_cat3(TS,".",F,S).
trace_string_x(rfld{T,F,_},S) :- trace_string_x(T,TS), str_cat3(TS,"^",F,S).
trace_string_x(index{T,Y,I},S) :- trace_string_x(T,TS), scalar_string_x(I,IS),
    tostring(Y,YS), str_cat_list([TS,"[",IS,",",YS,"]"],S).
trace_string_x(uk_index{T,_},S) :- trace_string_x(T,TS),
    str_cat_list(["uk_index{",TS,"}"],S).
trace_string_x(uk_recurse{T,RL},S) :- trace_string_x(T,TS),
    str_cat_list(["uk_recurse{",TS,"}"],S).

trace_string_x(uc_exp{P,E,_},S) :-
    (point_location(P,_,LINE), tostring(LINE,LS);
     ~point_location(P,_,_), LS="?"),
    (exp_string(E,ES); ~exp_string(E,_), ES="?"),
    str_cat_list(["uc_exp{",LS,",",ES,"}"],S).
trace_string_x(uc_val{P,T},S) :-
    (point_location(P,_,LINE), tostring(LINE,LS);
     ~point_location(P,_,_), LS="?"),
    trace_string_x(T,TS),
    str_cat_list(["uc_val{",LS,",",TS,"}"],S).
trace_string_x(uc_sum{I,P,T},S) :-
    (point_location(P,_,LINE), tostring(LINE,LS);
     ~point_location(P,_,_), LS="?"),
    trace_string_x(T,TS), tostring(I,IS),
    str_cat_list(["uc_sum{",IS,",",LS,",",TS,"}"],S).
+trace_string_x(empty,"").

dstring_sum(s_func,STR) :-
    cil_curfn(FN), cleanup_glob_string(FN,STR).
dstring_sum(s_loop{PH},STR) :-
    cil_stmt_x_location(PH,_,LINE), tostring(LINE,LS), str_cat("loop:",LS,STR).
dstring_sum(s_call{I},STR) :-
    cil_instr_x_location(I,_,LINE), tostring(LINE,LS),
    cil_instr_call(I,E), exp_string(E,ES),
    str_cat_list(["call:",LS,":",ES],STR).
dstring_sum(s_asm{I},STR) :-
    cil_instr_x_location(I,_,LINE), tostring(LINE,LS),
    str_cat("asm:",LS,STR).

% converting traces from/to composite-relative traces

comp_from_relative(STR,RT,T) :-
    trace_relative(RT), trace_compose(STR,RT,T).
comp_from_relative(STR,RT,RT) :- ~trace_relative(RT).
comp_to_relative(STR,T,RT) :- trace_sub(T,STR,RT).
comp_to_relative(STR,T,T) :- ~trace_sub(T,STR,_), trace_root(T,glob{_}).
comp_to_relative(rfld{STR,F,C},T,RT) :- comp_to_relative(STR,T,RXT),
    trace_relative(RXT), trace_compose(fld{empty,F,C},RXT,RT).
