
open Clpautil
open Spec
open Specio
module IO = Streamio

module A = Array
module H = Hashtbl

type fn = int

module StrHashType = struct
  type t = string
  let equal (s1:string) (s2:string) = (s1 = s2)
  let hash = H.hash
end
module StrHash = Hashcons.Make(StrHashType)
let id2str = StrHash.id2a
let str2id = StrHash.a2id

module StrListHashType = struct
  type t = int list
  let equal (s1:int list) (s2:int list) = (s1 = s2)
  let hash = H.hash
end
module StrListHash = Hashcons.Make(StrListHashType)
let id2fnlist = StrListHash.id2a
let fnlist2id = StrListHash.a2id

let val2fn = o str2id val2str
let fn2val = o str2val id2str

let cxt2val = o (list2val None) (List.map fn2val)

module VHashedType = struct
  type t = int
  let equal (v1 : int) (v2 : int) = (v1 = v2)
  let hash = H.hash
end
module FH = H.Make(VHashedType)
module FLH = H.Make(VHashedType)

let functions : int ref FH.t = FH.create 100  (* int is context depth *)
let function_callees: (fn * fn list) FH.t = FH.create 100
let function_callers: (fn * fn list) FH.t = FH.create 100

let rec fill_context_depth fn cxt =
  if not (FH.mem functions fn) then
    FH.add functions fn (ref 0);
  let depth = FH.find functions fn in
    depth := max !depth (List.length cxt);
    match cxt with
        [] -> ()
      | pfn::xcxt -> fill_context_depth pfn xcxt

let filled_functions : bool ref = ref false
let fill_functions() =
  if not !filled_functions then begin
    filled_functions := true;
    let snames = bdb_load_session_list "cil_callee_context" in
    List.iter (fun pname ->
      let sess_str = get_some (bdb_load_session pname) in
      let sess = get_some (load_session_str sess_str) in
      let (_,nargs) = pred2rep (get_ext_session_name sess) in
      let func = val2fn nargs.(0) in
        iter_ext_logic sess (fun ppred ->
          let (_,pargs) = pred2rep ppred in
          let cxt = List.map val2fn (val2list pargs.(0)) in
          let cfunc = val2fn pargs.(1) in
            fill_context_depth func cxt;
            FH.add function_callees func (cfunc,cxt);
            FH.add function_callers cfunc (func,cxt)))
      snames
  end

let rec adjust_context (func,cxt) =
  let depth = try !(FH.find functions func) with Not_found -> 0 in
   if depth = List.length cxt then [fnlist2id (func::cxt)]
   else if depth < List.length cxt then
     adjust_context (func, List.rev (List.tl (List.rev cxt)))
   else begin
     let last = if cxt <> [] then List.hd (List.rev cxt) else func in
     let callers = FH.find_all function_callers last in
     let nadjusts = List.map (fun (pfn,_) -> adjust_context (func,cxt@[pfn])) callers in
       List.concat nadjusts
   end

let rec context_match fncxt0 fncxt1 = match (fncxt0,fncxt1) with
    [], _ -> true
  | _, [] -> true
  | (fn0::tail0, fn1::tail1) -> fn0 = fn1 && context_match tail0 tail1  

let scc_reach : int FLH.t = FLH.create 100
let scc_contents : int FLH.t = FLH.create 100

let scc_overlap fncxt (seen_scc: unit FLH.t) =
  List.filter (FLH.mem seen_scc) (FLH.find_all scc_reach fncxt)

let rec crawl (seen: unit FLH.t) (seen_scc: unit FH.t) (fn,cxt) =
  let fncxts = adjust_context (fn,cxt) in
  List.iter (fun fncxt ->
    let xfncxt = id2fnlist fncxt in
    if FLH.mem seen fncxt then ()
    else if FLH.mem scc_contents fncxt then
      let scc_list = FLH.find_all scc_contents fncxt in
        List.iter (fun scc -> FH.replace seen_scc scc ()) scc_list
    else if scc_overlap fncxt seen_scc <> [] then ()
    else begin
      FLH.add seen fncxt ();
      let callees = FLH.find_all function_callees (List.hd xfncxt) in
        List.iter (fun (cfunc,cxt) ->
          assert (List.length xfncxt > List.length cxt);
          if context_match cxt (List.tl xfncxt) then
            crawl seen seen_scc (cfunc,xfncxt)) callees;
      let overlap = scc_overlap fncxt seen_scc in
        List.iter (fun scc -> FLH.add scc_contents fncxt scc) overlap
    end) fncxts

let rec find_path (seen: unit FLH.t) (fn,cxt) target =
  let fncxts = adjust_context (fn,cxt) in
  List.fold_left (fun prev fncxt ->
    let xfncxt = id2fnlist fncxt in
    if prev <> None then prev
    else if FLH.mem seen fncxt then None
    else if List.hd xfncxt = target then Some [target]
    else begin
      FLH.add seen fncxt ();
      let callees = FH.find_all function_callees (List.hd xfncxt) in
        List.fold_left (fun prev (cfunc,cxt) ->
          if prev <> None then prev
          else if context_match cxt (List.tl xfncxt) then
            (match find_path seen (cfunc,xfncxt) target with
                 Some npath -> Some ((List.hd xfncxt)::npath)
               | None -> None)
          else None) None callees
    end) None fncxts

let filled_sccs : fn list option ref = ref None
let fill_sccs scc_list =
  if !filled_sccs = None then begin
    filled_sccs := Some scc_list;
    List.iter (fun scc ->
      let seen = FLH.create 100 in
        crawl seen (FH.create 100) (scc, []);
        FLH.iter (fun fncxt () -> FLH.add scc_reach fncxt scc) seen) scc_list;
    List.iter (fun scc ->
      List.iter (fun fncxt -> FLH.add scc_contents fncxt scc)
        (adjust_context (scc, []))) scc_list;
  end else if scc_list <> get_some !filled_sccs then
    pred_unexpected "Mismatched SCC lists"

let build_all_callees func scc_list =
  fill_sccs scc_list;
  let cxts = adjust_context (func, []) in
  let seen = FLH.create 100 in
  let seen_scc = FH.create 100 in
    crawl seen seen_scc (func,[]);
  let callees = FH.create 100 in
  let callees_hash = ref 0 in
    FLH.iter (fun fncxt () ->
      if scc_overlap fncxt seen_scc = [] then
        let head = List.hd (id2fnlist fncxt) in
          if not (FH.mem callees head) then begin
            callees_hash := !callees_hash + (H.hash (id2str head));
            FH.add callees head ()
          end) seen;
    IO.printf "Crawled %s, found %d callees [%d]" (id2str func) (FH.length callees) !callees_hash;
    FH.iter (fun scc () -> IO.printf " [CallsSCC:%s]" (id2str scc)) seen_scc;
    List.iter (fun scc -> IO.printf " [ReachedSCC:%s]" (id2str scc))
      (rmdups_list compare (list_sort compare (List.concat (List.map (FLH.find_all scc_reach) cxts))));
    IO.printf "\n";
  let ext_cb = get_ext_callbacks() in
  let sess = rep2pred ("cil_all_callee",[|fn2val func|]) in
    ext_cb.cb_clear sess;
    FH.iter (fun scc () ->
      let sccfn = fn2val scc in
      let cpred = rep2pred ("scc_calls",[|sccfn|]) in
        ext_cb.cb_add sess cpred) seen_scc;
    List.iter (fun fncxt ->
      List.iter (fun scc ->
        let sccfn = fn2val scc in
        let ccxt = cxt2val (List.tl (id2fnlist fncxt)) in
        let cpred = rep2pred ("scc_reached",[|sccfn;ccxt|]) in
          ext_cb.cb_add sess cpred)
        (FLH.find_all scc_reach fncxt)) cxts;
    FLH.iter (fun fncxt () ->
      if scc_overlap fncxt seen_scc = [] then
        let cfn = fn2val (List.hd (id2fnlist fncxt)) in
        let ccxt = cxt2val (List.tl (id2fnlist fncxt)) in
        let cpred = rep2pred ("all_callee",[|cfn;ccxt|]) in
          ext_cb.cb_add sess cpred) seen

let print_path func target =
  let seen = FLH.create 100 in
    match find_path seen (func, []) target with
        Some path ->
          IO.printf "Path: %s -> %s\n" (id2str func) (id2str target);
          List.iter (fun func -> IO.printf "  %s\n" (id2str func)) path
      | None ->
          IO.printf "No Path\n"

let print_reachable func =
  let seen = FLH.create 100 in
    crawl seen (FH.create 100) (func, []);
    IO.printf "Reachable: %s\n" (id2str func);
    FLH.iter (fun fncxt () -> IO.printf "  %s\n" (id2str (List.hd (id2fnlist fncxt)))) seen

let find_handlers = ref []
let add_handlers = ref []

let _ = prepend add_handlers
  (("callgraph_build",[|vt_str;vt_list vt_str|]),
   fun (name,args) ->
     let func = val2fn args.(0) in
     let scc_list = List.map val2fn (val2list args.(1)) in
       fill_functions();
       build_all_callees func scc_list)

let _ = prepend find_handlers
  (("callgraph_path_exists",[|vt_str;vt_str|]),
   [[|true;true|],[Sk_zero;Sk_once]],
   fun (name,args) ->
     let func = val2fn args.(0) in
     let target = val2fn args.(1) in
     let seen = FLH.create 100 in
       fill_functions();
       match find_path seen (func, []) target with
           Some _ -> [name,args]
         | None -> [])

let _ = prepend add_handlers
  (("callgraph_print_path",[|vt_str;vt_str|]),
   fun (name,args) ->
     let func = val2fn args.(0) in
     let target = val2fn args.(1) in
       fill_functions();
       print_path func target)

let _ = prepend add_handlers
  (("callgraph_print_reachable",[|vt_str|]),
   fun (name,args) ->
     let func = val2fn args.(0) in
       fill_functions();
       print_reachable func)

let _ = register_package { empty_package with
  p_name = "callgraph";
  p_find_handlers = !find_handlers;
  p_add_handlers = !add_handlers;
}
