(* cilcvt.ml
 * 
 * Code to convert data structures from the CLPA logic database into something
 * more ML like.
 *
 * In particular, we convert CIL expression and lvalue data structures.
 *)

open Spec
open Clpautil

let _ = Cil.msvcMode := false
let _ = Cil.initCIL ()

(* Make a predicate of the form: name(v, V1, V2, V3, ..., Vnargs) *)
let mk_pred v pred nargs = 
  let args = Array.init nargs (function 
    | 0 -> v
    | x -> var2val ("V" ^ (string_of_int x))
  ) in
    rep2pred (pred, args)

let convert_list (session : t_pred option)
    (predicates : (t_pred * (t_val array -> 'a)) list) : 'a list =
  List.flatten
    (List.map (fun (pred, cvt_fn) ->
      List.map (o (o (o cvt_fn snd) pred2rep) snd)
        (logic_corpus_find session pred Sfa_none))
        predicates)

let convert_sumtype (session : t_pred option)
    (predicates : (t_pred * (t_val array -> 'a)) list) : 'a =
    match rmdups_list compare (convert_list session predicates) with
      | [] -> failwith ("No matches found for: " ^ 
               (string_of_option full_string_of_pred session) ^ " " ^
               (sep_string_of_list " " (o full_string_of_pred fst) predicates))
      | [v] -> v
      | _ -> failwith ("Too many matches found for: " ^
               (string_of_option full_string_of_pred session) ^ " " ^
               (sep_string_of_list " " (o full_string_of_pred fst) predicates))

let convert_sumtype_default (session : t_pred option)
    (predicates : (t_pred * (t_val array -> 'a)) list) (default: 'a) : 'a =
    match rmdups_list compare (convert_list session predicates) with
      | [] -> default
      | [v] -> v
      | _ -> failwith ("Too many matches found for: " ^
               (string_of_option full_string_of_pred session) ^ " " ^
               (sep_string_of_list " " (o full_string_of_pred fst) predicates))





let convert_flag (session : t_pred option) (predicate : t_pred) : bool =
  (logic_corpus_find session predicate Sfa_none) <> []
          
let convert_args (session : t_pred option)
    (pred : t_pred) (cvt_fn : t_val array -> int * 'a) : 'a list =
  List.map snd
    (List.sort (fun x y -> compare (fst x) (fst y))
        (List.map (o (o (o cvt_fn snd) pred2rep) snd)
            (logic_corpus_find session pred Sfa_none)))

let cil_of_clpa_constant (v : t_val) : Cil.constant =
  let predicates = [
    (mk_pred v "cil_const_int" 3,
      (fun a -> Cil.CInt64 (int64_of_big_int (val2int a.(2)), 
                           Transcil.val2ikind a.(1), None)) );
    (mk_pred v "cil_const_str" 2,
      (fun a -> Cil.CStr (val2str a.(1))) );
    (mk_pred v "cil_const_char" 2,
      (fun a -> Cil.CChr (char_of_int (val2sint a.(1)))) );
    (mk_pred v "cil_const_real" 3,
      (fun a -> Cil.CReal (val2flt a.(2), Transcil.val2fkind a.(1), None)) )
  ] in
    convert_sumtype None predicates
    
let cil_of_clpa_exp_location (v : t_val) : Cil.exp_location =
  let predicates = [
    (mk_pred v "cil_exp_x_location" 6,
      (fun a -> { Cil.exp_line_range = (val2sint a.(2), val2sint a.(3));
                  Cil.exp_file = val2str a.(1);
                  Cil.exp_byte_range = (val2sint a.(4), val2sint a.(5));
                  Cil.exp_macro_id = -1 })
    )
  ] in
    match convert_list None predicates with
      | [] ->
          { Cil.exp_line_range = (-1, -1);
            Cil.exp_file = "<unknown>";
            Cil.exp_byte_range = (-1, -1);
            Cil.exp_macro_id = -1 
          }
      | [l] -> l
      | _ -> failwith "Too many matching locations for expression"



let unique_id = Counter.init 0

(* Shared structures indexed by t_val *)
let compinfos : (string, Cil.compinfo) Hashtbl.t = Hashtbl.create 10
let enuminfos : (string, Cil.enuminfo) Hashtbl.t = Hashtbl.create 10
let varinfos  : (t_pred option * t_val, Cil.varinfo) Hashtbl.t  = 
  Hashtbl.create 10

(* Shared structures indexed by name *)
let varinfos_name  : (string, Cil.varinfo) Hashtbl.t  = Hashtbl.create 10

let get_enuminfo_by_name = Hashtbl.find enuminfos
let get_varinfo_by_name = Hashtbl.find varinfos_name

let reset () = 
  Hashtbl.clear compinfos;
  Hashtbl.clear enuminfos;
  Hashtbl.clear varinfos;
  Hashtbl.clear varinfos_name

let unknown_loc = {
  Cil.line = -1;
  Cil.file = "unknown";
  Cil.byte = -1;
  Cil.loc_macro_id = -1;
}

let string_of_cil_type t = Pretty.sprint 72 (Cil.d_type () t)

let rec cil_of_clpa_type (session : t_pred option) (v : t_val) : Cil.typ = 
  let attrs = Cil.addAttributes (convert_list session [
    (mk_pred v "cil_type_x_attr" 2, (fun a -> cil_of_clpa_attr session a.(1)))
  ]) [] in
  let predicates = [
    (mk_pred v "cil_type_void" 1,
      (fun a -> Cil.TVoid attrs));
    (mk_pred v "cil_type_int" 2,
      (fun a -> Cil.TInt (Transcil.val2ikind a.(1), attrs)));
    (mk_pred v "cil_type_float" 2,
      (fun a -> Cil.TFloat (Transcil.val2fkind a.(1), attrs)));
    (mk_pred v "cil_type_ptr" 2,
      (fun a -> Cil.TPtr (cil_of_clpa_type session a.(1), attrs)));
    (mk_pred v "cil_type_array" 3,
      (fun a -> 
        let e = try Some (cil_of_clpa_exp a.(2)) with Failure _ -> None in
        Cil.TArray (cil_of_clpa_type session a.(1), e, attrs)));
    (mk_pred v "cil_type_func" 3,
      (fun a -> 
        let args = convert_args session (mk_pred v "cil_type_func_arg" 4)
             (fun aa -> 
               let n = val2sint aa.(1) in
               let funargattrs = Cil.addAttributes (convert_list session [
                 (rep2pred ("cil_type_func_arg_attr", 
                            [|v;aa.(1);var2val "X"|]),
                  fun aaa -> cil_of_clpa_attr session aaa.(2)
                 )
               ]) [] in
               let arg = (val2str aa.(2), cil_of_clpa_type session aa.(3), 
                         funargattrs) in
                 (n, arg)
             )
        in
        Cil.TFun (cil_of_clpa_type session a.(1), Some args, 
                 Builtin.val2bool a.(2), attrs)));
    (mk_pred v "cil_type_named" 3,
      (fun a -> Cil.TNamed (cil_of_clpa_typeinfo session a.(1) a.(2), attrs)));
    (mk_pred v "cil_type_comp" 2,
      (fun a -> Cil.TComp (cil_of_clpa_comp (val2str a.(1)), attrs)));
    (mk_pred v "cil_type_enum" 2,
      (fun a -> Cil.TEnum (cil_of_clpa_enuminfo (val2str a.(1)), attrs)));
    (mk_pred v "cil_type_valist" 1,
      (fun a -> Cil.TBuiltin_va_list attrs));
  ] in
    convert_sumtype session predicates

and cil_of_clpa_typeinfo (session : t_pred option) (vname : t_val) 
    (vtyp : t_val) : Cil.typeinfo = {
  Cil.tname = val2str vname;
  Cil.ttype = cil_of_clpa_type session vtyp;
  Cil.treferenced = true
}

and cil_of_clpa_enuminfo (name : string) : Cil.enuminfo = 
  try Hashtbl.find enuminfos name
  with Not_found ->
    let predicates = [
      (rep2pred ("cil_enum_name", [|var2val "X";str2val name|]),
      (fun a -> (a.(0))))
      ] in
    let esession = Some (rep2pred ("cil_enum", [|str2val name|])) in
    let v = convert_sumtype esession predicates in

    let attrs = convert_list esession [
      (mk_pred v "cil_enum_attr" 2, (fun a -> cil_of_clpa_attr esession a.(1)))
    ] in
    let ei = {
      Cil.ename = name;
      Cil.eitems = [];
      Cil.eattr = Cil.addAttributes attrs [];
      Cil.ereferenced = true;
    } in
      Hashtbl.add enuminfos name ei;
      ei.Cil.eitems <- convert_args esession
        (mk_pred v "cil_enum_item" 4)
        (fun a -> (val2sint a.(1), 
                  (val2str a.(2), cil_of_clpa_exp a.(3), unknown_loc)));
      ei
    
and cil_of_clpa_comp (name : string) : Cil.compinfo = 
  try Hashtbl.find compinfos name
  with Not_found -> 
    let predicates = [
      (rep2pred ("cil_comp_name", [|var2val "X";str2val name;var2val "Y"|]),
      (fun a -> (a.(0), Builtin.val2bool a.(2))))
      ] in
    let csession = Some (rep2pred ("cil_comp", [|str2val name|])) in
    match convert_list csession predicates with
      | [] ->
          (* An undefined struct, to which we can have pointers in C *)
          let ci = {
            Cil.cstruct = true;
            Cil.cname = name;
            Cil.ckey =  Counter.allocate unique_id;
            Cil.cfields = [];
            Cil.cattr = [];
            Cil.cdefined = false;
            Cil.creferenced = true;
          } in
            Hashtbl.add compinfos name ci;
            ci

      | [(v, is_struct)] ->
          (* A struct with a definition *)
          let cattrs = convert_list csession [
            (mk_pred v "cil_comp_attr" 2, 
            (fun a -> cil_of_clpa_attr csession a.(1)))
          ] in
          let ci = {
            Cil.cstruct = is_struct;
            Cil.cname = name;
            Cil.ckey =  Counter.allocate unique_id;
            Cil.cfields = [];
            Cil.cattr = Cil.addAttributes cattrs [];
            Cil.cdefined = true;
            Cil.creferenced = true;
          } in
            Hashtbl.add compinfos name ci;
            ci.Cil.cfields <- convert_args 
              csession (mk_pred v "cil_comp_field" 3)
              (fun a -> (val2sint a.(1), cil_of_clpa_fieldinfo_comp ci a.(2)));
            ci

      | _ -> pred_unexpected "Too many matches for cil_comp_name"

and cil_of_clpa_fieldinfo_comp (comp : Cil.compinfo) (v : t_val) 
    : Cil.fieldinfo = 
  let csession = Some (rep2pred ("cil_comp", [|str2val comp.Cil.cname|])) in
  let fattrs = convert_list csession [
    (mk_pred v "cil_field_attr" 2, (fun a -> cil_of_clpa_attr csession a.(1)))
  ] in
  let predicates = [
    (mk_pred v "cil_field_name" 5,
    (fun a -> {
      Cil.fcomp = comp;
      Cil.fname = val2str a.(1);
      Cil.ftype = cil_of_clpa_type csession a.(2);
      Cil.fbitfield = None; (* XXX: handle bitfields *)
      Cil.fattr = Cil.addAttributes fattrs [];
      Cil.floc = unknown_loc;
    }))
  ] in
    convert_sumtype csession predicates

and cil_of_clpa_exp (v : t_val) : Cil.exp =
  let exp_loc = cil_of_clpa_exp_location v in
  let predicates = [
    (mk_pred v "cil_exp_const" 2,
      (fun a -> Cil.Const (cil_of_clpa_constant a.(1), exp_loc)));
    (mk_pred v "cil_exp_lval" 2,
      (fun a -> Cil.Lval (cil_of_clpa_lval a.(1), exp_loc)));
    (mk_pred v "cil_exp_sizeof" 2,
      (fun a -> Cil.SizeOf (cil_of_clpa_type None a.(1), exp_loc)));
    (mk_pred v "cil_exp_sizeofe" 2,
      (fun a -> Cil.SizeOfE (cil_of_clpa_exp a.(1), exp_loc)));
    (mk_pred v "cil_exp_sizeofstr" 2,
      (fun a -> Cil.SizeOfStr (val2str a.(1), exp_loc)));
    (mk_pred v "cil_exp_alignof" 2,
      (fun a -> Cil.AlignOf (cil_of_clpa_type None a.(1), exp_loc)));
    (mk_pred v "cil_exp_alignofe" 2,
      (fun a -> Cil.AlignOfE (cil_of_clpa_exp a.(1), exp_loc)));
    (mk_pred v "cil_exp_unop" 4,
      (fun a -> Cil.UnOp (Transcil.val2unop a.(1), cil_of_clpa_exp a.(2), 
                         cil_of_clpa_type None a.(3), exp_loc)));
    (mk_pred v "cil_exp_binop" 5,
      (fun a -> Cil.BinOp (Transcil.val2binop a.(1), cil_of_clpa_exp a.(2), 
                          cil_of_clpa_exp a.(3), cil_of_clpa_type None a.(4), 
                          exp_loc)));
    (mk_pred v "cil_exp_cast" 3,
      (fun a -> Cil.CastE (cil_of_clpa_type None a.(2), cil_of_clpa_exp a.(1), 
                          exp_loc)));
    (mk_pred v "cil_exp_addr" 2,
      (fun a -> Cil.AddrOf (cil_of_clpa_lval a.(1), exp_loc)));
    (mk_pred v "cil_exp_start" 2,
      (fun a -> Cil.StartOf (cil_of_clpa_lval a.(1), exp_loc)));
  ] in
    convert_sumtype None predicates

and cil_of_clpa_lval (v : t_val) : Cil.lval = 
  let predicates = [
    (mk_pred v "cil_lval_var" 3, (fun a -> 
      let vinfo = cil_of_clpa_varinfo None a.(1) in
      (Cil.Var vinfo, cil_of_clpa_offset vinfo.Cil.vtype a.(2))));
    (mk_pred v "cil_lval_mem" 3, (fun a -> 
      let exp = cil_of_clpa_exp a.(1) in
      let typ = match Cil.typeOf exp with
        | Cil.TPtr (typ, _) -> typ 
        | typ -> failwith ("Non-pointer expression passed as argument to " ^ 
                              "cil_lval_mem: " ^ (string_of_cil_type typ)) in
        (Cil.Mem exp, cil_of_clpa_offset typ a.(2))));
  ] in
    convert_sumtype None predicates

and cil_of_clpa_offset (typ: Cil.typ) (v : t_val) : Cil.offset =
  let predicates = [
    (mk_pred v "cil_off_none" 1, (fun _ -> Cil.NoOffset));
    (mk_pred v "cil_off_field" 3, (fun a -> 
      match Cil.unrollType typ with 
        | Cil.TComp (ci,_) ->
            let fi = cil_of_clpa_fieldinfo_comp ci a.(1) in
              Cil.Field (fi, cil_of_clpa_offset fi.Cil.ftype a.(2))
        | _ -> failwith 
            ("Field offset taken of non-composite lvalue with type "
              ^ (string_of_cil_type typ))
    ));
    (mk_pred v "cil_off_index" 3, (fun a -> 
      match Cil.unrollType typ with
        | Cil.TArray (etyp,_,_) ->
            Cil.Index (cil_of_clpa_exp a.(1), cil_of_clpa_offset etyp a.(2))
        | _ -> failwith "Index offset taken of non-array lvalue"
    ))
  ] in
    convert_sumtype None predicates

and cil_of_clpa_varinfo (session:t_pred option) (v : t_val) : Cil.varinfo = 
  try Hashtbl.find varinfos (session,v)
  with Not_found ->
    let (name, typ) = convert_sumtype session [
      (mk_pred v "cil_var_name" 3, (fun a -> 
        (val2str a.(1), cil_of_clpa_type session a.(2))))
    ] in
    let storage = convert_sumtype_default session [
      (mk_pred v "cil_var_static" 1, (fun _ -> Cil.Static));
      (mk_pred v "cil_var_register" 1, (fun _ -> Cil.Register));
    ] Cil.NoStorage in
    let attrs = convert_list session [
      (mk_pred v "cil_var_attr" 2, (fun a -> cil_of_clpa_attr None a.(1)))
    ] in
    let vi = {
      Cil.vname = name;
      Cil.vtype = typ;
      Cil.vattr = attrs;
      Cil.vstorage = storage;
      Cil.vglob = convert_flag session (mk_pred v "cil_var_global" 1);
      Cil.vinline = convert_flag session (mk_pred v "cil_var_inline" 1);
      Cil.vdecl = unknown_loc; (* XXX: I'm inaccurate, fix me *)
      Cil.vid = Counter.allocate unique_id;
      Cil.vaddrof = true; (* XXX: I'm inaccurate, fix me *)
      Cil.vreferenced = true;
      Cil.vdescr = Pretty.text name;
      Cil.vdescrpure = true;
    } in
      Hashtbl.add varinfos (session,v) vi;
      if session = None then
        Hashtbl.add varinfos_name name vi;
      vi

and cil_of_clpa_attr (session : t_pred option) (v : t_val) 
    : Cil.attribute = 
  let name = convert_sumtype session [
    (mk_pred v "cil_attr_name" 2, (fun a -> val2str a.(1)))
  ] in
  let args = convert_args session (mk_pred v "cil_attr_arg" 3)
    (fun a -> (val2sint a.(1), cil_of_clpa_attr_arg session a.(2))) in
    Cil.Attr (name, args)

 
and cil_of_clpa_attr_arg (session : t_pred option) (v : t_val) : Cil.attrparam 
    = 
  let predicates = [
    (mk_pred v "cil_attr_arg_int" 2, (fun a -> Cil.AInt (val2sint a.(1))));
    (mk_pred v "cil_attr_arg_str" 2, (fun a -> Cil.AStr (val2str a.(1))));
    (mk_pred v "cil_attr_arg_cons" 2, (fun a -> 
      match cil_of_clpa_attr session a.(1) with
        | Cil.Attr (name, args) -> Cil.ACons (name, args)
    ));
    (mk_pred v "cil_attr_arg_sizeof" 2, 
      (fun a -> Cil.ASizeOf (cil_of_clpa_type session a.(1))));
    (mk_pred v "cil_attr_arg_sizeofe" 2, 
      (fun a -> Cil.ASizeOfE (cil_of_clpa_attr_arg session a.(1))));
    (mk_pred v "cil_attr_arg_alignof" 2, 
      (fun a -> Cil.AAlignOf (cil_of_clpa_type session a.(1))));
    (mk_pred v "cil_attr_arg_alignofe" 2, 
      (fun a -> Cil.AAlignOfE (cil_of_clpa_attr_arg session a.(1))));
    (mk_pred v "cil_attr_arg_unop" 3, 
      (fun a -> Cil.AUnOp (Transcil.val2unop a.(1), 
                           cil_of_clpa_attr_arg session a.(2))
      )
    );
    (mk_pred v "cil_attr_arg_binop" 4, 
      (fun a -> Cil.ABinOp (Transcil.val2binop a.(1), 
                           cil_of_clpa_attr_arg session a.(2),
                           cil_of_clpa_attr_arg session a.(3))
      )
    );
    (mk_pred v "cil_attr_arg_dot" 3, 
      (fun a -> Cil.ADot (cil_of_clpa_attr_arg session a.(1), val2str a.(2))));
  ] in
    convert_sumtype session predicates

let cil_of_clpa_global_var (name:string) : Cil.varinfo =
  let predicates = [
    (rep2pred ("cil_var_name", [|var2val "X";str2val name;var2val "Y"|]),
    (fun a -> (a.(0))))
  ] in
  let gsession = Some (rep2pred ("cil_glob", [|str2val name|])) in
  let v = convert_sumtype gsession predicates in
    cil_of_clpa_varinfo gsession v
      
let cil_of_clpa_fundec (name:string option) : Cil.fundec =
  let session = match name with
    | None -> None
    | Some name -> Some (rep2pred ("cil_body", [|str2val name|])) in
  let (fn, fn_vinfo) = convert_sumtype session [
    (rep2pred ("cil_fundec_name", [|var2val "X";var2val "Y";var2val "Z"|]),
        (fun a -> (a.(0), cil_of_clpa_varinfo session a.(1))))
  ] in
  let formals = convert_args session (mk_pred fn "cil_fundec_formal" 3)
    (fun a -> (val2sint a.(1), cil_of_clpa_varinfo session a.(2))) in
  let locals = convert_list session [mk_pred fn "cil_fundec_local" 2,
    fun a -> cil_of_clpa_varinfo session a.(1)] in
  {
    Cil.svar = fn_vinfo;
    Cil.sformals = formals;
    Cil.slocals = locals;
    Cil.smaxid = 0;
    Cil.sbody = { Cil.battrs = []; Cil.bstmts = [] };
    Cil.smaxstmtid = None;
    Cil.sallstmts = [];
  }
  
let cil_of_clpa_instr (v:t_val) : Cil.instr =
    let predicates = [
      (mk_pred v "cil_instr_set" 3, fun a ->
          Cil.Set (cil_of_clpa_lval a.(1), cil_of_clpa_exp a.(2), unknown_loc)
      );
      (mk_pred v "cil_instr_call" 2, fun a ->
          let fnexp = cil_of_clpa_exp a.(1) in
          let retlv = convert_sumtype_default None [
              mk_pred v "cil_instr_call_ret" 2, fun a -> 
                  Some (cil_of_clpa_lval a.(1))
          ] None in
          let args = convert_args None (mk_pred v "cil_instr_call_arg" 3)
            (fun a -> (val2sint a.(1), cil_of_clpa_exp a.(2))) in
          Cil.Call (retlv, fnexp, args, unknown_loc)
      )             
    ] in
    convert_sumtype None predicates

