(* clpautil.ml
 *
 * Miscellaneous utility functions. 
 *)

module OrderedInt : sig
  type t = int
  val compare : t -> t -> int
  val to_string : t -> string
end
module IntMap : Map.S with type key = int

module OrderedString : sig
  type t = string
  val compare : t -> t -> int
  val to_string : t -> string
end

module StringSet : Set.S with type elt = string
module StringMap : Map.S with type key = string

type ('l, 'r) either =
    | Left of 'l
    | Right of 'r

val id : 'a -> 'a

val o   : ('b -> 'c) -> ('a -> 'b) -> 'a -> 'c
val o3  : ('c -> 'd) -> ('b -> 'c) -> ('a -> 'b) -> 'a -> 'd
val inv : ('a -> 'b -> 'c) -> 'b -> 'a -> 'c

val fst3 : 'a * 'b * 'c -> 'a
val snd3 : 'a * 'b * 'c -> 'b
val thd3 : 'a * 'b * 'c -> 'c

(* Get an option's value, raising Not_found if it is None *)
val get_some : 'a option -> 'a

val prepend : 'a list ref -> 'a -> unit
val swap : 'a * 'a -> 'a * 'a

val bytes_per_word : float

(* for help in using exception stack traces *)

val list_map        : ('a -> 'b) -> 'a list -> 'b list
val list_fold_right : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
val list_sort       : ('a -> 'a -> int) -> 'a list -> 'a list
val list_merge      : ('a -> 'a -> int) -> 'a list -> 'a list -> 'a list
val (@) : 'a list -> 'a list -> 'a list

(* folding *)

(* Split a list into two roughly evenly sized sublists. Preserves ordering. *)
val divide_list_in_2_stable : 'a list -> 'a list * 'a list

(* A combined map and fold_left, like in the Mercury standard library *)
val map_foldl: ('l -> 'a -> 'm * 'a) -> 'l list -> 'a -> 'm list * 'a

(* A combined map and filter, like a Haskell list comprehension *)
val map_filter: ('a -> 'b option) -> 'a list -> 'b list

(* Create a list of n copies of x *)
val repeat : 'a -> int -> 'a list

(* Search for an element satisfying a two-element predicate in a pair of
 * corresponding lists, raising Not_found if no such element exists *)
val find2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> 'a * 'b

val fold_string : ('a -> char -> 'a) -> 'a -> string -> 'a
val fold_array2 : ('a -> 'b -> 'c -> 'a) -> 'a -> 'b array -> 'c array -> 'a

val compare_lists : ('a -> 'a -> int) -> 'a list -> 'a list -> int
val compare_arrays : ('a -> 'a -> int) -> 'a array -> 'a array -> int
val exists_array : ('a -> bool) -> 'a array -> bool
val forall_array : ('a -> bool) -> 'a array -> bool

(* Test for duplicates in a sorted list *)
val list_hasdups : ('a -> 'a -> int) -> 'a list -> bool

(* Remove duplicates from a sorted list *)
val rmdups_list  : ('a -> 'a -> int) -> 'a list -> 'a list

val findopt_list : ('a -> 'b option) -> 'a list -> 'b
val find_num : (int -> bool) -> int -> int -> int
val fold_num : (int -> 'a -> 'a) -> 'a -> int -> int -> 'a
val fold_num_down : (int -> 'a -> 'a) -> 'a -> int -> int -> 'a
val list_iteri: (int -> 'a -> 'b) -> 'a list -> unit


(** list_init n f: Create the list [f 0; f 1; ... ; f (n-1)] *)
val list_init : int -> (int -> 'a) -> 'a list

(* strings *)

val string_map : (char -> string) -> string -> string
val string_find : string -> string -> int
val string_contains : string -> string -> bool
val string_begins : string -> string -> bool

val sep_string_of_list : string -> ('a -> string) -> 'a list -> string
val sep_string_of_array : string -> ('a -> string) -> 'a array -> string

val alloc_string : float -> string
val time_string : float -> string

val unreachable : unit -> 'a
val unimplemented : string -> 'a


(* Big integer <-> int64 conversion *)
val int64_of_big_int : Big_int.big_int -> Int64.t
val big_int_of_int64 : Int64.t -> Big_int.big_int

val string_of_option : ('a -> string) -> 'a option -> string
val string_of_pair : ('a -> string) -> ('b -> string) -> 'a * 'b -> string

(* Execute the function passed as an argument, returning Some x if it
   succeeds and None if it throws Not_found *)
val find_option : (unit -> 'a) -> 'a option
